@@ -612,6 +612,9 @@ void CControlSocket::OnTimer(wxTimerEvent& event)
 	if (m_pCurOpData && m_pCurOpData->waitForAsyncRequest)
 		return;
 
+	if (IsWaitingForLock())
+		return;
+
 	if (m_stopWatch.Time() > (timeout * 1000))
 	{
 		LogMessage(::Error, _("Connection timed out"));
@@ -827,6 +830,15 @@ void CControlSocket::OnObtainLock(wxCommandEvent& event)
 	UnlockCache();
 }
 
+bool CControlSocket::IsWaitingForLock()
+{
+	std::list<t_lockInfo>::iterator own = GetLockStatus();
+	if (own == m_lockInfoList.end())
+		return false;
+
+	return own->waiting == true;
+}
+
 void CControlSocket::InvalidateCurrentWorkingDir(const CServerPath& path)
 {
 	wxASSERT(!path.IsEmpty());
@@ -869,14 +881,12 @@ void CControlSocket::SendAsyncRequest(CAsyncRequestNotification* pNotification)
 // CRealControlSocket
 // ------------------
 
-BEGIN_EVENT_TABLE(CRealControlSocket, CControlSocket)
-	EVT_FZ_SOCKET(wxID_ANY, CRealControlSocket::OnSocketEvent)
-END_EVENT_TABLE()
-
 CRealControlSocket::CRealControlSocket(CFileZillaEnginePrivate *pEngine)
-	: CControlSocket(pEngine), CSocket(this, 0)
+	: CControlSocket(pEngine)
 {
-	m_pBackend = new CSocketBackend(this, this);
+	m_pSocket = new CSocket(this);
+
+	m_pBackend = new CSocketBackend(this, m_pSocket);
 	m_pProxyBackend = 0;
 
 	m_pSendBuffer = 0;
@@ -885,11 +895,13 @@ CRealControlSocket::CRealControlSocket(CFileZillaEnginePrivate *pEngine)
 
 CRealControlSocket::~CRealControlSocket()
 {
-	Close();
+	m_pSocket->Close();
 	if (m_pProxyBackend && m_pProxyBackend != m_pBackend)
 		delete m_pProxyBackend;
 	delete m_pBackend;
 	m_pBackend = 0;
+
+	delete m_pSocket;
 }
 
 bool CRealControlSocket::Send(const char *buffer, int len)
@@ -906,11 +918,10 @@ bool CRealControlSocket::Send(const char *buffer, int len)
 	}
 	else
 	{
-		m_pBackend->Write(buffer, len);
-		int numsent = 0;
-		if (m_pBackend->Error())
+		int error;
+		int written = m_pBackend->Write(buffer, len, error);
+		if (written < 0)
 		{
-			const int error = m_pBackend->LastError();
 			if (error != EAGAIN)
 			{
 				LogMessage(::Error, _("Could not write to socket: %s"), CSocket::GetErrorDescription(error).c_str());
@@ -919,19 +930,17 @@ bool CRealControlSocket::Send(const char *buffer, int len)
 				return false;
 			}
 		}
-		else
-			numsent = m_pBackend->LastCount();
 
-		if (numsent)
+		if (written)
 			m_pEngine->SetActive(false);
 
-		if (numsent < len)
+		if (written < len)
 		{
 			char *tmp = m_pSendBuffer;
-			m_pSendBuffer = new char[m_nSendBufferLen + len - numsent];
+			m_pSendBuffer = new char[m_nSendBufferLen + len - written];
 			memcpy(m_pSendBuffer, tmp, m_nSendBufferLen);
-			memcpy(m_pSendBuffer + m_nSendBufferLen, buffer, len - numsent);
-			m_nSendBufferLen += len - numsent;
+			memcpy(m_pSendBuffer + m_nSendBufferLen, buffer, len - written);
+			m_nSendBufferLen += len - written;
 			delete [] tmp;
 		}
 	}
@@ -944,9 +953,6 @@ void CRealControlSocket::OnSocketEvent(CSocketEvent &event)
 	if (!m_pBackend)
 		return;
 
-	if (event.GetId() != m_pBackend->GetId())
-		return;
-
 	switch (event.GetType())
 	{
 	case CSocketEvent::hostaddress:
@@ -970,7 +976,7 @@ void CRealControlSocket::OnSocketEvent(CSocketEvent &event)
 			if (m_pProxyBackend && !m_pProxyBackend->Detached())
 			{
 				m_pProxyBackend->Detach();
-				m_pBackend = new CSocketBackend(this, this);
+				m_pBackend = new CSocketBackend(this, m_pSocket);
 			}
 			OnConnect();
 		}
@@ -1009,10 +1015,10 @@ void CRealControlSocket::OnSend()
 			return;
 		}
 
-		m_pBackend->Write(m_pSendBuffer, m_nSendBufferLen);
-		if (m_pBackend->Error())
+		int error;
+		int written = m_pBackend->Write(m_pSendBuffer, m_nSendBufferLen, error);
+		if (written < 0)
 		{
-			const int error = m_pBackend->LastError();
 			if (error != EAGAIN)
 			{
 				LogMessage(::Error, _("Could not write to socket: %s"), CSocket::GetErrorDescription(error).c_str());
@@ -1023,15 +1029,13 @@ void CRealControlSocket::OnSend()
 			return;
 		}
 
-		int numsent = m_pBackend->LastCount();
-
-		if (numsent)
+		if (written)
 		{
 			SetAlive();
 			m_pEngine->SetActive(false);
 		}
 
-		if (numsent == m_nSendBufferLen)
+		if (written == m_nSendBufferLen)
 		{
 			m_nSendBufferLen = 0;
 			delete [] m_pSendBuffer;
@@ -1039,8 +1043,8 @@ void CRealControlSocket::OnSend()
 		}
 		else
 		{
-			memmove(m_pSendBuffer, m_pSendBuffer + numsent, m_nSendBufferLen - numsent);
-			m_nSendBufferLen -= numsent;
+			memmove(m_pSendBuffer, m_pSendBuffer + written, m_nSendBufferLen - written);
+			m_nSendBufferLen -= written;
 		}
 	}
 }
@@ -1090,7 +1094,7 @@ int CRealControlSocket::ContinueConnect()
 		port = m_pEngine->GetOptions()->GetOptionVal(OPTION_PROXY_PORT);
 
 		delete m_pBackend;
-		m_pProxyBackend = new CProxySocket(this, this, this);
+		m_pProxyBackend = new CProxySocket(this, m_pSocket, this);
 		m_pBackend = m_pProxyBackend;
 		int res = m_pProxyBackend->Handshake((enum CProxySocket::ProxyType)proxy_type,
 											 m_pCurrentServer->GetHost(), m_pCurrentServer->GetPort(),
@@ -1121,7 +1125,7 @@ int CRealControlSocket::ContinueConnect()
 	if (!IsIpAddress(host))
 		LogMessage(Status, _("Resolving address of %s"), host.c_str());
 
-	int res = CSocket::Connect(host, port);
+	int res = m_pSocket->Connect(host, port);
 
 	// Treat success same as EINPROGRESS, we wait for connect notification in any case
 	if (res && res != EINPROGRESS)
@@ -1143,7 +1147,7 @@ int CRealControlSocket::DoClose(int nErrorCode /*=FZ_REPLY_DISCONNECTED*/)
 
 void CRealControlSocket::ResetSocket()
 {
-	Close();
+	m_pSocket->Close();
 
 	if (m_pSendBuffer)
 	{
