@@ -4,7 +4,6 @@
 #include "socket.h"
 #include "logging_private.h"
 #include "backend.h"
-#include "socket.h"
 
 class COpData
 {
@@ -82,6 +81,7 @@ public:
 		: COpData(cmd_cwd)
 	{
 		tryMkdOnFail = false;
+		link_discovery = false;
 	}
 
 	virtual ~CChangeDirOpData()
@@ -92,6 +92,8 @@ public:
 	wxString subDir;
 	bool tryMkdOnFail;
 	CServerPath target;
+
+	bool link_discovery;
 };
 
 enum TransferEndReason
@@ -118,7 +120,7 @@ public:
 	virtual int Connect(const CServer &server) = 0;
 	virtual int Disconnect();
 	virtual void Cancel();
-	virtual int List(CServerPath path = CServerPath(), wxString subDir = _T(""), bool refresh = false) { return FZ_REPLY_NOTSUPPORTED; }
+	virtual int List(CServerPath path = CServerPath(), wxString subDir = _T(""), int flags = 0) { return FZ_REPLY_NOTSUPPORTED; }
 	virtual int FileTransfer(const wxString localFile, const CServerPath &remotePath,
 							 const wxString &remoteFile, bool download,
 							 const CFileTransferCommand::t_transferSettings& transferSettings) { return FZ_REPLY_NOTSUPPORTED; }
@@ -235,6 +237,8 @@ protected:
 	// On failure, the engine was not waiting for a lock.
 	enum locking_reason ObtainLockFromEvent();
 
+	bool IsWaitingForLock();
+
 #ifdef __VISUALC__
 	// Retarded compiler does not like my code
 	public:
@@ -266,7 +270,7 @@ protected:
 };
 
 class CProxySocket;
-class CRealControlSocket : public CControlSocket, public CSocket
+class CRealControlSocket : public CControlSocket, public CSocketEventHandler
 {
 public:
 	CRealControlSocket(CFileZillaEnginePrivate *pEngine);
@@ -275,13 +279,12 @@ public:
 	virtual int Connect(const CServer &server);
 	virtual int ContinueConnect();
 
-	virtual bool Connected() { return GetState() == CSocket::connected; }
+	virtual bool Connected() { return m_pSocket->GetState() == CSocket::connected; }
 
 protected:
 	virtual int DoClose(int nErrorCode = FZ_REPLY_DISCONNECTED);
 	void ResetSocket();
 
-	DECLARE_EVENT_TABLE();
 	virtual void OnSocketEvent(CSocketEvent &event);
 	virtual void OnConnect();
 	virtual void OnReceive();
@@ -290,6 +293,8 @@ protected:
 	
 	virtual bool Send(const char *buffer, int len);
 
+	CSocket* m_pSocket;
+
 	CBackend* m_pBackend;
 	CProxySocket* m_pProxyBackend;
 
