@@ -69,7 +69,7 @@ public:
 		if (type > 2)
 			return _T("");
 
-		if (m_str[type] != _T(""))
+		if (!m_str[type].IsEmpty())
 			return m_str[type];
 
 		if (!type)
@@ -317,6 +317,9 @@ public:
 			m_len = wxStrlen(p);
 
 		m_parsePos = 0;
+
+		m_Tokens.reserve(10);
+		m_LineEndTokens.reserve(10);
 	}
 
 	~CLine()
@@ -575,6 +578,16 @@ CDirectoryListingParser::CDirectoryListingParser(CControlSocket* pControlSocket,
 		//Slovenian month names
 		m_MonthNamesMap[_T("avg")] = 8;
 
+		//Icelandic
+#if wxUSE_UNICODE
+		m_MonthNamesMap[_T("ma\x00ed")] = 5;
+		m_MonthNamesMap[_T("j\x00fan")] = 6;
+		m_MonthNamesMap[_T("j\x00fal")] = 7;
+		m_MonthNamesMap[_T("\x00e1g")] = 8;
+		m_MonthNamesMap[_T("n\x00f3v")] = 11;
+#endif
+		m_MonthNamesMap[_T("des")] = 12;
+
 		//There are more languages and thus month
 		//names, but as long as nobody reports a
 		//problem, I won't add them, there are way
@@ -698,7 +711,6 @@ CDirectoryListing CDirectoryListingParser::Parse(const CServerPath &path)
 	}
 	else
 	{
-		listing.SetCount(m_entryList.size());
 		listing.Assign(m_entryList);
 	}
 
@@ -723,7 +735,7 @@ bool CDirectoryListingParser::ParseLine(CLine *pLine, const enum ServerType serv
 			goto done;
 	}
 
-	res = ParseAsUnix(pLine, entry);
+	res = ParseAsUnix(pLine, entry, true); // Common 'ls -l'
 	if (res)
 		goto done;
 	res = ParseAsDos(pLine, entry);
@@ -766,7 +778,13 @@ bool CDirectoryListingParser::ParseLine(CLine *pLine, const enum ServerType serv
 		res = ParseAsIBM_MVS_PDS2(pLine, entry);
 		if (res)
 			goto done;
+		res = ParseAsIBM_MVS_Tape(pLine, entry);
+		if (res)
+			goto done;
 	}
+	res = ParseAsUnix(pLine, entry, false); // 'ls -l' but without the date/time
+	if (res)
+		goto done;
 
 	// Some servers just send a list of filenames. If a line could not be parsed,
 	// check if it's a filename. If that's the case, store it for later, else clear
@@ -826,7 +844,7 @@ done:
 	return true;
 }
 
-bool CDirectoryListingParser::ParseAsUnix(CLine *pLine, CDirentry &entry)
+bool CDirectoryListingParser::ParseAsUnix(CLine *pLine, CDirentry &entry, bool expect_date)
 {
 	int index = 0;
 	CToken token;
@@ -912,8 +930,16 @@ bool CDirectoryListingParser::ParseAsUnix(CLine *pLine, CDirentry &entry)
 			entry.ownerGroup += token.GetString(1);
 		}
 
-		if (!ParseUnixDateTime(pLine, index, entry))
-			continue;
+		if (expect_date)
+		{
+			if (!ParseUnixDateTime(pLine, index, entry))
+				continue;
+		}
+		else
+		{
+			entry.hasDate = false;
+			entry.hasTime = false;
+		}
 
 		// Get the filename
 		if (!pLine->GetToken(++index, token, 1))
@@ -1087,11 +1113,8 @@ bool CDirectoryListingParser::ParseUnixDateTime(CLine *pLine, int &index, CDiren
 		// Check month name
 		while (strMonth.Right(1) == _T(",") || strMonth.Right(1) == _T("."))
 			strMonth.RemoveLast();
-		strMonth.MakeLower();
-		std::map<wxString, int>::iterator iter = m_MonthNamesMap.find(strMonth);
-		if (iter == m_MonthNamesMap.end())
+		if (!GetMonthFromName(strMonth, month))
 			return false;
-		month = iter->second;
 	}
 
 	// Get time/year field
@@ -1223,11 +1246,8 @@ bool CDirectoryListingParser::ParseShortDate(CToken &token, CDirentry &entry, bo
 
 		// Check month name
 		wxString dateMonth = token.GetString().Mid(0, pos);
-		dateMonth.MakeLower();
-		std::map<wxString, int>::iterator iter = m_MonthNamesMap.find(dateMonth);
-		if (iter == m_MonthNamesMap.end())
+		if (!GetMonthFromName(dateMonth, month))
 			return false;
-		month = iter->second;
 		gotMonth = true;
 		gotMonthName = true;
 	}
@@ -1312,11 +1332,8 @@ bool CDirectoryListingParser::ParseShortDate(CToken &token, CDirentry &entry, bo
 		// Month field in yyyy-mm-dd or dd-mm-yyyy
 		// Check month name
 		wxString dateMonth = token.GetString().Mid(pos + 1, pos2 - pos - 1);
-		dateMonth.MakeLower();
-		std::map<wxString, int>::iterator iter = m_MonthNamesMap.find(dateMonth);
-		if (iter == m_MonthNamesMap.end())
+		if (!GetMonthFromName(dateMonth, month))
 			return false;
-		month = iter->second;
 		gotMonth = true;
 	}
 	else
@@ -1715,19 +1732,19 @@ bool CDirectoryListingParser::ParseAsVms(CLine *pLine, CDirentry &entry)
 		const int len = token.GetLength();
 		if (len > 2 && token[0] == '(' && token[len - 1] == ')')
 		{
-			if (entry.permissions != _T(""))
+			if (!entry.permissions.IsEmpty())
 				entry.permissions += _T(" ");
 			entry.permissions += token.GetString().Mid(1, len - 2);
 		}
 		else if (len > 2 && token[0] == '[' && token[len - 1] == ']')
 		{
-			if (entry.ownerGroup != _T(""))
+			if (!entry.ownerGroup.IsEmpty())
 				entry.ownerGroup += _T(" ");
 			entry.ownerGroup += token.GetString().Mid(1, len - 2);
 		}
 		else
 		{
-			if (entry.permissions != _T(""))
+			if (!entry.permissions.IsEmpty())
 				entry.permissions += _T(" ");
 			entry.ownerGroup += token.GetString();
 		}
@@ -1867,9 +1884,8 @@ bool CDirectoryListingParser::ParseOther(CLine *pLine, CDirentry &entry)
 
 		// Get date
 		wxString dateMonth = token.GetString();
-		dateMonth.MakeLower();
-		std::map<wxString, int>::const_iterator iter = m_MonthNamesMap.find(dateMonth);
-		if (iter == m_MonthNamesMap.end())
+		int month = 0;
+		if (!GetMonthFromName(dateMonth, month))
 		{
 			// OS/2 or nortel.VxWorks
 			entry.dir = false;
@@ -1902,7 +1918,9 @@ bool CDirectoryListingParser::ParseOther(CLine *pLine, CDirentry &entry)
 				return false;
 
 			entry.name = token.GetString();
-			if (!skippedCount && entry.name.Right(5).MakeUpper() == _T("<DIR>"))
+			wxString type = entry.name.Right(5);
+			MakeLowerAscii(type);
+			if (!skippedCount && type == _T("<dir>"))
 			{
 				entry.dir = true;
 				entry.name = entry.name.Left(entry.name.Length() - 5);
@@ -1912,8 +1930,6 @@ bool CDirectoryListingParser::ParseOther(CLine *pLine, CDirentry &entry)
 		}
 		else
 		{
-			int month = iter->second;
-
 			// Get day
 			if (!pLine->GetToken(++index, token))
 				return false;
@@ -2124,10 +2140,10 @@ CLine *CDirectoryListingParser::GetLine(bool breakAtEnd /*=false*/)
 		else
 		{
 			wxString str(res, wxConvUTF8);
-			if (str == _T(""))
+			if (str.IsEmpty())
 			{
 				str = wxString(res, wxConvLocal);
-				if (str == _T(""))
+				if (str.IsEmpty())
 					str = wxString(res, wxConvISO8859_1);
 			}
 			buffer = new wxChar[str.Len() + 1];
@@ -2472,6 +2488,41 @@ bool CDirectoryListingParser::ParseAsIBM_MVS_PDS2(CLine *pLine, CDirentry &entry
 	return true;
 }
 
+bool CDirectoryListingParser::ParseAsIBM_MVS_Tape(CLine *pLine, CDirentry &entry)
+{
+	int index = 0;
+	CToken token;
+
+	// volume
+	if (!pLine->GetToken(index++, token))
+		return false;
+
+	// unit
+	if (!pLine->GetToken(index++, token))
+		return false;
+
+	if (token.GetString().CmpNoCase(_T("Tape")))
+		return false;
+
+	// dsname
+	if (!pLine->GetToken(index++, token))
+		return false;
+
+	entry.name = token.GetString();
+	entry.dir = false;
+	entry.link = false;
+	entry.ownerGroup = _T("");
+	entry.permissions = _T("");
+	entry.size = -1;
+	entry.hasDate = false;
+	entry.hasTime = false;
+
+	if (pLine->GetToken(index++, token))
+		return false;
+
+	return true;
+}
+
 bool CDirectoryListingParser::ParseComplexFileSize(CToken& token, wxLongLong& size, int blocksize /*=-1*/)
 {
 	if (token.IsNumeric())
@@ -2575,7 +2626,7 @@ bool CDirectoryListingParser::ParseAsMlsd(CLine *pLine, CDirentry &entry)
 		return false;
 
 	wxString facts = token.GetString();
-	if (facts == _T(""))
+	if (facts.IsEmpty())
 		return false;
 
 	entry.dir = false;
@@ -2586,7 +2637,7 @@ bool CDirectoryListingParser::ParseAsMlsd(CLine *pLine, CDirentry &entry)
 	entry.ownerGroup = _T("");
 	entry.permissions = _T("");
 
-	while (facts != _T(""))
+	while (!facts.IsEmpty())
 	{
 		int delim = facts.Find(';');
 		if (delim < 3)
@@ -2596,7 +2647,8 @@ bool CDirectoryListingParser::ParseAsMlsd(CLine *pLine, CDirentry &entry)
 		if (pos < 1 || (pos + 2) >= delim)
 			return false;
 
-		wxString factname = facts.Left(pos).Lower();
+		wxString factname = facts.Left(pos);
+		MakeLowerAscii(factname);
 		wxString value = facts.Mid(pos + 1, delim - pos - 1);
 		if (factname == _T("type"))
 		{
@@ -2628,7 +2680,7 @@ bool CDirectoryListingParser::ParseAsMlsd(CLine *pLine, CDirentry &entry)
 
 			if (*time)
 			{
-				if (!dateTime.ParseFormat(time, _T("%H%M"), dateTime))
+				if (!dateTime.ParseFormat(time, _T("%H%M%S"), dateTime))
 					return false;
 				entry.hasTime = true;
 			}
@@ -2882,3 +2934,20 @@ bool CDirectoryListingParser::ParseAsHPNonstop(CLine *pLine, CDirentry &entry)
 		
 	return true;
 }
+
+bool CDirectoryListingParser::GetMonthFromName(const wxString& name, int &month)
+{
+	std::map<wxString, int>::iterator iter = m_MonthNamesMap.find(name.Lower());
+	if (iter == m_MonthNamesMap.end())
+	{
+		wxString lower(name);
+		MakeLowerAscii(lower);
+		iter = m_MonthNamesMap.find(lower);
+		if (iter == m_MonthNamesMap.end())
+			return false;
+	}
+	
+	month = iter->second;
+	
+	return true;
+}
