@@ -6,6 +6,7 @@
 #include "logging_private.h"
 #include "httpcontrolsocket.h"
 #include "ratelimiter.h"
+#include "pathcache.h"
 
 class wxFzEngineEvent : public wxEvent
 {
@@ -65,6 +66,9 @@ CFileZillaEnginePrivate::CFileZillaEnginePrivate()
 	m_engineList.push_back(this);
 	m_retryCount = 0;
 
+	static int id = 0;
+	m_engine_id = id++;
+
 	m_pLogging = new CLogging(this);
 }
 
@@ -108,8 +112,10 @@ void CFileZillaEnginePrivate::OnEngineEvent(wxFzEngineEvent &event)
 	case engineCancel:
 		if (!IsBusy())
 			break;
-
-		m_pControlSocket->Cancel();
+		if (m_pControlSocket)
+			m_pControlSocket->Cancel();
+		else if (m_pCurrentCommand)
+			ResetOperation(FZ_REPLY_CANCELED);
 		break;
 	case engineTransferEnd:
 		if (m_pControlSocket)
@@ -333,40 +339,54 @@ int CFileZillaEnginePrivate::List(const CListCommand &command)
 	if (command.GetPath().IsEmpty() && command.GetSubDir() != _T(""))
 		return FZ_REPLY_SYNTAXERROR;
 
-	bool refresh = command.Refresh();
-	if (!command.Refresh() && !command.GetPath().IsEmpty())
+	if (command.GetFlags() & LIST_FLAG_LINK && command.GetSubDir() == _T(""))
+		return FZ_REPLY_SYNTAXERROR;
+
+	bool refresh = (command.GetFlags() & LIST_FLAG_REFRESH) != 0;
+	bool avoid = (command.GetFlags() & LIST_FLAG_AVOID) != 0;
+	if (refresh && avoid)
+		return FZ_REPLY_SYNTAXERROR;
+
+	if (!refresh && !command.GetPath().IsEmpty())
 	{
 		const CServer* pServer = m_pControlSocket->GetCurrentServer();
 		if (pServer)
 		{
-			CDirectoryListing *pListing = new CDirectoryListing;
-			CDirectoryCache cache;
-			bool is_outdated = false;
-			bool found = cache.Lookup(*pListing, *pServer, command.GetPath(), command.GetSubDir(), true, is_outdated);
-			if (found && !is_outdated)
+			CServerPath path(CPathCache::Lookup(*pServer, command.GetPath(), command.GetSubDir()));
+			if (!path.IsEmpty())
 			{
-				if (pListing->m_hasUnsureEntries)
-					refresh = true;
-				else
+				CDirectoryListing *pListing = new CDirectoryListing;
+				CDirectoryCache cache;
+				bool is_outdated = false;
+				bool found = cache.Lookup(*pListing, *pServer, path, true, is_outdated);
+				if (found && !is_outdated)
 				{
-					m_lastListDir = pListing->path;
-					m_lastListTime = wxDateTime::Now();
-					CDirectoryListingNotification *pNotification = new CDirectoryListingNotification(pListing->path);
-					AddNotification(pNotification);
-					delete pListing;
-					return FZ_REPLY_OK;
+					if (pListing->m_hasUnsureEntries)
+						refresh = true;
+					else
+					{
+						if (!avoid)
+						{
+							m_lastListDir = pListing->path;
+							m_lastListTime = wxDateTime::Now();
+							CDirectoryListingNotification *pNotification = new CDirectoryListingNotification(pListing->path);
+							AddNotification(pNotification);
+						}
+						delete pListing;
+						return FZ_REPLY_OK;
+					}
 				}
+				if (is_outdated)
+					refresh = true;
+				delete pListing;
 			}
-			if (is_outdated)
-				refresh = true;
-			delete pListing;
 		}
 	}
 	if (IsBusy())
 		return FZ_REPLY_BUSY;
 
 	m_pCurrentCommand = command.Clone();
-	return m_pControlSocket->List(command.GetPath(), command.GetSubDir(), refresh);
+	return m_pControlSocket->List(command.GetPath(), command.GetSubDir(), command.GetFlags());
 }
 
 int CFileZillaEnginePrivate::FileTransfer(const CFileTransferCommand &command)
