@@ -3,6 +3,8 @@
 #include "httpcontrolsocket.h"
 #include <errno.h>
 
+#define FZ_REPLY_REDIRECTED FZ_REPLY_ALREADYCONNECTED
+
 // Connect is special for HTTP: It is done on a per-command basis, so we need
 // to establish a connection before each command.
 class CHttpConnectOpData : public COpData
@@ -184,29 +186,33 @@ bool CHttpControlSocket::SetAsyncRequestReply(CAsyncRequestNotification *pNotifi
 
 void CHttpControlSocket::OnReceive()
 {
-	if (!m_pRecvBuffer)
-	{
-		m_pRecvBuffer = new char[m_recvBufferLen];
-		m_recvBufferPos = 0;
-	}
-
-	int id = GetId();
+	DoReceive();
+}
 
+int CHttpControlSocket::DoReceive()
+{
 	do
 	{
-		if (GetState() != connected && GetState() != closing)
-			return;
+		const enum CSocket::SocketState state = m_pSocket->GetState();
+		if (state != CSocket::connected && state != CSocket::closing)
+			return 0;
+
+		if (!m_pRecvBuffer)
+		{
+			m_pRecvBuffer = new char[m_recvBufferLen];
+			m_recvBufferPos = 0;
+		}
 
 		unsigned int len = m_recvBufferLen - m_recvBufferPos;
 		int error;
-		int read = Read(m_pRecvBuffer + m_recvBufferPos, len, error);
+		int read = m_pSocket->Read(m_pRecvBuffer + m_recvBufferPos, len, error);
 		if (read == -1)
 		{
 			if (error != EAGAIN)
 			{
 				ResetOperation(FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED);
 			}
-			return;
+			return 0;
 		}
 
 		m_pEngine->SetActive(true);
@@ -215,7 +221,7 @@ void CHttpControlSocket::OnReceive()
 		{
 			// Just ignore all further data
 			m_recvBufferPos = 0;
-			return;
+			return 0;
 		}
 
 		m_recvBufferPos += read;
@@ -225,17 +231,21 @@ void CHttpControlSocket::OnReceive()
 			if (!read)
 			{
 				ResetOperation(FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED);
-				return;
+				return 0;
 			}
 
-			ParseHeader(m_pHttpOpData);
+			int res = ParseHeader(m_pHttpOpData);
+			if ((res & FZ_REPLY_REDIRECTED) == FZ_REPLY_REDIRECTED)
+				return FZ_REPLY_REDIRECTED;
+			if (res != FZ_REPLY_WOULDBLOCK)
+				return 0;
 		}
 		else if (m_pHttpOpData->m_transferEncoding == CHttpOpData::chunked)
 		{
 			if (!read)
 			{
 				ResetOperation(FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED);
-				return;
+				return 0;
 			}
 			OnChunkedData(m_pHttpOpData);
 		}
@@ -245,7 +255,7 @@ void CHttpControlSocket::OnReceive()
 			{
 				wxASSERT(!m_recvBufferPos);
 				ProcessData(0, 0);
-				return;
+				return 0;
 			}
 			else
 			{
@@ -255,7 +265,9 @@ void CHttpControlSocket::OnReceive()
 			}
 		}
 	}
-	while (GetId() == id);
+	while (m_pSocket);
+
+	return 0;
 }
 
 void CHttpControlSocket::OnConnect()
@@ -443,9 +455,9 @@ int CHttpControlSocket::DoInternalConnect()
 
 	if (m_pBackend)
 		delete m_pBackend;
-	m_pBackend = new CSocketBackend(this, this);
+	m_pBackend = new CSocketBackend(this, m_pSocket);
 
-	int res = CSocket::Connect(pData->host, pData->port);
+	int res = m_pSocket->Connect(pData->host, pData->port);
 	if (!res)
 		return FZ_REPLY_OK;
 
@@ -633,7 +645,10 @@ int CHttpControlSocket::ParseHeader(CHttpOpData* pData)
 					// International domain names
 					host = ConvertDomainName(host);
 
-					return InternalConnect(host, port);
+					int res = InternalConnect(host, port);
+					if (res == FZ_REPLY_WOULDBLOCK)
+						res |= FZ_REPLY_REDIRECTED;
+					return res;
 				}
 
 				pData->m_gotHeader = true;
@@ -789,12 +804,12 @@ int CHttpControlSocket::OnChunkedData(CHttpOpData* pData)
 				}
 				else if (*q >= 'A' && *q <= 'F')
 				{
-					pData->m_chunkData.size *= 10;
+					pData->m_chunkData.size *= 16;
 					pData->m_chunkData.size += *q - 'A' + 10;
 				}
 				else if (*q >= 'a' && *q <= 'f')
 				{
-					pData->m_chunkData.size *= 10;
+					pData->m_chunkData.size *= 16;
 					pData->m_chunkData.size += *q - 'a' + 10;
 				}
 				else if (*q == ';' || *q == ' ')
@@ -852,7 +867,7 @@ int CHttpControlSocket::ResetOperation(int nErrorCode)
 
 void CHttpControlSocket::OnClose(int error)
 {
-	LogMessage(Debug_Verbose, _T("CRealControlSocket::OnClose(%d)"), error);
+	LogMessage(Debug_Verbose, _T("CHttpControlSocket::OnClose(%d)"), error);
 
 	if (error)
 	{
@@ -861,7 +876,8 @@ void CHttpControlSocket::OnClose(int error)
 		return;
 	}
 	
-	OnReceive();
+	if (DoReceive() == FZ_REPLY_REDIRECTED)
+		return; // Socket got closed already
 
 	// HTTP socket isn't connected outside operations
 	if (!m_pCurOpData)
@@ -903,7 +919,7 @@ void CHttpControlSocket::ResetHttpData(CHttpOpData* pData)
 {
 	wxASSERT(pData);
 
-	delete m_pRecvBuffer;
+	delete [] m_pRecvBuffer;
 	m_pRecvBuffer = 0;
 
 	pData->m_gotHeader = false;
