@@ -389,6 +389,7 @@ int CSftpControlSocket::Connect(const CServer &server)
 	m_pid = wxExecute(executable + _T(" -v"), wxEXEC_ASYNC, m_pProcess);
 	if (!m_pid)
 	{
+		LogMessage(Debug_Warning, _T("wxExecute failed"));
 		delete m_pProcess;
 		m_pProcess = 0;
 		DoClose();
@@ -398,6 +399,7 @@ int CSftpControlSocket::Connect(const CServer &server)
 	m_pInputThread = new CSftpInputThread(this, m_pProcess);
 	if (!m_pInputThread->Init())
 	{
+		LogMessage(Debug_Warning, _T("Thread creation failed"));
 		delete m_pInputThread;
 		m_pInputThread = 0;
 		m_pProcess->Detach();
@@ -512,10 +514,16 @@ int CSftpControlSocket::ConnectSend()
 			wxString user = m_pEngine->GetOptions()->GetOption(OPTION_PROXY_USER);
 			if (user != _T(""))
 				cmd += _T(" \"") + user + _T("\"");
+
+			wxString show = cmd;
+
 			wxString pass = m_pEngine->GetOptions()->GetOption(OPTION_PROXY_PASS);
 			if (pass != _T(""))
+			{
 				cmd += _T(" \"") + pass + _T("\"");
-			res = Send(cmd);
+				show += _T(" \"") + wxString('*', pass.Len()) + _T("\"");
+			}
+			res = Send(cmd, show);
 		}
 		break;
 	case connect_keys:
@@ -620,6 +628,9 @@ void CSftpControlSocket::OnSftpEvent(wxCommandEvent& event)
 						else
 							LogMessage(::Error, _("Server sent an additional login prompt. You need to use the interactive login type."));
 						DoClose(FZ_REPLY_CRITICALERROR | FZ_REPLY_PASSWORDFAILED);
+
+						for (;iter != messages.end(); iter++)
+							delete *iter;
 						return;
 					}
 
@@ -737,7 +748,7 @@ bool CSftpControlSocket::Send(wxString cmd, const wxString& show /*=_T("")*/)
 {
 	SetWait(true);
 
-	if (show != _T(""))
+	if (!show.empty())
 		LogMessageRaw(Command, show);
 	else
 		LogMessageRaw(Command, cmd);
@@ -864,6 +875,7 @@ public:
 	{
 		pParser = 0;
 		mtime_index = 0;
+		fallback_to_current = false;
 	}
 
 	virtual ~CSftpListOpData()
@@ -879,6 +891,7 @@ public:
 	// Set to true to get a directory listing even if a cache
 	// lookup can be made after finding out true remote directory
 	bool refresh;
+	bool fallback_to_current;
 
 	CDirectoryListing directoryListing;
 	int mtime_index;
@@ -892,7 +905,7 @@ enum listStates
 	list_mtime
 };
 
-int CSftpControlSocket::List(CServerPath path /*=CServerPath()*/, wxString subDir /*=_T("")*/, bool refresh /*=false*/)
+int CSftpControlSocket::List(CServerPath path /*=CServerPath()*/, wxString subDir /*=_T("")*/, int flags /*=0*/)
 {
 	LogMessage(Status, _("Retrieving directory listing..."));
 
@@ -918,9 +931,10 @@ int CSftpControlSocket::List(CServerPath path /*=CServerPath()*/, wxString subDi
 		path.SetType(m_pCurrentServer->GetType());
 	pData->path = path;
 	pData->subDir = subDir;
-	pData->refresh = refresh;
+	pData->refresh = (flags & LIST_FLAG_REFRESH) != 0;
+	pData->fallback_to_current = (flags & LIST_FLAG_FALLBACK_CURRENT) != 0;
 
-	int res = ChangeDir(path, subDir);
+	int res = ChangeDir(path, subDir, (flags & LIST_FLAG_LINK) != 0);
 	if (res != FZ_REPLY_OK)
 		return res;
 
@@ -968,7 +982,7 @@ int CSftpControlSocket::ListParseResponse(bool successful, const wxString& reply
 			return res;
 
 		CDirectoryCache cache;
-		cache.Store(pData->directoryListing, *m_pCurrentServer, pData->path, pData->subDir);
+		cache.Store(pData->directoryListing, *m_pCurrentServer);
 
 		m_pEngine->SendDirectoryListingNotification(m_CurrentPath, !pData->pNextOpData, true, false);
 
@@ -1031,7 +1045,7 @@ int CSftpControlSocket::ListParseResponse(bool successful, const wxString& reply
 		}
 
 		CDirectoryCache cache;
-		cache.Store(pData->directoryListing, *m_pCurrentServer, pData->path, pData->subDir);
+		cache.Store(pData->directoryListing, *m_pCurrentServer);
 
 		m_pEngine->SendDirectoryListingNotification(m_CurrentPath, !pData->pNextOpData, true, false);
 
@@ -1122,8 +1136,21 @@ int CSftpControlSocket::ListSubcommandResult(int prevResult)
 
 	if (prevResult != FZ_REPLY_OK)
 	{
-		ResetOperation(prevResult);
-		return FZ_REPLY_ERROR;
+		if (pData->fallback_to_current)
+		{
+			// List current directory instead
+			pData->fallback_to_current = false;
+			pData->path.Clear();
+			pData->subDir = _T("");
+			int res = ChangeDir();
+			if (res != FZ_REPLY_OK)
+				return res;
+		}
+		else
+		{
+			ResetOperation(prevResult);
+			return FZ_REPLY_ERROR;
+		}
 	}
 
 	if (pData->path.IsEmpty())
@@ -1138,12 +1165,9 @@ int CSftpControlSocket::ListSubcommandResult(int prevResult)
 
 		int hasUnsureEntries;
 		bool is_outdated = false;
-		bool found = cache.DoesExist(*m_pCurrentServer, m_CurrentPath, _T(""), hasUnsureEntries, is_outdated);
+		bool found = cache.DoesExist(*m_pCurrentServer, m_CurrentPath, hasUnsureEntries, is_outdated);
 		if (found)
 		{
-			if (!pData->path.IsEmpty() && pData->subDir != _T(""))
-				cache.AddParent(*m_pCurrentServer, m_CurrentPath, pData->path, pData->subDir);
-
 			// We're done if listins is recent and has no outdated entries
 			if (!is_outdated && !hasUnsureEntries)
 			{
@@ -1217,7 +1241,7 @@ enum cwdStates
 	cwd_cwd_subdir
 };
 
-int CSftpControlSocket::ChangeDir(CServerPath path /*=CServerPath()*/, wxString subDir /*=_T("")*/)
+int CSftpControlSocket::ChangeDir(CServerPath path /*=CServerPath()*/, wxString subDir /*=_T("")*/, bool link_discovery /*=false*/)
 {
 	enum cwdStates state = cwd_init;
 
@@ -1277,6 +1301,7 @@ int CSftpControlSocket::ChangeDir(CServerPath path /*=CServerPath()*/, wxString
 	pData->path = path;
 	pData->subDir = subDir;
 	pData->target = target;
+	pData->link_discovery = link_discovery;
 
 	if (pData->pNextOpData && pData->pNextOpData->opId == cmd_transfer &&
 		!static_cast<CSftpFileTransferOpData *>(pData->pNextOpData)->download)
@@ -1348,7 +1373,16 @@ int CSftpControlSocket::ChangeDirParseResponse(bool successful, const wxString&
 		break;
 	case cwd_cwd_subdir:
 		if (!successful || reply == _T(""))
-			error = true;
+		{
+			if (pData->link_discovery)
+			{
+				LogMessage(Debug_Info, _T("Symlink does not link to a directory, probably a file"));
+				ResetOperation(FZ_REPLY_LINKNOTDIR);
+				return FZ_REPLY_ERROR;
+			}
+			else
+				error = true;
+		}
 		else if (ParsePwdReply(reply))
 		{
 			CPathCache cache;
@@ -1643,7 +1677,7 @@ int CSftpControlSocket::FileTransferSubcommandResult(int prevResult)
 			}
 			if (pData->opState == filetransfer_waitlist)
 			{
-				int res = List(CServerPath(), _T(""), true);
+				int res = List(CServerPath(), _T(""), LIST_FLAG_REFRESH);
 				if (res != FZ_REPLY_OK)
 					return res;
 				ResetOperation(FZ_REPLY_INTERNALERROR);
@@ -1971,14 +2005,16 @@ int CSftpControlSocket::Mkdir(const CServerPath& path)
 		LogMessage(Status, _("Creating directory '%s'..."), path.GetPath().c_str());
 
 	CMkdirOpData *pData = new CMkdirOpData;
-	pData->opState = mkd_findparent;
 	pData->path = path;
 
 	if (!m_CurrentPath.IsEmpty())
 	{
 		// Unless the server is broken, a directory already exists if current directory is a subdir of it.
 		if (m_CurrentPath == path || m_CurrentPath.IsSubdirOf(path, false))
+		{
+			delete pData;
 			return FZ_REPLY_OK;
+		}
 
 		if (m_CurrentPath.IsParentOf(path, false))
 			pData->commonParent = m_CurrentPath;
@@ -1986,11 +2022,18 @@ int CSftpControlSocket::Mkdir(const CServerPath& path)
 			pData->commonParent = path.GetCommonParent(m_CurrentPath);
 	}
 
-	pData->currentPath = path.GetParent();
-	pData->segments.push_back(path.GetLastSegment());
+	if (!path.HasParent())
+		pData->opState = mkd_tryfull;
+	else
+	{
+		pData->currentPath = path.GetParent();
+		pData->segments.push_back(path.GetLastSegment());
 
-	if (pData->currentPath == m_CurrentPath)
-		pData->opState = mkd_mkdsub;
+		if (pData->currentPath == m_CurrentPath)
+			pData->opState = mkd_mkdsub;
+		else
+			pData->opState = mkd_findparent;
+	}
 
 	pData->pNextOpData = m_pCurOpData;
 	m_pCurOpData = pData;
@@ -2260,19 +2303,22 @@ int CSftpControlSocket::RemoveDir(const CServerPath& path /*=CServerPath()*/, co
 	pData->path = path;
 	pData->subDir = subDir;
 
-	CServerPath fullPath = pData->path;
-
-	if (!fullPath.AddSegment(subDir))
+	CServerPath fullPath = CPathCache::Lookup(*m_pCurrentServer, pData->path, pData->subDir);
+	if (fullPath.IsEmpty())
 	{
-		LogMessage(::Error, _("Path cannot be constructed for directory %s and subdir %s"), path.GetPath().c_str(), subDir.c_str());
-		return FZ_REPLY_ERROR;
+		CServerPath fullPath = pData->path;
+
+		if (!fullPath.AddSegment(subDir))
+		{
+			LogMessage(::Error, _("Path cannot be constructed for directory %s and subdir %s"), path.GetPath().c_str(), subDir.c_str());
+			return FZ_REPLY_ERROR;
+		}
 	}
 
 	CDirectoryCache cache;
 	cache.InvalidateFile(*m_pCurrentServer, path, subDir);
 
-	CPathCache pathCache;
-	pathCache.InvalidatePath(*m_pCurrentServer, pData->path, pData->subDir);
+	CPathCache::InvalidatePath(*m_pCurrentServer, pData->path, pData->subDir);
 
 	m_pEngine->InvalidateCurrentWorkingDirs(fullPath);
 	if (!Send(_T("rmdir ") + WildcardEscape(QuoteFilename(fullPath.GetPath())),
@@ -2308,7 +2354,7 @@ int CSftpControlSocket::RemoveDirParseResponse(bool successful, const wxString&
 	}
 
 	CDirectoryCache cache;
-	cache.RemoveDir(*m_pCurrentServer, pData->path, pData->subDir);
+	cache.RemoveDir(*m_pCurrentServer, pData->path, pData->subDir, CPathCache::Lookup(*m_pCurrentServer, pData->path, pData->subDir));
 	m_pEngine->SendDirectoryListingNotification(pData->path, false, true, false);
 
 	return ResetOperation(FZ_REPLY_OK);
@@ -2552,14 +2598,13 @@ int CSftpControlSocket::RenameSend()
 			wxString fromQuoted = QuoteFilename(pData->m_cmd.GetFromPath().FormatFilename(pData->m_cmd.GetFromFile(), !pData->m_useAbsolute));
 			wxString toQuoted = QuoteFilename(pData->m_cmd.GetToPath().FormatFilename(pData->m_cmd.GetToFile(), !pData->m_useAbsolute && pData->m_cmd.GetFromPath() == pData->m_cmd.GetToPath()));
 
-			CPathCache pathCache;
-			pathCache.InvalidatePath(*m_pCurrentServer, pData->m_cmd.GetFromPath(), pData->m_cmd.GetFromFile());
-			pathCache.InvalidatePath(*m_pCurrentServer, pData->m_cmd.GetToPath(), pData->m_cmd.GetToFile());
+			CPathCache::InvalidatePath(*m_pCurrentServer, pData->m_cmd.GetFromPath(), pData->m_cmd.GetFromFile());
+			CPathCache::InvalidatePath(*m_pCurrentServer, pData->m_cmd.GetToPath(), pData->m_cmd.GetToFile());
 
 			if (wasDir)
 			{
 				// Need to invalidate current working directories
-				CServerPath path = pathCache.Lookup(*m_pCurrentServer, pData->m_cmd.GetFromPath(), pData->m_cmd.GetFromFile());
+				CServerPath path = CPathCache::Lookup(*m_pCurrentServer, pData->m_cmd.GetFromPath(), pData->m_cmd.GetFromFile());
 				if (path.IsEmpty())
 				{
 					path = pData->m_cmd.GetFromPath();
@@ -2631,10 +2676,10 @@ void CSftpControlSocket::OnQuotaRequest(enum CRateLimiter::rate_direction direct
 		AddToStream(wxString::Format(_T("-%d%d\n"), (int)direction, b));
 		UpdateUsage(direction, b);
 	}
+	else if (bytes == 0)
+		Wait(direction);
 	else if (bytes < 0)
 		AddToStream(wxString::Format(_T("-%d-\n"), (int)direction));
-	else
-		Wait(direction);
 }
 
 
