@@ -65,6 +65,10 @@ public:
 		if (def != wxDragCopy && def != wxDragMove)
 			return wxDragError;
 
+		CDragDropManager* pDragDropManager = CDragDropManager::Get();
+		if (pDragDropManager)
+			pDragDropManager->pDropTarget = m_pLocalListView;
+
 		wxString subdir;
 		int flags;
 		int hit = m_pLocalListView->HitTest(wxPoint(x, y), flags, 0);
@@ -75,17 +79,14 @@ public:
 				subdir = data->name;
 		}
 
-		wxString dir;
+		CLocalPath dir = m_pLocalListView->m_pState->GetLocalDir();
 		if (subdir != _T(""))
 		{
-			dir = CState::Canonicalize(m_pLocalListView->m_dir, subdir);
-			if (dir == _T(""))
+			if (!dir.ChangePath(subdir))
 				return wxDragError;
 		}
-		else
-			dir = m_pLocalListView->m_dir;
 
-		if (!CState::LocalDirIsWriteable(dir))
+		if (!dir.IsWriteable())
 			return wxDragError;
 
 		if (!GetData())
@@ -187,22 +188,22 @@ public:
 
 		const wxString& subdir = DisplayDropHighlight(wxPoint(x, y));
 
+		CLocalPath dir = m_pLocalListView->m_pState->GetLocalDir();
 		if (subdir == _T(""))
 		{
 			const CDragDropManager* pDragDropManager = CDragDropManager::Get();
 			if (pDragDropManager && pDragDropManager->localParent == m_pLocalListView->m_dir)
 				return wxDragNone;
-
-			if (!CState::LocalDirIsWriteable(m_pLocalListView->m_dir))
-				return wxDragNone;
 		}
 		else
 		{
-			wxString dir = CState::Canonicalize(m_pLocalListView->m_dir, subdir);
-			if (dir == _T("") || !CState::LocalDirIsWriteable(dir))
+			if (!dir.ChangePath(subdir))
 				return wxDragNone;
 		}
 
+		if (!dir.IsWriteable())
+			return wxDragNone;
+
 		if (def == wxDragLink)
 			def = wxDragCopy;
 
@@ -241,6 +242,7 @@ BEGIN_EVENT_TABLE(CLocalListView, CFileListCtrl<CLocalFileData>)
 	EVT_LIST_BEGIN_DRAG(wxID_ANY, CLocalListView::OnBeginDrag)
 	EVT_MENU(XRCID("ID_OPEN"), CLocalListView::OnMenuOpen)
 	EVT_MENU(XRCID("ID_EDIT"), CLocalListView::OnMenuEdit)
+	EVT_MENU(XRCID("ID_ENTER"), CLocalListView::OnMenuEnter)
 END_EVENT_TABLE()
 
 CLocalListView::CLocalListView(wxWindow* pParent, CState *pState, CQueueView *pQueue)
@@ -314,7 +316,7 @@ bool CLocalListView::DisplayDir(wxString dirname)
 	m_fileData.clear();
 	m_indexMapping.clear();
 
-	m_hasParent = CState::LocalDirHasParent(dirname);
+	m_hasParent = CLocalPath(dirname).HasLogicalParent();
 
 	if (m_hasParent)
 	{
@@ -374,7 +376,7 @@ regular_dir:
 		bool wasLink;
 		while (local_filesystem.GetNextFile(data.name, wasLink, data.dir, &data.size, &data.lastModified, &data.attributes))
 		{
-			if (data.name == _T(""))
+			if (data.name.IsEmpty())
 			{
 				wxGetApp().DisplayEncodingWarning();
 				continue;
@@ -432,36 +434,41 @@ regular_dir:
 
 	ReselectItems(selectedNames, focused);
 
-	Refresh();
+	RefreshListOnly();
 
 	return true;
 }
 
-wxString FormatSize(const wxLongLong& size, bool add_bytes_suffix = false)
+wxString FormatSize(const wxLongLong& size, bool add_bytes_suffix, int format, bool thousands_separator, int num_decimal_places)
 {
-	COptions* const pOptions = COptions::Get();
-	const int format = pOptions->GetOptionVal(OPTION_SIZE_FORMAT);
-
 	if (!format)
 	{
-		if (!pOptions->GetOptionVal(OPTION_SIZE_USETHOUSANDSEP))
-			return size.ToString();
-
+		static wxString sep;
+		static bool separator_initialized = false;
+		if (!separator_initialized)
+		{
+			separator_initialized = true;
 #ifdef __WXMSW__
-		wxChar sep[5];
-		int count = ::GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_STHOUSAND, sep, 5);
-		if (!count)
-			return size.ToString();
+			wxChar tmp[5];
+			int count = ::GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_STHOUSAND, tmp, 5);
+			if (count)
+				sep = tmp;
 #else
-		char* chr = nl_langinfo(THOUSEP);
-		if (!chr || !*chr)
-			return size.ToString();
-		wxChar sep = chr[0];
+			char* chr = nl_langinfo(THOUSEP);
+			if (chr && *chr)
+			{
+#if wxUSE_UNICODE
+				sep = wxString(chr, wxConvLibc);
+#else
+				sep = chr;
 #endif
+			}
+#endif
+		}
 
 		wxString tmp = size.ToString();
 		const int len = tmp.Len();
-		if (len <= 3)
+		if (len <= 3 || !thousands_separator || sep.empty())
 		{
 			if (!add_bytes_suffix)
 				return tmp;
@@ -489,31 +496,114 @@ wxString FormatSize(const wxLongLong& size, bool add_bytes_suffix = false)
 		}
 	}
 
+	wxString places;
+
 	int divider;
 	if (format == 3)
 		divider = 1000;
 	else
 		divider = 1024;
 
-	// We always round up. Set to true if there's a reminder
-	bool r2 = false;
-
 	// Exponent (2^(10p) or 10^(3p) depending on option
 	int p = 0;
 
 	wxLongLong r = size;
+	int remainder = 0;
+	bool clipped = false;
 	while (r > divider && p < 6)
 	{
 		const wxLongLong rr = r / divider;
-		if (rr * divider != r)
-			r2 = true;
+		if (remainder != 0)
+			clipped = true;
+		remainder = (r - rr * divider).GetLo();
 		r = rr;
 		p++;
 	}
-	if (r2)
-		r++;
+	if (!num_decimal_places)
+	{
+		if (remainder != 0 || clipped)
+			r++;
+	}
+	else if (p) // Don't add decimal places on exact bytes
+	{
+		if (format != 3)
+		{
+			// Binary, need to convert 1024 into range from 1-1000
+			if (clipped)
+			{
+				remainder++;
+				clipped = false;
+			}
+			remainder = (int)ceil((double)remainder * 1000 / 1024);
+		}
+
+		int max;
+		switch (num_decimal_places)
+		{
+		default:
+		case 1:
+			max = 9;
+			divider = 100;
+			break;
+		case 2:
+			max = 99;
+			divider = 10;
+			break;
+		case 3:
+			max = 999;
+			break;
+		}
+
+		if (num_decimal_places != 3)
+		{
+			if (remainder % divider)
+				clipped = true;
+			remainder /= divider;
+		}
+
+		if (clipped)
+			remainder++;
+		if (remainder > max)
+		{
+			r++;
+			remainder = 0;
+		}
+
+		places.Printf(_T("%d"), remainder);
+		const int len = places.Len();
+		for (int i = len; i < num_decimal_places; i++)
+			places = _T("0") + places;
+	}
 
 	wxString result = r.ToString();
+	if (places != _T(""))
+	{
+#ifdef __WXMSW__
+		wxChar sep[5];
+		int count = ::GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_SDECIMAL, sep, 5);
+		if (!count)
+		{
+			sep[0] = '.';
+			sep[1] = 0;
+		}
+#else
+		wxString sep;
+		char* chr = nl_langinfo(RADIXCHAR);
+		if (!chr || !*chr)
+			sep = _T(".");
+		else
+		{
+#if wxUSE_UNICODE
+			sep = wxString(chr, wxConvLibc);
+#else
+			sep = chr;
+#endif
+		}
+#endif
+
+		result += sep;
+		result += places;
+	}
 	result += ' ';
 	if (!p)
 		return result + _T("B");
@@ -530,6 +620,16 @@ wxString FormatSize(const wxLongLong& size, bool add_bytes_suffix = false)
 	return result;
 }
 
+wxString FormatSize(const wxLongLong& size, bool add_bytes_suffix = false)
+{
+	COptions* const pOptions = COptions::Get();
+	const int format = pOptions->GetOptionVal(OPTION_SIZE_FORMAT);
+	const bool thousands_separator = pOptions->GetOptionVal(OPTION_SIZE_USETHOUSANDSEP) != 0;
+	const int num_decimal_places = pOptions->GetOptionVal(OPTION_SIZE_DECIMALPLACES);
+
+	return FormatSize(size, add_bytes_suffix, format, thousands_separator, num_decimal_places);
+}
+
 // See comment to OnGetItemText
 int CLocalListView::OnGetItemImage(long item) const
 {
@@ -595,17 +695,34 @@ void CLocalListView::OnItemActivated(wxListEvent &event)
 
 	if (data->dir)
 	{
-		if (IsComparing())
-			ExitComparisonMode();
-
-		wxString error;
-		if (!m_pState->SetLocalDir(data->name, &error))
+		const int action = COptions::Get()->GetOptionVal(OPTION_DOUBLECLICK_ACTION_DIRECTORY);
+		if (action == 3)
 		{
-			if (error != _T(""))
-				wxMessageBox(error, _("Failed to change directory"), wxICON_INFORMATION);
-			else
-				wxBell();
+			// No action
+			wxBell();
+			return;
 		}
+
+		if (!action || data->name == _T(".."))
+		{
+			// Enter action
+
+			if (IsComparing())
+				ExitComparisonMode();
+
+			wxString error;
+			if (!m_pState->SetLocalDir(data->name, &error))
+			{
+				if (error != _T(""))
+					wxMessageBox(error, _("Failed to change directory"), wxICON_INFORMATION);
+				else
+					wxBell();
+			}
+			return;
+		}
+
+		wxCommandEvent evt(0, action == 1 ? XRCID("ID_UPLOAD") : XRCID("ID_ADDTOQUEUE"));
+		OnMenuUpload(evt);
 		return;
 	}
 
@@ -615,6 +732,22 @@ void CLocalListView::OnItemActivated(wxListEvent &event)
 		return;
 	}
 
+	const int action = COptions::Get()->GetOptionVal(OPTION_DOUBLECLICK_ACTION_FILE);
+	if (action == 3)
+	{
+		// No action
+		wxBell();
+		return;
+	}
+
+	if (action == 2)
+	{
+		// View / Edit action
+		wxCommandEvent evt;
+		OnMenuEdit(evt);
+		return;
+	}
+
 	const CServer* pServer = m_pState->GetServer();
 	if (!pServer)
 	{
@@ -631,10 +764,47 @@ void CLocalListView::OnItemActivated(wxListEvent &event)
 
 	wxFileName fn(m_dir, data->name);
 
-	m_pQueue->QueueFile(false, false, fn.GetFullPath(), data->name, path, *pServer, data->size);
+	const bool queue_only = action == 1;
+
+	m_pQueue->QueueFile(queue_only, false, fn.GetFullPath(), data->name, path, *pServer, data->size);
 	m_pQueue->QueueFile_Finish(true);
 }
 
+void CLocalListView::OnMenuEnter(wxCommandEvent &event)
+{
+	int item = GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
+	if (item == -1)
+	{
+		wxBell();
+		return;
+	}
+
+	if (GetNextItem(item, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED) != -1)
+	{
+		wxBell();
+		return;
+	}
+
+	CLocalFileData *data = GetData(item);
+	if (!data || !data->dir)
+	{
+		wxBell();
+		return;
+	}
+
+	if (IsComparing())
+		ExitComparisonMode();
+
+	wxString error;
+	if (!m_pState->SetLocalDir(data->name, &error))
+	{
+		if (error != _T(""))
+			wxMessageBox(error, _("Failed to change directory"), wxICON_INFORMATION);
+		else
+			wxBell();
+	}
+}
+
 #ifdef __WXMSW__
 void CLocalListView::DisplayDrives()
 {
@@ -997,9 +1167,9 @@ public:
 		CLocalFileData &data1 = m_fileData[a];
 		CLocalFileData &data2 = m_fileData[b];
 
-		if (data1.fileType == _T(""))
+		if (data1.fileType.IsEmpty())
 			data1.fileType = m_pListView->GetType(data1.name, data1.dir, m_pListView->m_dir);
-		if (data2.fileType == _T(""))
+		if (data2.fileType.IsEmpty())
 			data2.fileType = m_pListView->GetType(data2.name, data2.dir, m_pListView->m_dir);
 
 		CMP(CmpDir, data1, data2);
@@ -1109,6 +1279,7 @@ void CLocalListView::OnContextMenu(wxContextMenuEvent& event)
 	}
 	if (!count || fillCount == count)
 	{
+		pMenu->Delete(XRCID("ID_ENTER"));
 		pMenu->Enable(XRCID("ID_UPLOAD"), false);
 		pMenu->Enable(XRCID("ID_ADDTOQUEUE"), false);
 		pMenu->Enable(XRCID("ID_DELETE"), false);
@@ -1117,12 +1288,19 @@ void CLocalListView::OnContextMenu(wxContextMenuEvent& event)
 	}
 	else if (count > 1)
 	{
+		pMenu->Delete(XRCID("ID_ENTER"));
 		pMenu->Enable(XRCID("ID_RENAME"), false);
 		pMenu->Enable(XRCID("ID_OPEN"), false);
 		pMenu->Enable(XRCID("ID_EDIT"), false);
 	}
-	else if (selectedDir)
-		pMenu->Enable(XRCID("ID_EDIT"), false);
+	else
+	{
+		// Exactly one item selected
+		if (selectedDir)
+			pMenu->Enable(XRCID("ID_EDIT"), false);
+		else
+			pMenu->Delete(XRCID("ID_ENTER"));
+	}
 	if (fillCount)
 		pMenu->Enable(XRCID("ID_OPEN"), false);
 
@@ -1279,7 +1457,7 @@ void CLocalListView::OnMenuRename(wxCommandEvent& event)
 	EditLabel(item);
 }
 
-void CLocalListView::OnChar(wxKeyEvent& event)
+void CLocalListView::OnKeyDown(wxKeyEvent& event)
 {
 	const int code = event.GetKeyCode();
 	if (code == WXK_DELETE)
@@ -1322,30 +1500,9 @@ void CLocalListView::OnChar(wxKeyEvent& event)
 		event.Skip();
 }
 
-void CLocalListView::OnKeyDown(wxKeyEvent& event)
-{
-	const int code = event.GetKeyCode();
-	const int mods = event.GetModifiers();
-	if (code == 'A' && (mods == wxMOD_CMD || mods == (wxMOD_CONTROL | wxMOD_META)))
-	{
-		for (unsigned int i = m_hasParent ? 1 : 0; i < m_indexMapping.size(); i++)
-		{
-			const CLocalFileData *data = GetData(i);
-			if (data && data->flags != fill)
-				SetSelection(i, true);
-			else
-				SetSelection(i, false);
-		}
-		if (m_pFilelistStatusBar)
-			m_pFilelistStatusBar->SelectAll();
-	}
-	else
-		OnChar(event);
-}
-
 void CLocalListView::OnBeginLabelEdit(wxListEvent& event)
 {
-	if (!m_pState->LocalDirIsWriteable(m_dir))
+	if (!m_pState->GetLocalDir().IsWriteable())
 	{
 		event.Veto();
 		return;
@@ -1448,7 +1605,7 @@ void CLocalListView::OnEndLabelEdit(wxListEvent& event)
 		return;
 	}
 
-	if (!m_pState->LocalDirIsWriteable(m_dir))
+	if (!m_pState->GetLocalDir().IsWriteable())
 	{
 		event.Veto();
 		return;
@@ -1547,7 +1704,7 @@ void CLocalListView::ApplyCurrentFilter()
 	ReselectItems(selectedNames, focused);
 
 	if (!IsComparing())
-		Refresh();
+		RefreshListOnly();
 }
 
 std::list<wxString> CLocalListView::RememberSelectedItems(wxString& focused)
@@ -1658,7 +1815,7 @@ void CLocalListView::ReselectItems(const std::list<wxString>& selectedNames, wxS
 void CLocalListView::OnStateChange(enum t_statechange_notifications notification, const wxString& data)
 {
 	if (notification == STATECHANGE_LOCAL_DIR)
-		DisplayDir(m_pState->GetLocalDir());
+		DisplayDir(m_pState->GetLocalDir().GetPath());
 	else if (notification == STATECHANGE_APPLYFILTER)
 		ApplyCurrentFilter();
 	else
@@ -1716,10 +1873,16 @@ void CLocalListView::OnBeginDrag(wxListEvent& event)
 	source.SetData(obj);
 	int res = source.DoDragDrop(wxDrag_AllowMove);
 
+	bool handled_internally = pDragDropManager->pDropTarget != 0;
+
 	pDragDropManager->Release();
 
-	if (res == wxDragCopy || res == wxDragMove)
+	if (!handled_internally && (res == wxDragCopy || res == wxDragMove))
+	{
+		// We only need to refresh local side if the operation got handled
+		// externally, the internal handlers do this for us already
 		m_pState->RefreshLocal();
+	}
 }
 
 void CLocalListView::RefreshFile(const wxString& file)
@@ -1804,7 +1967,7 @@ void CLocalListView::RefreshFile(const wxString& file)
 		{
 			if (m_sortColumn)
 				SortList();
-			Refresh(false);
+			RefreshListOnly(false);
 		}
 		return;
 	}
@@ -1855,7 +2018,7 @@ void CLocalListView::RefreshFile(const wxString& file)
 				prevState = state;
 			}
 		}
-		Refresh();
+		RefreshListOnly();
 	}
 	else
 	{
@@ -1972,7 +2135,7 @@ void CLocalListView::FinishComparison()
 
 	ComparisonRestoreSelections();
 
-	Refresh();
+	RefreshListOnly();
 
 	CComparableListing* pOther = GetOther();
 	if (!pOther)
@@ -2015,7 +2178,7 @@ wxString CLocalListView::GetItemText(int item, unsigned int column)
 		if (data->flags == fill)
 			return _T("");
 
-		if (data->fileType == _T(""))
+		if (data->fileType.IsEmpty())
 			data->fileType = GetType(data->name, data->dir, m_dir);
 
 		return data->fileType;
@@ -2035,14 +2198,14 @@ void CLocalListView::OnMenuEdit(wxCommandEvent& event)
 	const CServer* const pServer = m_pState->GetServer();
 	if (!pServer)
 	{
-		wxBell();
+		wxMessageBox(_("Cannot edit file, not connected to any server."), _("Editing failed"), wxICON_EXCLAMATION);
 		return;
 	}
 
 	const CServerPath& path = m_pState->GetRemotePath();
 	if (path.IsEmpty())
 	{
-		wxBell();
+		wxMessageBox(_("Cannot edit file, remote path unknown."), _("Editing failed"), wxICON_EXCLAMATION);
 		return;
 	}
 
@@ -2083,13 +2246,37 @@ void CLocalListView::OnMenuEdit(wxCommandEvent& event)
 	wxFileName fn(m_dir, data->name);
 
 	bool dangerous = false;
-	if (!pEditHandler->CanOpen(CEditHandler::local, fn.GetFullPath(), dangerous) || dangerous)
+	bool program_exists = false;
+	wxString cmd = pEditHandler->CanOpen(CEditHandler::local, fn.GetFullPath(), dangerous, program_exists);
+	if (cmd.empty())
 	{
-		wxMessageBox(wxString::Format(_("The file '%s' could not be opened:\nNo program has been associated on your system with this file type."), fn.GetFullPath().c_str()), _("Opening failed"), wxICON_EXCLAMATION);
+		CNewAssociationDialog dlg(this);
+		if (!dlg.Show(fn.GetFullName()))
+			return;
+		cmd = pEditHandler->CanOpen(CEditHandler::local, fn.GetFullPath(), dangerous, program_exists);
+		if (cmd.empty())
+		{
+			wxMessageBox(wxString::Format(_("The file '%s' could not be opened:\nNo program has been associated on your system with this file type."), fn.GetFullPath().c_str()), _("Opening failed"), wxICON_EXCLAMATION);
+			return;
+		}
+	}
+	if (!program_exists)
+	{
+		wxString msg = wxString::Format(_("The file '%s' cannot be opened:\nThe associated program (%s) could not be found.\nPlease check your filetype associations."), fn.GetFullPath().c_str(), cmd.c_str());
+		wxMessageBox(msg, _("Cannot edit file"), wxICON_EXCLAMATION);
 		return;
 	}
+	if (dangerous)
+	{
+		int res = wxMessageBox(_("The selected file would be executed directly.\nThis can be dangerous and damage your system.\nDo you really want to continue?"), _("Dangerous filetype"), wxICON_EXCLAMATION | wxYES_NO);
+		if (res != wxYES)
+		{
+			wxBell();
+			return;
+		}
+	}
 
-	CEditHandler::fileState state = pEditHandler->GetFileState(CEditHandler::local, fn.GetFullPath());
+	CEditHandler::fileState state = pEditHandler->GetFileState(fn.GetFullPath());
 	switch (state)
 	{
 	case CEditHandler::upload:
@@ -2104,14 +2291,15 @@ void CLocalListView::OnMenuEdit(wxCommandEvent& event)
 				wxBell();
 				return;
 			}
-			pEditHandler->StartEditing(CEditHandler::local, fn.GetFullPath());
+			pEditHandler->StartEditing(fn.GetFullPath());
 			return;
 		}
 	default:
 		break;
 	}
 
-	if (!pEditHandler->AddFile(CEditHandler::local, fn.GetFullPath(), path, *pServer))
+	wxString file = fn.GetFullPath();
+	if (!pEditHandler->AddFile(CEditHandler::local, file, path, *pServer))
 	{
 		wxMessageBox(wxString::Format(_("The file '%s' could not be opened:\nThe associated command failed"), fn.GetFullPath().c_str()), _("Opening failed"), wxICON_EXCLAMATION);
 		return;
@@ -2161,26 +2349,39 @@ void CLocalListView::OnMenuOpen(wxCommandEvent& event)
 
 	if (data->dir)
 	{
-		wxString dir = CState::Canonicalize(m_dir, data->name);
-		wxLaunchDefaultBrowser(CState::GetAsURL(dir));
+		CLocalPath path(m_dir);
+		if (!path.ChangePath(data->name))
+		{
+			wxBell();
+			return;
+		}
+
+		wxLaunchDefaultBrowser(CState::GetAsURL(path.GetPath()));
 		return;
 	}
 
 	wxFileName fn(m_dir, data->name);
 
-	wxString cmd = pEditHandler->GetSystemOpenCommand(fn.GetFullPath());
+	bool program_exists = false;
+	wxString cmd = pEditHandler->GetSystemOpenCommand(fn.GetFullPath(), program_exists);
 	if (cmd == _T(""))
 	{
 		int pos = data->name.Find('.') == -1;
 		if (pos == -1 || (pos == 0 && data->name.Mid(1).Find('.') == -1))
-			cmd = pEditHandler->GetOpenCommand(fn.GetFullPath());
+			cmd = pEditHandler->GetOpenCommand(fn.GetFullPath(), program_exists);
 	}
 	if (cmd == _T(""))
 	{
 		wxMessageBox(wxString::Format(_("The file '%s' could not be opened:\nNo program has been associated on your system with this file type."), fn.GetFullPath().c_str()), _("Opening failed"), wxICON_EXCLAMATION);
 		return;
 	}
-
+	if (!program_exists)
+	{
+		wxString msg = wxString::Format(_("The file '%s' cannot be opened:\nThe associated program (%s) could not be found.\nPlease check your filetype associations."), fn.GetFullPath().c_str(), cmd.c_str());
+		wxMessageBox(msg, _("Cannot edit file"), wxICON_EXCLAMATION);
+		return;
+	}
+	
 	if (wxExecute(cmd))
 		return;
 
