@@ -24,6 +24,8 @@
 #include "local_filesys.h"
 #include "statusbar.h"
 #include "recursive_operation.h"
+#include "auto_ascii_files.h"
+#include "dragdropmanager.h"
 
 #ifdef _DEBUG
 #define new DEBUG_NEW
@@ -52,6 +54,10 @@ public:
 		if (!GetData())
 			return wxDragError;
 
+		CDragDropManager* pDragDropManager = CDragDropManager::Get();
+		if (pDragDropManager)
+			pDragDropManager->pDropTarget = m_pQueueView;
+
 		if (m_pDataObject->GetReceivedFormat() == m_pFileDataObject->GetFormat())
 		{
 			CState* const pState = m_pQueueView->m_pMainFrame->GetState();
@@ -84,14 +90,12 @@ public:
 				return wxDragNone;
 			}
 
-			const wxString& target = pState->GetLocalDir();
-#ifdef __WXMSW__
-			if (target == _T("\\"))
+			const CLocalPath& target = pState->GetLocalDir();
+			if (!target.IsWriteable())
 			{
 				wxBell();
 				return wxDragNone;
 			}
-#endif
 
 			if (!pState->DownloadDroppedFiles(m_pRemoteDataObject, target, true))
 				return wxDragNone;
@@ -114,6 +118,15 @@ public:
 			return def;
 		}
 
+		CDragDropManager* pDragDropManager = CDragDropManager::Get();
+		if (pDragDropManager && !pDragDropManager->remoteParent.IsEmpty())
+		{
+			// Drag from remote to queue, check if local path is writeable
+			CState* const pState = m_pQueueView->m_pMainFrame->GetState();
+			if (!pState->GetLocalDir().IsWriteable())
+				return wxDragNone;			
+		}
+
 		def = wxDragCopy;
 
 		return def;
@@ -188,21 +201,61 @@ public:
 
 		m_didSendEvent = false;
 		m_threadWaiting = false;
+		m_throttleWait = false;
+		m_processing_entries = false;
+
+		t_dirPair* pair = new t_dirPair;
+		pair->localPath = pFolderItem->GetLocalPath().c_str();
+		pair->remotePath.SetSafePath(pFolderItem->GetRemotePath().GetSafePath().c_str());
+		m_dirsToCheck.push_back(pair);
 	}
 
 	virtual ~CFolderProcessingThread()
 	{
-		for (std::list<t_newEntry*>::iterator iter = m_entryList.begin(); iter != m_entryList.end(); iter++)
+		for (std::list<CFolderProcessingEntry*>::iterator iter = m_entryList.begin(); iter != m_entryList.end(); iter++)
+			delete *iter;
+		for (std::list<t_dirPair*>::iterator iter = m_dirsToCheck.begin(); iter != m_dirsToCheck.end(); iter++)
 			delete *iter;
 	}
 
-	void GetFiles(std::list<t_newEntry*> &entryList)
+	void GetFiles(std::list<CFolderProcessingEntry*> &entryList)
 	{
 		wxASSERT(entryList.empty());
 		wxMutexLocker locker(m_sync);
 		entryList.swap(m_entryList);
 
 		m_didSendEvent = false;
+		m_processing_entries = true;
+
+		if (m_throttleWait)
+		{
+			m_throttleWait = false;
+			m_condition.Signal();
+		}
+	}
+
+	class t_dirPair : public CFolderProcessingEntry
+	{
+	public:
+		t_dirPair() : CFolderProcessingEntry(CFolderProcessingEntry::dir) {}
+		wxString localPath;
+		CServerPath remotePath;
+	};
+
+	void ProcessDirectory(const wxString& localPath, CServerPath remotePath, const wxString& name)
+	{
+		wxMutexLocker locker(m_sync);
+
+		t_dirPair* pair = new t_dirPair;
+
+		{
+			pair->localPath = (localPath + CLocalFileSystem::path_separator + name).c_str();
+
+			remotePath.AddSegment(name);
+			pair->remotePath.SetSafePath(remotePath.GetSafePath().c_str());
+		}
+
+		m_dirsToCheck.push_back(pair);
 
 		if (m_threadWaiting)
 		{
@@ -211,9 +264,30 @@ public:
 		}
 	}
 
+	void CheckFinished()
+	{
+		m_sync.Lock();
+		wxASSERT(m_processing_entries);
+
+		m_processing_entries = false;
+
+		if (m_threadWaiting && (!m_dirsToCheck.empty() || m_entryList.empty()))
+		{
+			m_threadWaiting = false;
+			m_condition.Signal();
+		}
+
+		m_sync.Unlock();
+	}
+
+	CFolderScanItem* GetFolderScanItem()
+	{
+		return m_pFolderItem;
+	}
+
 protected:
 
-	void AddEntry(t_newEntry* entry)
+	void AddEntry(CFolderProcessingEntry* entry)
 	{
 		m_sync.Lock();
 		m_entryList.push_back(entry);
@@ -228,7 +302,7 @@ protected:
 			send = false;
 			if (m_entryList.size() >= 100)
 			{
-				m_threadWaiting = true;
+				m_throttleWait = true;
 				m_condition.Wait();
 			}
 		}
@@ -237,6 +311,9 @@ protected:
 		else
 			send = true;
 
+		if (send)
+			m_didSendEvent = true;
+
 		m_sync.Unlock();
 
 		if (send)
@@ -251,156 +328,110 @@ protected:
 
 	ExitCode Entry()
 	{
+#ifdef __WXDEBUG__
 		wxMutexGuiEnter();
 		wxASSERT(m_pFolderItem->GetTopLevelItem() && m_pFolderItem->GetTopLevelItem()->GetType() == QueueItemType_Server);
 		wxMutexGuiLeave();
+#endif
 
 		wxASSERT(!m_pFolderItem->Download());
 
 		wxString currentLocalPath;
 		CServerPath currentRemotePath;
 
-		wxDir* pDir = 0;
+		CLocalFileSystem localFileSystem;
 
-		while (!TestDestroy())
+		while (!TestDestroy() && !m_pFolderItem->m_remove)
 		{
-			if (m_pFolderItem->m_remove)
-			{
-				wxCommandEvent evt(fzEVT_FOLDERTHREAD_COMPLETE, wxID_ANY);
-				wxPostEvent(m_pOwner, evt);
-
-				delete pDir;
-
-				return 0;
-			}
-			bool found;
-			wxString file;
-			if (!pDir)
-			{
-				if (!m_pFolderItem->m_dirsToCheck.empty())
-				{
-					const CFolderScanItem::t_dirPair& pair = m_pFolderItem->m_dirsToCheck.front();
-
-					wxLogNull nullLog;
-
-					pDir = new wxDir(pair.localPath);
-
-					if (pDir->IsOpened())
-					{
-						currentLocalPath = pair.localPath;
-						currentRemotePath = pair.remotePath;
-
-						found = pDir->GetFirst(&file);
-
-						if (!found)
-						{
-							// Empty directory
-
-							t_newEntry *entry = new t_newEntry;
-							entry->remotePath.SetSafePath(pair.remotePath.GetSafePath().c_str());
-
-							AddEntry(entry);
-						}
-					}
-					else
-						found = false;
-					m_pFolderItem->m_dirsToCheck.pop_front();
-				}
-				else
-				{
-					wxCommandEvent evt(fzEVT_FOLDERTHREAD_COMPLETE, wxID_ANY);
-					wxPostEvent(m_pOwner, evt);
-
-					if (pDir)
-						delete pDir;
-					return 0;
-				}
-			}
-			else
-				found = pDir->GetNext(&file);
-
-			while (found && !TestDestroy() && !m_pFolderItem->m_remove)
-			{
-				const wxString& fullName = currentLocalPath + wxFileName::GetPathSeparator() + file;
-
-				bool isLink;
-				wxLongLong size;
-				int attributes;
-				CLocalFileSystem::local_fileType type = CLocalFileSystem::GetFileInfo(fullName, isLink, &size, 0, &attributes);
-
-				if (type == CLocalFileSystem::dir)
-				{
-					if (isLink)
-					{
-						found = pDir->GetNext(&file);
-						continue;
-					}
-
-					if (!m_filters.FilenameFiltered(file, currentLocalPath, true, -1, true, attributes))
-					{
-						CFolderScanItem::t_dirPair pair;
-						pair.localPath = fullName;
-						pair.remotePath = currentRemotePath;
-						pair.remotePath.AddSegment(file);
-						m_pFolderItem->m_dirsToCheck.push_back(pair);
-					}
-				}
-				else if (type == CLocalFileSystem::file)
-				{
-					if (!m_filters.FilenameFiltered(file, currentLocalPath, false, size, true, attributes))
-					{
-						t_newEntry* entry = new t_newEntry;
-
-						{
-							entry->localFile = fullName.c_str();
-							entry->remoteFile = file.c_str();
-							entry->remotePath.SetSafePath(currentRemotePath.GetSafePath().c_str());
-							entry->size = size;
-						}
-
-						AddEntry(entry);
-					}
-				}
-
-				found = pDir->GetNext(&file);
-			}
-
-			bool send;
 			m_sync.Lock();
-			if (!m_didSendEvent && m_entryList.size())
-				send = true;
-			else
-				send = false;
+			if (m_dirsToCheck.empty())
+			{
+				if (!m_didSendEvent && !m_entryList.empty())
+				{
+					m_didSendEvent = true;
+					m_sync.Unlock();
+					wxCommandEvent evt(fzEVT_FOLDERTHREAD_FILES, wxID_ANY);
+					wxPostEvent(m_pOwner, evt);
+					continue;
+				}
+
+				if (!m_didSendEvent && !m_processing_entries)
+				{
+					m_sync.Unlock();
+					break;
+				}
+				m_threadWaiting = true;
+				m_condition.Wait();
+				if (m_dirsToCheck.empty())
+				{
+					m_sync.Unlock();
+					break;
+				}
+				m_sync.Unlock();
+				continue;
+			}
+
+			const t_dirPair *pair = m_dirsToCheck.front();
+			m_dirsToCheck.pop_front();
+
 			m_sync.Unlock();
 
-			if (send)
+			if (!localFileSystem.BeginFindFiles(pair->localPath, false))
 			{
-				// We send the notification after leaving the critical section, else we
-				// could get into a deadlock. wxWidgets event system does internal
-				// locking.
-				wxCommandEvent evt(fzEVT_FOLDERTHREAD_FILES, wxID_ANY);
-				wxPostEvent(m_pOwner, evt);
+				delete pair;
+				continue;
 			}
 
-			delete pDir;
-			pDir = 0;
+			t_dirPair* pair2 = new t_dirPair;
+	
+			{
+				pair2->localPath = pair->localPath.c_str();
+				pair2->remotePath.SetSafePath(pair->remotePath.GetSafePath().c_str());
+			}
+
+			AddEntry(pair2);
+
+			t_newEntry* entry = new t_newEntry;
+
+			wxString name;
+			bool is_link;
+			bool is_dir;
+			while (localFileSystem.GetNextFile(name, is_link, is_dir, &entry->size, &entry->time, &entry->attributes))
+			{
+				if (is_link)
+					continue;
+				entry->name = name.c_str();
+				entry->dir = is_dir;
+
+				AddEntry(entry);
+
+				entry = new t_newEntry;
+			}
+			delete entry;
+
+			delete pair;
 		}
 
+		wxCommandEvent evt(fzEVT_FOLDERTHREAD_COMPLETE, wxID_ANY);
+		wxPostEvent(m_pOwner, evt);
+
 		return 0;
 	}
 
+	std::list<t_dirPair*> m_dirsToCheck;
+
 	// Access has to be guarded by m_sync
-	std::list<t_newEntry*> m_entryList;
+	std::list<CFolderProcessingEntry*> m_entryList;
 
 	CQueueView* m_pOwner;
 	CFolderScanItem* m_pFolderItem;
 
-	CFilterManager m_filters;
-
 	wxMutex m_sync;
 	wxCondition m_condition;
 	bool m_threadWaiting;
+	bool m_throttleWait;
 	bool m_didSendEvent;
+	bool m_processing_entries;
 };
 
 CQueueView::CQueueView(CQueue* parent, int index, CMainFrame* pMainFrame, CAsyncRequestQueue *pAsyncRequestQueue)
@@ -477,7 +508,10 @@ bool CQueueView::QueueFile(const bool queueOnly, const bool download, const wxSt
 	else
 	{
 		fileItem = new CFileItem(pServerItem, queueOnly, download, localFile, remoteFile, remotePath, size);
-		fileItem->m_transferSettings.binary = ShouldUseBinaryMode(download ? remoteFile : wxFileName(localFile).GetFullName(), remotePath.GetType());
+		if (download)
+			fileItem->m_transferSettings.binary = !CAutoAsciiFiles::TransferRemoteAsAscii(remoteFile, remotePath.GetType());
+		else
+			fileItem->m_transferSettings.binary = !CAutoAsciiFiles::TransferLocalAsAscii(localFile, remotePath.GetType());
 		fileItem->m_edit = edit;
 	}
 
@@ -496,12 +530,15 @@ void CQueueView::QueueFile_Finish(const bool start)
 		m_pMainFrame->GetState()->NotifyHandlers(STATECHANGE_QUEUEPROCESSING);
 	}
 
-	m_waitStatusLineUpdate = true;
-	AdvanceQueue();
-	m_waitStatusLineUpdate = false;
-	UpdateStatusLinePositions();
+	if (m_activeMode)
+	{
+		m_waitStatusLineUpdate = true;
+		AdvanceQueue(false);
+		m_waitStatusLineUpdate = false;
+	}
 
-	Refresh(false);
+	UpdateStatusLinePositions();
+	RefreshListOnly(false);
 }
 
 bool CQueueView::QueueFiles(const bool queueOnly, const wxString& localPath, const CRemoteDataObject& dataObject)
@@ -517,25 +554,12 @@ bool CQueueView::QueueFiles(const bool queueOnly, const wxString& localPath, con
 
 		CFileItem* fileItem;
 		fileItem = new CFileItem(pServerItem, queueOnly, true, localPath + iter->name, iter->name, dataObject.GetServerPath(), iter->size);
-		fileItem->m_transferSettings.binary = ShouldUseBinaryMode(iter->name, dataObject.GetServerPath().GetType());
+		fileItem->m_transferSettings.binary = !CAutoAsciiFiles::TransferRemoteAsAscii(iter->name, dataObject.GetServerPath().GetType());
 
 		InsertItem(pServerItem, fileItem);
 	}
 
-	CommitChanges();
-
-	if (!m_activeMode && !queueOnly)
-	{
-		m_activeMode = 1;
-		m_pMainFrame->GetState()->NotifyHandlers(STATECHANGE_QUEUEPROCESSING);
-	}
-
-	m_waitStatusLineUpdate = true;
-	AdvanceQueue();
-	m_waitStatusLineUpdate = false;
-	UpdateStatusLinePositions();
-
-	Refresh(false);
+	QueueFile_Finish(!queueOnly);
 
 	return true;
 }
@@ -590,6 +614,8 @@ void CQueueView::ProcessNotification(t_EngineData* pEngineData, CNotification* p
 	case nId_logmsg:
 		m_pMainFrame->GetStatusView()->AddToLog(reinterpret_cast<CLogmsgNotification *>(pNotification));
 		delete pNotification;
+		if (COptions::Get()->GetOptionVal(OPTION_MESSAGELOG_POSITION) == 2)
+			m_pQueue->Highlight(3);
 		break;
 	case nId_operation:
 		ProcessReply(pEngineData, reinterpret_cast<COperationNotification*>(pNotification));
@@ -842,6 +868,9 @@ bool CQueueView::TryStartNextTransfer()
 
 		wxRect rect;
 		GetItemRect(lineIndex + 1, rect);
+#ifndef __WXMSW__
+		rect.y -= GetMainWindow()->GetPosition().y;
+#endif
 		m_allowBackgroundErase = false;
 		if (!pEngineData->pStatusLineCtrl)
 			pEngineData->pStatusLineCtrl = new CStatusLineCtrl(this, pEngineData, rect);
@@ -879,19 +908,24 @@ void CQueueView::ProcessReply(t_EngineData* pEngineData, COperationNotification*
 	if ((replyCode & FZ_REPLY_CANCELED) == FZ_REPLY_CANCELED)
 	{
 		enum ResetReason reason;
-		if (pEngineData->pItem && pEngineData->pItem->m_remove)
-			reason = remove;
-		else
+		if (pEngineData->pItem)
 		{
-			if (pEngineData->pItem->GetType() == QueueItemType_File && ((CFileItem*)pEngineData->pItem)->m_madeProgress)
+			if (pEngineData->pItem->m_remove)
+				reason = remove;
+			else
 			{
-				CFileItem* pItem = (CFileItem*)pEngineData->pItem;
-				pItem->m_madeProgress = false;
-				pItem->m_autoResume = true;
+				if (pEngineData->pItem->GetType() == QueueItemType_File && ((CFileItem*)pEngineData->pItem)->m_madeProgress)
+				{
+					CFileItem* pItem = (CFileItem*)pEngineData->pItem;
+					pItem->m_madeProgress = false;
+					pItem->m_autoResume = true;
+				}
+				reason = reset;
 			}
-			reason = reset;
+			pEngineData->pItem->m_statusMessage = _("Interrupted by user");
 		}
-		pEngineData->pItem->m_statusMessage = _("Interrupted by user");
+		else
+			reason = reset;
 		ResetEngine(*pEngineData, reason);
 		return;
 	}
@@ -909,7 +943,12 @@ void CQueueView::ProcessReply(t_EngineData* pEngineData, COperationNotification*
 			pEngineData->state = t_EngineData::none;
 		break;
 	case t_EngineData::connect:
-		if (replyCode == FZ_REPLY_OK)
+		if (!pEngineData->pItem)
+		{
+			ResetEngine(*pEngineData, reset);
+			return;
+		}
+		else if (replyCode == FZ_REPLY_OK)
 		{
 			if (pEngineData->pItem->GetType() == QueueItemType_File)
 				pEngineData->state = t_EngineData::transfer;
@@ -946,6 +985,11 @@ void CQueueView::ProcessReply(t_EngineData* pEngineData, COperationNotification*
 		}
 		break;
 	case t_EngineData::transfer:
+		if (!pEngineData->pItem)
+		{
+			ResetEngine(*pEngineData, reset);
+			return;
+		}
 		if (replyCode == FZ_REPLY_OK)
 		{
 			ResetEngine(*pEngineData, success);
@@ -1063,6 +1107,14 @@ void CQueueView::ResetEngine(t_EngineData& data, const enum ResetReason reason)
 
 	if (data.pItem)
 	{
+		CServerItem* pServerItem = (CServerItem*)data.pItem->GetTopLevelItem();
+		if (pServerItem)
+		{
+			wxASSERT(pServerItem->m_activeCount > 0);
+			if (pServerItem->m_activeCount > 0)
+				pServerItem->m_activeCount--;
+		}
+
 		if (data.pItem->GetType() == QueueItemType_File)
 		{
 			wxASSERT(data.pStatusLineCtrl);
@@ -1091,11 +1143,13 @@ void CQueueView::ResetEngine(t_EngineData& data, const enum ResetReason reason)
 			{
 				CEditHandler* pEditHandler = CEditHandler::Get();
 				wxASSERT(pEditHandler);
-				wxFileName fn(pFileItem->GetLocalFile());
 				if (pFileItem->m_edit == CEditHandler::remote)
-					pEditHandler->FinishTransfer(reason == success, pFileItem->m_edit, fn.GetFullName());
+				{
+					const CServerItem* pServerItem = (const CServerItem*)pFileItem->GetTopLevelItem();
+					pEditHandler->FinishTransfer(reason == success, pFileItem->GetRemoteFile(), pFileItem->GetRemotePath(), pServerItem->GetServer());
+				}
 				else
-					pEditHandler->FinishTransfer(reason == success, pFileItem->m_edit, fn.GetFullPath());
+					pEditHandler->FinishTransfer(reason == success, pFileItem->GetLocalFile());
 				if (reason == success)
 					pFileItem->m_edit = CEditHandler::none;
 			}
@@ -1172,11 +1226,10 @@ void CQueueView::ResetEngine(t_EngineData& data, const enum ResetReason reason)
 		else
 			RemoveItem(data.pItem, true);
 		data.pItem = 0;
-
-		wxASSERT(m_activeCount > 0);
-		if (m_activeCount > 0)
-			m_activeCount--;
 	}
+	wxASSERT(m_activeCount > 0);
+	if (m_activeCount > 0)
+		m_activeCount--;
 	data.active = false;
 
 	if (data.state == t_EngineData::waitprimary)
@@ -1187,13 +1240,6 @@ void CQueueView::ResetEngine(t_EngineData& data, const enum ResetReason reason)
 	}
 
 	data.state = t_EngineData::none;
-	CServerItem* item = GetServerItem(data.lastServer);
-	if (item)
-	{
-		wxASSERT(item->m_activeCount > 0);
-		if (item->m_activeCount > 0)
-			item->m_activeCount--;
-	}
 
 	AdvanceQueue();
 
@@ -1516,6 +1562,8 @@ void CQueueView::CheckQueueState()
 		}
 		*/
 
+		TryRefreshListings();
+
 		m_pMainFrame->GetState()->NotifyHandlers(STATECHANGE_QUEUEPROCESSING);
 
 		if (!m_quit)
@@ -1545,6 +1593,10 @@ void CQueueView::UpdateStatusLinePositions()
 	m_lastTopItem = GetTopItem();
 	int bottomItem = m_lastTopItem + GetCountPerPage();
 
+#ifndef __WXMSW__
+	int header_offset = GetMainWindow()->GetPosition().y;
+#endif
+
 	for (std::list<CStatusLineCtrl*>::iterator iter = m_statusLineList.begin(); iter != m_statusLineList.end(); iter++)
 	{
 		CStatusLineCtrl *pCtrl = *iter;
@@ -1561,6 +1613,9 @@ void CQueueView::UpdateStatusLinePositions()
 			pCtrl->Show(false);
 			continue;
 		}
+#ifndef __WXMSW__
+		rect.y -= header_offset;
+#endif
 		m_allowBackgroundErase = bottomItem + 1 >= m_itemCount;
 		pCtrl->SetSize(rect);
 		m_allowBackgroundErase = false;
@@ -1606,7 +1661,7 @@ bool CQueueView::QueueFolder(bool queueOnly, bool download, const wxString& loca
 	m_queuedFolders[download ? 0 : 1].push_back(folderItem);
 	ProcessFolderItems();
 
-	Refresh(false);
+	RefreshListOnly(false);
 
 	return true;
 }
@@ -1649,7 +1704,7 @@ void CQueueView::ProcessUploadFolderItems()
 	m_pFolderProcessingThread->Create();
 	m_pFolderProcessingThread->Run();
 
-	Refresh(false);
+	RefreshListOnly(false);
 }
 
 void CQueueView::OnFolderThreadComplete(wxCommandEvent& event)
@@ -1659,6 +1714,13 @@ void CQueueView::OnFolderThreadComplete(wxCommandEvent& event)
 
 	wxASSERT(!m_queuedFolders[1].empty());
 	CFolderScanItem* pItem = m_queuedFolders[1].front();
+	if (pItem->m_dir_is_empty)
+	{
+		CServerItem* pServerItem = (CServerItem*)pItem->GetTopLevelItem();
+		CFileItem* fileItem = new CFolderItem(pServerItem, pItem->Queued(), pItem->m_current_remote_path, _T(""));
+		InsertItem(pServerItem, fileItem);
+		QueueFile_Finish(!pItem->Queued());
+	}
 	m_queuedFolders[1].pop_front();
 
 	RemoveItem(pItem, true);
@@ -1670,45 +1732,72 @@ void CQueueView::OnFolderThreadComplete(wxCommandEvent& event)
 	ProcessUploadFolderItems();
 }
 
-bool CQueueView::QueueFiles(const std::list<t_newEntry*> &entryList, bool queueOnly, bool download, CServerItem* pServerItem, const enum CFileExistsNotification::OverwriteAction defaultFileExistsAction)
+int CQueueView::QueueFiles(const std::list<CFolderProcessingEntry*> &entryList, bool queueOnly, bool download, CServerItem* pServerItem, const enum CFileExistsNotification::OverwriteAction defaultFileExistsAction)
 {
 	wxASSERT(pServerItem);
+	
+	CFolderScanItem* pFolderScanItem = m_pFolderProcessingThread->GetFolderScanItem();
 
-	for (std::list<t_newEntry*>::const_iterator iter = entryList.begin(); iter != entryList.end(); iter++)
+	int added = 0;
+
+	CFilterManager filters;
+	for (std::list<CFolderProcessingEntry*>::const_iterator iter = entryList.begin(); iter != entryList.end(); iter++)
 	{
-		const t_newEntry* entry = *iter;
-
-		CFileItem* fileItem;
-		if (entry->localFile != _T(""))
+		const CFolderProcessingEntry* entry = *iter;
+		if (entry->m_type == CFolderProcessingEntry::dir)
 		{
-			fileItem = new CFileItem(pServerItem, queueOnly, download, entry->localFile, entry->remoteFile, entry->remotePath, entry->size);
-			fileItem->m_transferSettings.binary = ShouldUseBinaryMode(download ? entry->remoteFile : wxFileName(entry->localFile).GetFullName(), entry->remotePath.GetType());
-			fileItem->m_defaultFileExistsAction = defaultFileExistsAction;
+			if (m_pFolderProcessingThread->GetFolderScanItem()->m_dir_is_empty)
+			{
+				CFileItem* fileItem = new CFolderItem(pServerItem, queueOnly, pFolderScanItem->m_current_remote_path, _T(""));
+				InsertItem(pServerItem, fileItem);
+				added++;
+			}
+
+			const CFolderProcessingThread::t_dirPair* entry = (const CFolderProcessingThread::t_dirPair*)*iter;
+			pFolderScanItem->m_current_local_path = entry->localPath;
+			pFolderScanItem->m_current_remote_path = entry->remotePath;
+			pFolderScanItem->m_dir_is_empty = true;
+			delete entry;
 		}
 		else
-			fileItem = new CFolderItem(pServerItem, queueOnly, entry->remotePath, _T(""));
+		{
+			const t_newEntry* entry = (const t_newEntry*)*iter;
 
-		delete entry;
+			if (filters.FilenameFiltered(entry->name, pFolderScanItem->m_current_local_path, entry->dir, entry->size, true, entry->attributes))
+			{
+				delete entry;
+				continue;
+			}
 
-		InsertItem(pServerItem, fileItem);
+			pFolderScanItem->m_dir_is_empty = false;
+
+			if (entry->dir)
+			{
+				m_pFolderProcessingThread->ProcessDirectory(pFolderScanItem->m_current_local_path, pFolderScanItem->m_current_remote_path, entry->name);
+				delete entry;
+				continue;
+			}
+
+			CFileItem* fileItem = new CFileItem(pServerItem, queueOnly, download, pFolderScanItem->m_current_local_path + CLocalFileSystem::path_separator + entry->name, entry->name, pFolderScanItem->m_current_remote_path, entry->size);
+
+			if (download)
+				fileItem->m_transferSettings.binary = !CAutoAsciiFiles::TransferRemoteAsAscii(entry->name, pFolderScanItem->m_current_remote_path.GetType());
+			else
+				fileItem->m_transferSettings.binary = !CAutoAsciiFiles::TransferLocalAsAscii(entry->name, pFolderScanItem->m_current_remote_path.GetType());
+
+			fileItem->m_defaultFileExistsAction = defaultFileExistsAction;
+
+			delete entry;
+
+			InsertItem(pServerItem, fileItem);
+
+			added++;
+		}
 	}
 
-	CommitChanges();
+	QueueFile_Finish(!queueOnly);
 
-	if (!m_activeMode && !queueOnly)
-	{
-		m_activeMode = 1;
-		m_pMainFrame->GetState()->NotifyHandlers(STATECHANGE_QUEUEPROCESSING);
-	}
-
-	m_waitStatusLineUpdate = true;
-	AdvanceQueue();
-	m_waitStatusLineUpdate = false;
-	UpdateStatusLinePositions();
-
-	Refresh(false);
-
-	return true;
+	return added;
 }
 
 void CQueueView::SaveQueue()
@@ -1766,9 +1855,10 @@ void CQueueView::LoadQueue()
 
 	pDocument->RemoveChild(pQueue);
 
-	if (!pDocument->GetDocument()->SaveFile(file.GetFullPath().mb_str()))
+	wxString error;
+	if (!SaveXmlFile(file.GetFullPath(), pDocument, &error))
 	{
-		wxString msg = wxString::Format(_("Could not write \"%s\", the queue could not be saved."), file.GetFullPath().c_str());
+		wxString msg = wxString::Format(_("Could not write \"%s\", the queue could not be saved.\n%s"), file.GetFullPath().c_str(), error.c_str());
 		wxMessageBox(msg, _("Error writing xml file"), wxICON_ERROR);
 	}
 
@@ -1865,7 +1955,7 @@ void CQueueView::ImportQueue(TiXmlElement* pElement, bool updateSelections)
 		CommitChanges();
 	}
 	else
-		Refresh();
+		RefreshListOnly();
 }
 
 void CQueueView::SettingsChanged()
@@ -1891,72 +1981,8 @@ void CQueueView::SettingsChanged()
 		}
 	}
 
-	m_asciiFiles.clear();
-	wxString extensions = COptions::Get()->GetOption(OPTION_ASCIIFILES);
-	wxString ext;
-	int pos = extensions.Find(_T("|"));
-	while (pos != -1)
-	{
-		if (!pos)
-		{
-			if (ext != _T(""))
-			{
-				ext.Replace(_T("\\\\"), _T("\\"));
-				m_asciiFiles.push_back(ext);
-				ext = _T("");
-			}
-		}
-		else if (extensions.c_str()[pos - 1] != '\\')
-		{
-			ext += extensions.Left(pos);
-			ext.Replace(_T("\\\\"), _T("\\"));
-			m_asciiFiles.push_back(ext);
-			ext = _T("");
-		}
-		else
-		{
-			ext += extensions.Left(pos - 1) + _T("|");
-		}
-		extensions = extensions.Mid(pos + 1);
-		pos = extensions.Find(_T("|"));
-	}
-	ext += extensions;
-	ext.Replace(_T("\\\\"), _T("\\"));
-	m_asciiFiles.push_back(ext);
-
-	AdvanceQueue();
-}
-
-// Defined in RemoteListView.cpp
-wxString StripVMSRevision(const wxString& name);
-
-bool CQueueView::ShouldUseBinaryMode(wxString filename, enum ServerType type)
-{
-	int mode = COptions::Get()->GetOptionVal(OPTION_ASCIIBINARY);
-	if (mode == 1)
-		return false;
-	else if (mode == 2)
-		return true;
-
-	if (type == VMS)
-		filename = StripVMSRevision(filename);
-
-	if (filename[0] == '.')
-		return COptions::Get()->GetOptionVal(OPTION_ASCIIDOTFILE) == 0;
-
-	int pos = filename.Find('.', true);
-	if (pos == -1)
-		return COptions::Get()->GetOptionVal(OPTION_ASCIINOEXT) == 0;
-	filename = filename.Mid(pos + 1);
-
-	if (filename == _T(""))
-		return true;
-
-	for (std::list<wxString>::const_iterator iter = m_asciiFiles.begin(); iter != m_asciiFiles.end(); iter++)
-		if (!filename.CmpNoCase(*iter))
-			return false;
-
-	return true;
+	if (m_activeMode)
+		AdvanceQueue();
 }
 
 void CQueueView::OnPostScroll()
@@ -2114,7 +2140,7 @@ void CQueueView::RemoveAll()
 	CalculateQueueSize();
 
 	CheckQueueState();
-	Refresh();
+	RefreshListOnly();
 }
 
 void CQueueView::OnRemoveSelected(wxCommandEvent& event)
@@ -2192,7 +2218,7 @@ void CQueueView::OnRemoveSelected(wxCommandEvent& event)
 	m_waitStatusLineUpdate = false;
 	UpdateStatusLinePositions();
 
-	Refresh();
+	RefreshListOnly();
 }
 
 bool CQueueView::StopItem(CFileItem* item)
@@ -2275,11 +2301,12 @@ void CQueueView::OnFolderThreadFiles(wxCommandEvent& event)
 	wxASSERT(!m_queuedFolders[1].empty());
 	CFolderScanItem* pItem = m_queuedFolders[1].front();
 
-	std::list<t_newEntry*> entryList;
+	std::list<CFolderProcessingEntry*> entryList;
 	m_pFolderProcessingThread->GetFiles(entryList);
-	QueueFiles(entryList, pItem->Queued(), false, (CServerItem*)pItem->GetTopLevelItem(), pItem->m_defaultFileExistsAction);
+	int added = QueueFiles(entryList, pItem->Queued(), false, (CServerItem*)pItem->GetTopLevelItem(), pItem->m_defaultFileExistsAction);
+	m_pFolderProcessingThread->CheckFinished();
 
-	pItem->m_count += entryList.size();
+	pItem->m_count += added;
 	pItem->m_statusMessage = wxString::Format(_("%d files added to queue"), pItem->GetCount());
 	RefreshItem(pItem);
 }
@@ -2305,11 +2332,14 @@ void CQueueView::OnSetDefaultFileExistsAction(wxCommandEvent &event)
 	if (!dlg.Load(this, true))
 		return;
 
-	enum CFileExistsNotification::OverwriteAction downloadAction, uploadAction;
-	if (!dlg.Run(&downloadAction, &uploadAction))
-		return;
+	// Get current default action for the item
+	enum CFileExistsNotification::OverwriteAction downloadAction = CFileExistsNotification::unknown;
+	enum CFileExistsNotification::OverwriteAction uploadAction = CFileExistsNotification::unknown;
+	bool has_upload = false;
+	bool has_download = false;
+	bool download_unknown = false;
+	bool upload_unknown = false;
 
-	std::list<long> selectedItems;
 	long item = -1;
 	while (true)
 	{
@@ -2324,22 +2354,95 @@ void CQueueView::OnSetDefaultFileExistsAction(wxCommandEvent &event)
 		switch (pItem->GetType())
 		{
 		case QueueItemType_FolderScan:
+			if (uploadAction == CFileExistsNotification::unknown)
+				uploadAction = ((CFolderScanItem*)pItem)->m_defaultFileExistsAction;
+			else if (((CFolderScanItem*)pItem)->m_defaultFileExistsAction != uploadAction)
+				upload_unknown = true;
+			has_upload = true;
+			break;
+		case QueueItemType_File:
+			{
+				CFileItem *pFileItem = (CFileItem*)pItem;
+				if (pFileItem->Download())
+				{
+					if (downloadAction == CFileExistsNotification::unknown)
+						downloadAction = pFileItem->m_defaultFileExistsAction;
+					else if (pFileItem->m_defaultFileExistsAction != downloadAction)
+						download_unknown = true;
+					has_download = true;
+				}
+				else
+				{
+					if (uploadAction == CFileExistsNotification::unknown)
+						uploadAction = pFileItem->m_defaultFileExistsAction;
+					else if (pFileItem->m_defaultFileExistsAction != uploadAction)
+						upload_unknown = true;
+					has_upload = true;
+				}
+			}
+			break;
+		case QueueItemType_Server:
+			{
+				download_unknown = true;
+				upload_unknown = true;
+				has_download = true;
+				has_upload = true;
+			}
+			break;
+		default:
+			break;
+		}
+	}
+	if (download_unknown)
+		downloadAction = CFileExistsNotification::unknown;
+	if (upload_unknown)
+		uploadAction = CFileExistsNotification::unknown;
+
+	if (!dlg.Run(has_download ? &downloadAction : 0, has_upload ? &uploadAction : 0))
+		return;
+
+	item = -1;
+	while (true)
+	{
+		item = GetNextItem(item, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
+		if (item == -1)
+			break;
+
+		CQueueItem* pItem = GetQueueItem(item);
+		if (!pItem)
+			continue;
+
+		switch (pItem->GetType())
+		{
+		case QueueItemType_FolderScan:
+			if (!has_upload)
+				break;
 			((CFolderScanItem*)pItem)->m_defaultFileExistsAction = uploadAction;
 			break;
 		case QueueItemType_File:
 			{
 				CFileItem *pFileItem = (CFileItem*)pItem;
 				if (pFileItem->Download())
+				{
+					if (!has_download)
+						break;
 					pFileItem->m_defaultFileExistsAction = downloadAction;
+				}
 				else
+				{
+					if (!has_upload)
+						break;
 					pFileItem->m_defaultFileExistsAction = uploadAction;
+				}
 			}
 			break;
 		case QueueItemType_Server:
 			{
 				CServerItem *pServerItem = (CServerItem*)pItem;
-				pServerItem->SetDefaultFileExistsAction(downloadAction, download);
-				pServerItem->SetDefaultFileExistsAction(uploadAction, upload);
+				if (has_download)
+					pServerItem->SetDefaultFileExistsAction(downloadAction, download);
+				if (has_upload)
+					pServerItem->SetDefaultFileExistsAction(uploadAction, upload);
 			}
 			break;
 		default:
@@ -2373,8 +2476,6 @@ t_EngineData* CQueueView::GetIdleEngine(const CServer* pServer /*=0*/)
 
 void CQueueView::TryRefreshListings()
 {
-	// TODO: This function is currently unused
-
 	if (m_quit)
 		return;
 
@@ -2384,25 +2485,32 @@ void CQueueView::TryRefreshListings()
 		return;
 
 	const CDirectoryListing* const pListing = pState->GetRemoteDir();
-	if (!pListing || !pListing->m_hasUnsureEntries)
+	if (!pListing)
 		return;
 
-	// See if there's an engine with is already listing
+	// See if there's an engine that is already listing
 	for (unsigned int i = 1; i < m_engineData.size(); i++)
 	{
-		if (!m_engineData[i]->active)
+		if (!m_engineData[i]->active || m_engineData[i]->state != t_EngineData::list)
 			continue;
 
-		if (m_engineData[i]->pItem)
-			continue;
-
-		if (m_engineData[i]->pEngine->IsConnected() && m_engineData[i]->lastServer == *pServer)
+		if (m_engineData[i]->lastServer == *pServer)
 		{
 			// This engine is already listing a directory on the current server
 			return;
 		}
 	}
 
+	if (m_last_refresh_server == *pServer && m_last_refresh_path == pListing->path &&
+		m_last_refresh_listing_time == pListing->m_firstListTime)
+	{
+		// Do not try to refresh same directory multiple times
+		return;
+	}
+	m_last_refresh_server = *pServer;
+	m_last_refresh_path = pListing->path;
+	m_last_refresh_listing_time = pListing->m_firstListTime;
+	
 	t_EngineData* pEngineData = GetIdleEngine(pServer);
 	if (!pEngineData)
 		return;
@@ -2410,13 +2518,14 @@ void CQueueView::TryRefreshListings()
 	if (!pEngineData->pEngine->IsConnected() || pEngineData->lastServer != *pServer)
 		return;
 
-	CListCommand command(pListing->path);
+	CListCommand command(pListing->path, _T(""), LIST_FLAG_AVOID);
 	int res = pEngineData->pEngine->Command(command);
 	if (res != FZ_REPLY_WOULDBLOCK)
 		return;
 
 	pEngineData->active = true;
 	pEngineData->state = t_EngineData::list;
+	m_activeCount++;
 }
 
 void CQueueView::OnAskPassword(wxCommandEvent& event)
@@ -2491,7 +2600,7 @@ void CQueueView::UpdateItemSize(CFileItem* pItem, wxLongLong size)
 	DisplayQueueSize();
 }
 
-void CQueueView::AdvanceQueue()
+void CQueueView::AdvanceQueue(bool refresh /*=true*/)
 {
 	static bool insideAdvanceQueue = false;
 	if (insideAdvanceQueue)
@@ -2522,11 +2631,12 @@ void CQueueView::AdvanceQueue()
 				continue;
 
 			m_engineData[i]->m_idleDisconnectTimer = new wxTimer(this);
-			m_engineData[i]->m_idleDisconnectTimer->Start(30000, true);
+			m_engineData[i]->m_idleDisconnectTimer->Start(60000, true);
 		}
 	}
 
-	Refresh(false);
+	if (refresh)
+		RefreshListOnly(false);
 
 	insideAdvanceQueue = false;
 
@@ -2656,7 +2766,7 @@ void CQueueView::OnSetPriority(wxCommandEvent& event)
 		pItem->SetPriority(priority);
 	}
 
-	Refresh();
+	RefreshListOnly();
 }
 
 void CQueueView::OnExclusiveEngineRequestGranted(wxCommandEvent& event)
