@@ -66,6 +66,10 @@ public:
 		if (!GetData())
 			return wxDragError;
 
+		CDragDropManager* pDragDropManager = CDragDropManager::Get();
+		if (pDragDropManager)
+			pDragDropManager->pDropTarget = m_pRemoteListView;
+
 		if (m_pDataObject->GetReceivedFormat() == m_pFileDataObject->GetFormat())
 		{
 			wxString subdir;
@@ -157,6 +161,8 @@ public:
 				new CRenameCommand(m_pRemoteDataObject->GetServerPath(), info.name, target, info.name)
 				);
 		}
+
+		// Refresh remote listing
 		m_pRemoteListView->m_pState->m_pCommandQueue->ProcessCommand(
 			new CListCommand()
 			);
@@ -272,6 +278,7 @@ public:
 		: wxWindow(parent, wxID_ANY, wxPoint(0, 60), wxDefaultSize),
 		m_text(_T("<") + text + _T(">"))
 	{
+		SetForegroundColour(parent->GetForegroundColour());
 		SetBackgroundColour(parent->GetBackgroundColour());
 		GetTextExtent(m_text, &m_textSize.x, &m_textSize.y);
 	}
@@ -297,7 +304,7 @@ protected:
 		wxPaintDC paintDc(this);
 
 		paintDc.SetFont(GetFont());
-		paintDc.SetTextForeground(wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOWTEXT));
+		paintDc.SetTextForeground(GetForegroundColour());
 
 		paintDc.DrawText(m_text, 0, 0);
 	};
@@ -321,13 +328,13 @@ BEGIN_EVENT_TABLE(CRemoteListView, CFileListCtrl<CGenericFileData>)
 	EVT_MENU(XRCID("ID_DELETE"), CRemoteListView::OnMenuDelete)
 	EVT_MENU(XRCID("ID_RENAME"), CRemoteListView::OnMenuRename)
 	EVT_MENU(XRCID("ID_CHMOD"), CRemoteListView::OnMenuChmod)
-	EVT_CHAR(CRemoteListView::OnChar)
 	EVT_KEY_DOWN(CRemoteListView::OnKeyDown)
 	EVT_LIST_BEGIN_LABEL_EDIT(wxID_ANY, CRemoteListView::OnBeginLabelEdit)
 	EVT_LIST_END_LABEL_EDIT(wxID_ANY, CRemoteListView::OnEndLabelEdit)
 	EVT_SIZE(CRemoteListView::OnSize)
 	EVT_LIST_BEGIN_DRAG(wxID_ANY, CRemoteListView::OnBeginDrag)
 	EVT_MENU(XRCID("ID_EDIT"), CRemoteListView::OnMenuEdit)
+	EVT_MENU(XRCID("ID_ENTER"), CRemoteListView::OnMenuEnter)
 END_EVENT_TABLE()
 
 CRemoteListView::CRemoteListView(wxWindow* pParent, CState *pState, CQueueView* pQueue)
@@ -370,12 +377,16 @@ CRemoteListView::CRemoteListView(wxWindow* pParent, CState *pState, CQueueView*
 	InitDateFormat();
 
 	EnablePrefixSearch(true);
+
+	m_pLinkResolveState = 0;
 }
 
 CRemoteListView::~CRemoteListView()
 {
 	wxString str = wxString::Format(_T("%d %d"), m_sortDirection, m_sortColumn);
 	COptions::Get()->SetOption(OPTION_REMOTEFILELIST_SORTORDER, str);
+
+	delete m_pLinkResolveState;
 }
 
 // See comment to OnGetItemText
@@ -419,10 +430,99 @@ bool CRemoteListView::IsItemValid(unsigned int item) const
 	return true;
 }
 
+void CRemoteListView::UpdateDirectoryListing_Added(const CDirectoryListing *pDirectoryListing)
+{
+	const unsigned int to_add = pDirectoryListing->GetCount() - m_pDirectoryListing->GetCount();
+	m_pDirectoryListing = pDirectoryListing;
+
+	m_indexMapping[0] = pDirectoryListing->GetCount();
+
+	std::list<unsigned int> added;
+
+	CFilterManager filter;
+	const wxString path = m_pDirectoryListing->path.GetPath();
+
+	CGenericFileData last = m_fileData.back();
+	m_fileData.pop_back();
+
+	for (unsigned int i = pDirectoryListing->GetCount() - to_add; i < pDirectoryListing->GetCount(); i++)
+	{
+		const CDirentry& entry = (*pDirectoryListing)[i];
+		CGenericFileData data;
+		data.flags = normal;
+		data.icon = entry.dir ? m_dirIcon : -2;
+		m_fileData.push_back(data);
+
+		if (filter.FilenameFiltered(entry.name, path, entry.dir, entry.size, false, 0))
+			continue;
+
+		if (m_pFilelistStatusBar)
+		{
+			if (entry.dir)
+				m_pFilelistStatusBar->AddDirectory();
+			else
+				m_pFilelistStatusBar->AddFile(entry.size);
+		}
+
+		// Find correct position in index mapping
+		std::vector<unsigned int>::iterator start = m_indexMapping.begin();
+		if (m_hasParent)
+			start++;
+		CFileListCtrl<CGenericFileData>::CSortComparisonObject compare = GetSortComparisonObject();
+		std::vector<unsigned int>::iterator insertPos = std::lower_bound(start, m_indexMapping.end(), i, compare);
+		compare.Destroy();
+
+		const int item = insertPos - m_indexMapping.begin();
+		m_indexMapping.insert(insertPos, i);
+
+		for (std::list<unsigned int>::iterator iter = added.begin(); iter != added.end(); iter++)
+		{
+			unsigned int &pos = *iter;
+			if (pos >= (unsigned int)item)
+				pos++;
+		}
+		added.push_back(item);
+	}
+
+	m_fileData.push_back(last);
+
+	std::list<bool> selected;
+	unsigned int start;
+	added.push_back(m_indexMapping.size());
+	start = added.front();
+
+	SetItemCount(m_indexMapping.size());
+
+	for (unsigned int i = start; i < m_indexMapping.size(); i++)
+	{
+		if (i == added.front())
+		{
+			selected.push_front(false);
+			added.pop_front();
+		}
+		bool is_selected = GetItemState(i, wxLIST_STATE_SELECTED) != 0;
+		selected.push_back(is_selected);
+
+		bool should_selected = selected.front();
+		selected.pop_front();
+		if (is_selected != should_selected)
+			SetSelection(i, should_selected);
+	}
+
+	if (m_pFilelistStatusBar)
+		m_pFilelistStatusBar->SetHidden(m_pDirectoryListing->GetCount() + 1 - m_indexMapping.size());
+
+	wxASSERT(m_indexMapping.size() == pDirectoryListing->GetCount() + 1);
+}
+
 void CRemoteListView::UpdateDirectoryListing_Removed(const CDirectoryListing *pDirectoryListing)
 {
 	const unsigned int removed = m_pDirectoryListing->GetCount() - pDirectoryListing->GetCount();
-	wxASSERT(removed);
+	if (!removed)
+	{
+		m_pDirectoryListing = pDirectoryListing;
+		return;
+	}
 	wxASSERT(!IsComparing());
 
 	std::list<unsigned int> removedItems;
@@ -442,14 +542,6 @@ void CRemoteListView::UpdateDirectoryListing_Removed(const CDirectoryListing *pD
 			continue;
 		}
 
-		if (m_pFilelistStatusBar)
-		{
-			if (oldEntry.dir)
-				m_pFilelistStatusBar->RemoveDirectory();
-			else
-				m_pFilelistStatusBar->RemoveFile(oldEntry.size);
-		}
-
 		removedItems.push_back(j++);
 	}
 	for (; j < m_pDirectoryListing->GetCount(); j++)
@@ -491,12 +583,20 @@ void CRemoteListView::UpdateDirectoryListing_Removed(const CDirectoryListing *pD
 		bool isSelected = GetItemState(i, wxLIST_STATE_SELECTED) != 0;
 		
 		// Update statusbar info
-		if (isSelected && removed && m_pFilelistStatusBar)
+		if (removed && m_pFilelistStatusBar)
 		{
-			if ((*m_pDirectoryListing)[index].dir)
-				m_pFilelistStatusBar->UnselectDirectory();
+			const CDirentry& oldEntry = (*m_pDirectoryListing)[index];
+			if (isSelected)
+			{
+				if (oldEntry.dir)
+					m_pFilelistStatusBar->UnselectDirectory();
+				else
+					m_pFilelistStatusBar->UnselectFile(oldEntry.size);
+			}
+			if (oldEntry.dir)
+				m_pFilelistStatusBar->RemoveDirectory();
 			else
-				m_pFilelistStatusBar->UnselectFile((*m_pDirectoryListing)[index].size);
+				m_pFilelistStatusBar->RemoveFile(oldEntry.size);
 		}
 
 		// Update index
@@ -547,7 +647,7 @@ void CRemoteListView::UpdateDirectoryListing_Removed(const CDirectoryListing *pD
 	if (m_pFilelistStatusBar)
 		m_pFilelistStatusBar->SetHidden(m_pDirectoryListing->GetCount() + 1 - m_indexMapping.size());
 
-	SetItemCount(m_indexMapping.size());
+	SaveSetItemCount(m_indexMapping.size());
 }
 
 bool CRemoteListView::UpdateDirectoryListing(const CDirectoryListing *pDirectoryListing)
@@ -586,11 +686,13 @@ bool CRemoteListView::UpdateDirectoryListing(const CDirectoryListing *pDirectory
 
 	if (unsure & (CDirectoryListing::unsure_dir_added | CDirectoryListing::unsure_file_added))
 	{
-		// Todo: Would be nice to handle this too
-		return false;
+		if (unsure & (CDirectoryListing::unsure_dir_removed | CDirectoryListing::unsure_file_removed))
+			return false; // Cannot handle both at the same time unfortunately
+		UpdateDirectoryListing_Added(pDirectoryListing);
+		return true;
 	}
 
-	wxASSERT(pDirectoryListing->GetCount() < m_pDirectoryListing->GetCount());
+	wxASSERT(pDirectoryListing->GetCount() <= m_pDirectoryListing->GetCount());
 	UpdateDirectoryListing_Removed(pDirectoryListing);
 	return true;
 }
@@ -622,7 +724,7 @@ void CRemoteListView::SetDirectoryListing(const CDirectoryListing *pDirectoryLis
 			wxASSERT(m_pDirectoryListing->GetCount() + 1 >= (unsigned int)GetItemCount());
 			wxASSERT(m_indexMapping[0] == m_pDirectoryListing->GetCount());
 
-			Refresh();
+			RefreshListOnly();
 
 			return;
 		}
@@ -753,7 +855,7 @@ void CRemoteListView::SetDirectoryListing(const CDirectoryListing *pDirectoryLis
 	else
 	{
 		ReselectItems(selectedNames, prevFocused);
-		Refresh(eraseBackground);
+		RefreshListOnly(eraseBackground);
 	}
 }
 
@@ -953,9 +1055,9 @@ public:
 
 		CGenericFileData &type1 = m_fileData[a];
 		CGenericFileData &type2 = m_fileData[b];
-		if (type1.fileType == _T(""))
+		if (type1.fileType.IsEmpty())
 			type1.fileType = m_pRemoteListView->GetType(data1.name, data1.dir);
-		if (type2.fileType == _T(""))
+		if (type2.fileType.IsEmpty())
 			type2.fileType = m_pRemoteListView->GetType(data2.name, data2.dir);
 
 		CMP(CmpStringNoCase, type1.fileType, type2.fileType);
@@ -1093,23 +1195,72 @@ void CRemoteListView::OnItemActivated(wxListEvent &event)
 		const CDirentry& entry = (*m_pDirectoryListing)[index];
 		const wxString& name = entry.name;
 
+		const CServer* pServer = m_pState->GetServer();
+		if (!pServer)
+		{
+			wxBell();
+			return;
+		}
+
 		if (entry.dir)
 		{
-			if (IsComparing())
-				ExitComparisonMode();
-			m_pState->m_pCommandQueue->ProcessCommand(new CListCommand(m_pDirectoryListing->path, name));
+			const int action = COptions::Get()->GetOptionVal(OPTION_DOUBLECLICK_ACTION_DIRECTORY);
+			if (action == 3)
+			{
+				// No action
+				wxBell();
+				return;
+			}
+
+			if (!action)
+			{
+				if (IsComparing())
+					ExitComparisonMode();
+				if (entry.link)
+				{
+					delete m_pLinkResolveState;
+					m_pLinkResolveState = new t_linkResolveState;
+					m_pLinkResolveState->remote_path = m_pDirectoryListing->path;
+					m_pLinkResolveState->link = name;
+					m_pLinkResolveState->local_path = m_pState->GetLocalDir().GetPath();
+					m_pLinkResolveState->server = *pServer;
+				}
+				m_pState->ChangeRemoteDir(m_pDirectoryListing->path, name, entry.link ? LIST_FLAG_LINK : 0);
+			}
+			else
+			{
+				wxCommandEvent evt(0, action == 1 ? XRCID("ID_DOWNLOAD") : XRCID("ID_ADDTOQUEUE"));
+				OnMenuDownload(evt);
+			}
 		}
 		else
 		{
-			const CServer* pServer = m_pState->GetServer();
-			if (!pServer)
+			const int action = COptions::Get()->GetOptionVal(OPTION_DOUBLECLICK_ACTION_FILE);
+			if (action == 3)
+			{
+				// No action
+				wxBell();
+				return;
+			}
+
+			if (action == 2)
+			{
+				// View / Edit action
+				wxCommandEvent evt;
+				OnMenuEdit(evt);
+				return;
+			}
+
+			const bool queue_only = action == 1;
+
+			const CLocalPath local_path = m_pState->GetLocalDir();
+			if (!local_path.IsWriteable())
 			{
 				wxBell();
 				return;
 			}
 
-			wxFileName fn = wxFileName(m_pState->GetLocalDir(), name);
-			m_pQueue->QueueFile(false, true, fn.GetFullPath(), name, m_pDirectoryListing->path, *pServer, entry.size);
+			m_pQueue->QueueFile(queue_only, true, local_path.GetPath() + name, name, m_pDirectoryListing->path, *pServer, entry.size);
 			m_pQueue->QueueFile_Finish(true);
 		}
 	}
@@ -1118,7 +1269,75 @@ void CRemoteListView::OnItemActivated(wxListEvent &event)
 		if (IsComparing())
 			ExitComparisonMode();
 
-		m_pState->m_pCommandQueue->ProcessCommand(new CListCommand(m_pDirectoryListing->path, _T("..")));
+		m_pState->ChangeRemoteDir(m_pDirectoryListing->path, _T(".."));
+	}
+}
+
+void CRemoteListView::OnMenuEnter(wxCommandEvent &event)
+{
+	if (!m_pState->IsRemoteIdle())
+	{
+		wxBell();
+		return;
+	}
+
+	int item = GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
+
+	if (GetNextItem(item, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED) != -1)
+	{
+		wxBell();
+		return;
+	}
+
+	if (item)
+	{
+		int index = GetItemIndex(item);
+		if (index == -1)
+		{
+			wxBell();
+			return;
+		}
+		if (m_fileData[index].flags == fill)
+		{
+			wxBell();
+			return;
+		}
+
+		const CDirentry& entry = (*m_pDirectoryListing)[index];
+		const wxString& name = entry.name;
+
+		const CServer* pServer = m_pState->GetServer();
+		if (!pServer)
+		{
+			wxBell();
+			return;
+		}
+
+		if (!entry.dir)
+		{
+			wxBell();
+			return;
+		}
+
+		if (IsComparing())
+			ExitComparisonMode();
+		if (entry.link)
+		{
+			delete m_pLinkResolveState;
+			m_pLinkResolveState = new t_linkResolveState;
+			m_pLinkResolveState->remote_path = m_pDirectoryListing->path;
+			m_pLinkResolveState->link = name;
+			m_pLinkResolveState->local_path = m_pState->GetLocalDir().GetPath();
+			m_pLinkResolveState->server = *pServer;
+		}
+		m_pState->ChangeRemoteDir(m_pDirectoryListing->path, name, entry.link ? LIST_FLAG_LINK : 0);
+	}
+	else
+	{
+		if (IsComparing())
+			ExitComparisonMode();
+
+		m_pState->ChangeRemoteDir(m_pDirectoryListing->path, _T(".."));
 	}
 }
 
@@ -1136,6 +1355,7 @@ void CRemoteListView::OnContextMenu(wxContextMenuEvent& event)
 
 	if (!m_pState->IsRemoteConnected() || !m_pState->IsRemoteIdle())
 	{
+		pMenu->Delete(XRCID("ID_ENTER"));
 		pMenu->Enable(XRCID("ID_DOWNLOAD"), false);
 		pMenu->Enable(XRCID("ID_ADDTOQUEUE"), false);
 		pMenu->Enable(XRCID("ID_MKDIR"), false);
@@ -1144,8 +1364,7 @@ void CRemoteListView::OnContextMenu(wxContextMenuEvent& event)
 		pMenu->Enable(XRCID("ID_CHMOD"), false);
 		pMenu->Enable(XRCID("ID_EDIT"), false);
 	}
-	else if ((GetItemCount() && GetItemState(0, wxLIST_STATE_SELECTED)) ||
-		GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED) == -1)
+	else if ((GetItemCount() && GetItemState(0, wxLIST_STATE_SELECTED)))
 	{
 		pMenu->Enable(XRCID("ID_DOWNLOAD"), false);
 		pMenu->Enable(XRCID("ID_ADDTOQUEUE"), false);
@@ -1154,6 +1373,16 @@ void CRemoteListView::OnContextMenu(wxContextMenuEvent& event)
 		pMenu->Enable(XRCID("ID_CHMOD"), false);
 		pMenu->Enable(XRCID("ID_EDIT"), false);
 	}
+	else if (GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED) == -1)
+	{
+		pMenu->Delete(XRCID("ID_ENTER"));
+		pMenu->Enable(XRCID("ID_DOWNLOAD"), false);
+		pMenu->Enable(XRCID("ID_ADDTOQUEUE"), false);
+		pMenu->Enable(XRCID("ID_DELETE"), false);
+		pMenu->Enable(XRCID("ID_RENAME"), false);
+		pMenu->Enable(XRCID("ID_CHMOD"), false);
+		pMenu->Enable(XRCID("ID_EDIT"), false);
+	}
 	else
 	{
 		int count = 0;
@@ -1176,6 +1405,7 @@ void CRemoteListView::OnContextMenu(wxContextMenuEvent& event)
 		}
 		if (!count || fillCount == count)
 		{
+			pMenu->Delete(XRCID("ID_ENTER"));
 			pMenu->Enable(XRCID("ID_DOWNLOAD"), false);
 			pMenu->Enable(XRCID("ID_ADDTOQUEUE"), false);
 			pMenu->Enable(XRCID("ID_DELETE"), false);
@@ -1187,11 +1417,15 @@ void CRemoteListView::OnContextMenu(wxContextMenuEvent& event)
 		{
 			if (selectedDir)
 				pMenu->Enable(XRCID("ID_EDIT"), false);
+			else
+				pMenu->Delete(XRCID("ID_ENTER"));
 			if (count > 1)
+			{
+				pMenu->Delete(XRCID("ID_ENTER"));
 				pMenu->Enable(XRCID("ID_RENAME"), false);
+			}
 
-			const wxString& localDir = m_pState->GetLocalDir();
-			if (!m_pState->LocalDirIsWriteable(localDir))
+			if (!m_pState->GetLocalDir().IsWriteable())
 			{
 				pMenu->Enable(XRCID("ID_DOWNLOAD"), false);
 				pMenu->Enable(XRCID("ID_ADDTOQUEUE"), false);
@@ -1208,8 +1442,8 @@ void CRemoteListView::OnMenuDownload(wxCommandEvent& event)
 	// Make sure selection is valid
 	bool idle = m_pState->IsRemoteIdle();
 
-	const wxString& localDir = m_pState->GetLocalDir();
-	if (!m_pState->LocalDirIsWriteable(localDir))
+	const CLocalPath localDir = m_pState->GetLocalDir();
+	if (!localDir.IsWriteable())
 	{
 		wxBell();
 		return;
@@ -1243,14 +1477,14 @@ void CRemoteListView::OnMenuDownload(wxCommandEvent& event)
 	TransferSelectedFiles(localDir, event.GetId() == XRCID("ID_ADDTOQUEUE"));
 }
 
-void CRemoteListView::TransferSelectedFiles(const wxString& localDir, bool queueOnly)
+void CRemoteListView::TransferSelectedFiles(const CLocalPath& local_parent, bool queueOnly)
 {
 	bool idle = m_pState->IsRemoteIdle();
 
 	CRecursiveOperation* pRecursiveOperation = m_pState->GetRecursiveOperationHandler();
 	wxASSERT(pRecursiveOperation);
 
-	wxASSERT(m_pState->LocalDirIsWriteable(localDir));
+	wxASSERT(local_parent.IsWriteable());
 
 	const CServer* pServer = m_pState->GetServer();
 	if (!pServer)
@@ -1281,20 +1515,18 @@ void CRemoteListView::TransferSelectedFiles(const wxString& localDir, bool queue
 		{
 			if (!idle)
 				continue;
-			wxFileName fn = wxFileName(localDir, _T(""));
-			fn.AppendDir(name);
+			CLocalPath local_path(local_parent);
+			local_path.AddSegment(name);
 			CServerPath remotePath = m_pDirectoryListing->path;
 			if (remotePath.AddSegment(name))
 			{
-				//m_pQueue->QueueFolder(event.GetId() == XRCID("ID_ADDTOQUEUE"), true, fn.GetFullPath(), remotePath, *pServer);
-				pRecursiveOperation->AddDirectoryToVisit(m_pDirectoryListing->path, name, fn.GetFullPath());
+				pRecursiveOperation->AddDirectoryToVisit(m_pDirectoryListing->path, name, local_path.GetPath(), entry.link);
 				startRecursive = true;
 			}
 		}
 		else
 		{
-			wxFileName fn = wxFileName(localDir, name);
-			m_pQueue->QueueFile(queueOnly, true, fn.GetFullPath(), name, m_pDirectoryListing->path, *pServer, entry.size);
+			m_pQueue->QueueFile(queueOnly, true, local_parent.GetPath() + name, name, m_pDirectoryListing->path, *pServer, entry.size);
 			added = true;
 		}
 	}
@@ -1324,15 +1556,15 @@ void CRemoteListView::OnMenuMkdir(wxCommandEvent& event)
 	CServerPath path = m_pDirectoryListing->path;
 
 	// Append a long segment which does (most likely) not exist in the path and
-	// replace it with "New folder" later. This way we get the exact position of
-	// "New folder" and can preselect it in the dialog.
+	// replace it with "New directory" later. This way we get the exact position of
+	// "New directory" and can preselect it in the dialog.
 	wxString tmpName = _T("25CF809E56B343b5A12D1F0466E3B37A49A9087FDCF8412AA9AF8D1E849D01CF");
 	if (path.AddSegment(tmpName))
 	{
 		wxString pathName = path.GetPath();
 		int pos = pathName.Find(tmpName);
 		wxASSERT(pos != -1);
-		wxString newName = _("New folder");
+		wxString newName = _("New directory");
 		pathName.Replace(tmpName, newName);
 		dlg.SetValue(pathName);
 		dlg.SelectText(pos, pos + newName.Length());
@@ -1462,7 +1694,7 @@ void CRemoteListView::OnMenuRename(wxCommandEvent& event)
 	EditLabel(item);
 }
 
-void CRemoteListView::OnChar(wxKeyEvent& event)
+void CRemoteListView::OnKeyDown(wxKeyEvent& event)
 {
 	int code = event.GetKeyCode();
 	if (code == WXK_DELETE)
@@ -1499,33 +1731,12 @@ void CRemoteListView::OnChar(wxKeyEvent& event)
 		if (IsComparing())
 			ExitComparisonMode();
 
-		m_pState->m_pCommandQueue->ProcessCommand(new CListCommand(m_pDirectoryListing->path, _T("..")));
+		m_pState->ChangeRemoteDir(m_pDirectoryListing->path, _T(".."));
 	}
 	else
 		event.Skip();
 }
 
-void CRemoteListView::OnKeyDown(wxKeyEvent& event)
-{
-	const int code = event.GetKeyCode();
-	const int mods = event.GetModifiers();
-	if (code == 'A' && (mods == wxMOD_CMD || mods == (wxMOD_CONTROL | wxMOD_META)))
-	{
-		for (unsigned int i = 1; i < m_indexMapping.size(); i++)
-		{
-			int index = GetItemIndex(i);
-			if (index != -1 && m_fileData[index].flags != fill)
-				SetSelection(i, true);
-			else
-				SetSelection(i, false);
-		}
-		if (m_pFilelistStatusBar)
-			m_pFilelistStatusBar->SelectAll();
-	}
-	else
-		OnChar(event);
-}
-
 void CRemoteListView::OnBeginLabelEdit(wxListEvent& event)
 {
 	if (!m_pState->IsRemoteIdle())
@@ -1790,12 +2001,12 @@ void CRemoteListView::OnMenuChmod(wxCommandEvent& event)
 		// Refresh listing. This gets done implicitely by the recursive operation, so
 		// only it if not recursing.
 		if (pRecursiveOperation->GetOperationMode() != CRecursiveOperation::recursive_chmod)
-			m_pState->m_pCommandQueue->ProcessCommand(new CListCommand(m_pDirectoryListing->path));
+			m_pState->ChangeRemoteDir(m_pDirectoryListing->path);
 	}
 	else
 	{
 		pChmodDlg->Destroy();
-		m_pState->m_pCommandQueue->ProcessCommand(new CListCommand(m_pDirectoryListing->path));
+		m_pState->ChangeRemoteDir(m_pDirectoryListing->path);
 	}
 
 }
@@ -1865,7 +2076,7 @@ void CRemoteListView::ApplyCurrentFilter()
 
 	ReselectItems(selectedNames, focused);
 	if (!IsComparing())
-		Refresh();
+		RefreshListOnly();
 }
 
 std::list<wxString> CRemoteListView::RememberSelectedItems(wxString& focused)
@@ -2175,7 +2386,7 @@ void CRemoteListView::OnBeginDrag(wxListEvent& event)
 			continue;
 		const CDirentry& entry = (*m_pDirectoryListing)[index];
 
-		pRemoteDataObject->AddFile(entry.name, entry.dir, entry.size);
+		pRemoteDataObject->AddFile(entry.name, entry.dir, entry.size, entry.link);
 	}
 
 	pRemoteDataObject->Finalize();
@@ -2264,8 +2475,8 @@ void CRemoteListView::OnBeginDrag(wxListEvent& event)
 				}
 			}
 
-			wxString target = ext->GetTarget();
-			if (target == _T(""))
+			CLocalPath target(ext->GetTarget());
+			if (target.empty())
 			{
 				delete ext;
 				ext = 0;
@@ -2285,11 +2496,13 @@ void CRemoteListView::OnBeginDrag(wxListEvent& event)
 #endif
 }
 
-bool CRemoteListView::DownloadDroppedFiles(const CRemoteDataObject* pRemoteDataObject, wxString path, bool queueOnly)
+bool CRemoteListView::DownloadDroppedFiles(const CRemoteDataObject* pRemoteDataObject, const CLocalPath& path, bool queueOnly)
 {
 	if (!m_pState->IsRemoteIdle())
 		return false;
 
+	wxASSERT(!path.empty());
+
 	CRecursiveOperation* pRecursiveOperation = m_pState->GetRecursiveOperationHandler();
 	wxASSERT(pRecursiveOperation);
 
@@ -2299,7 +2512,7 @@ bool CRemoteListView::DownloadDroppedFiles(const CRemoteDataObject* pRemoteDataO
 		if (!iter->dir)
 			continue;
 
-		pRecursiveOperation->AddDirectoryToVisit(pRemoteDataObject->GetServerPath(), iter->name, path + iter->name);
+		pRecursiveOperation->AddDirectoryToVisit(pRemoteDataObject->GetServerPath(), iter->name, path.GetPath() + iter->name, iter->link);
 	}
 
 	if (IsComparing())
@@ -2379,10 +2592,28 @@ void CRemoteListView::OnMenuEdit(wxCommandEvent& event)
 		return;
 	}
 
+	const CServerPath path = m_pDirectoryListing->path;
+	const CServer server = *m_pState->GetServer();
+
 	bool dangerous = false;
-	if (!pEditHandler->CanOpen(CEditHandler::remote, entry.name, dangerous))
+	bool program_exists = false;
+	wxString cmd = pEditHandler->CanOpen(CEditHandler::remote, entry.name, dangerous, program_exists);
+	if (cmd.empty())
 	{
-		wxMessageBox(_("Selected file cannot be opened.\nNo default editor has been set or filetype association is missing or incorrect."), _("Cannot edit file"), wxICON_STOP);
+		CNewAssociationDialog dlg(this);
+		if (!dlg.Show(entry.name))
+			return;
+		cmd = pEditHandler->CanOpen(CEditHandler::remote, entry.name, dangerous, program_exists);
+		if (cmd.empty())
+		{
+			wxMessageBox(wxString::Format(_("The file '%s' could not be opened:\nNo program has been associated on your system with this file type."), entry.name.c_str()), _("Opening failed"), wxICON_EXCLAMATION);
+			return;
+		}
+	}
+	if (!program_exists)
+	{
+		wxString msg = wxString::Format(_("The file '%s' cannot be opened:\nThe associated program (%s) could not be found.\nPlease check your filetype associations."), entry.name.c_str(), cmd.c_str());
+		wxMessageBox(msg, _("Cannot edit file"), wxICON_EXCLAMATION);
 		return;
 	}
 	if (dangerous)
@@ -2393,15 +2624,9 @@ void CRemoteListView::OnMenuEdit(wxCommandEvent& event)
 			wxBell();
 			return;
 		}
-
-		if (!m_pState->IsRemoteConnected() || !m_pDirectoryListing)
-		{
-			wxBell();
-			return;
-		}
 	}
 
-	CEditHandler::fileState state = pEditHandler->GetFileState(CEditHandler::remote, entry.name);
+	CEditHandler::fileState state = pEditHandler->GetFileState(entry.name, path, server);
 	switch (state)
 	{
 	case CEditHandler::download:
@@ -2434,38 +2659,32 @@ void CRemoteListView::OnMenuEdit(wxCommandEvent& event)
 
 			if (XRCCTRL(dlg, "ID_REOPEN", wxRadioButton)->GetValue())
 			{
-				pEditHandler->StartEditing(CEditHandler::remote, entry.name);
+				pEditHandler->StartEditing(entry.name, path, server);
 				return;
 			}
 			else
 			{
-				if (!pEditHandler->Remove(CEditHandler::remote, entry.name))
+				if (!pEditHandler->Remove(entry.name, path, server))
 				{
 					wxMessageBox(_("The selected file is still opened in some other program, please close it."), _("Selected file still being edited"), wxICON_EXCLAMATION);
 					return;
 				}
 			}
-
-			if (!m_pState->IsRemoteConnected() || !m_pDirectoryListing)
-			{
-				wxBell();
-				return;
-			}
 		}
 		break;
 	default:
 		break;
 	}
 
-	if (!pEditHandler->AddFile(CEditHandler::remote, entry.name, m_pDirectoryListing->path, *m_pState->GetServer()))
+	wxString file = entry.name;
+	if (!pEditHandler->AddFile(CEditHandler::remote, file, path, server))
 	{
 		wxFAIL;
 		wxBell();
 		return;
 	}
 
-	wxFileName fn = wxFileName(localDir, entry.name);
-	m_pQueue->QueueFile(false, true, fn.GetFullPath(), entry.name, m_pDirectoryListing->path, *m_pState->GetServer(), entry.size, CEditHandler::remote);
+	m_pQueue->QueueFile(false, true, file, entry.name, path, server, entry.size, CEditHandler::remote);
 	m_pQueue->QueueFile_Finish(true);
 }
 
@@ -2591,7 +2810,7 @@ void CRemoteListView::FinishComparison()
 
 	ComparisonRestoreSelections();
 
-	Refresh();
+	RefreshListOnly();
 }
 
 wxListItemAttr* CRemoteListView::OnGetItemAttr(long item) const
@@ -2684,7 +2903,7 @@ wxString CRemoteListView::GetItemText(int item, unsigned int column)
 	else if (column == 2)
 	{
 		CGenericFileData& data = m_fileData[index];
-		if (data.fileType == _T(""))
+		if (data.fileType.IsEmpty())
 		{
 			const CDirentry& entry = (*m_pDirectoryListing)[index];
 			if (m_pDirectoryListing->path.GetType() == VMS)
@@ -2764,3 +2983,19 @@ wxLongLong CRemoteListView::ItemGetSize(int index) const
 {
 	return (*m_pDirectoryListing)[index].size;
 }
+
+void CRemoteListView::LinkIsNotDir(const CServerPath& path, const wxString& link)
+{
+	if (!m_pLinkResolveState)
+		return;
+
+	if (m_pLinkResolveState->remote_path == path && m_pLinkResolveState->link == link)
+	{
+		wxFileName fn(m_pLinkResolveState->local_path, link);
+		m_pQueue->QueueFile(false, true, fn.GetFullPath(), link, m_pLinkResolveState->remote_path, m_pLinkResolveState->server, -1);
+		m_pQueue->QueueFile_Finish(true);
+	}
+
+	delete m_pLinkResolveState;
+	m_pLinkResolveState = 0;
+}
