@@ -36,7 +36,8 @@ public:
 	// If files are locked, they won't be removed though
 	void Release();
 
-	enum fileState GetFileState(enum fileType type, const wxString& fileName) const;
+	enum fileState GetFileState(const wxString& fileName) const; // Local files
+	enum fileState GetFileState(const wxString& fileName, const CServerPath& remotePath, const CServer& server) const; // Remote files
 
 	// Returns the number of files in given state
 	// pServer may be set only if state isn't unknown
@@ -44,14 +45,16 @@ public:
 
 	// Adds the file that doesn't exist yet. (Has to be in unknown state)
 	// The initial state will be download
-	bool AddFile(enum fileType type, const wxString& fileName, const CServerPath& remotePath, const CServer& server);
+	bool AddFile(enum fileType type, wxString& fileName, const CServerPath& remotePath, const CServer& server);
 
 	// Tries to unedit and remove file
-	bool Remove(enum fileType type, const wxString& fileName);
+	bool Remove(const wxString& fileName); // Local files
+	bool Remove(const wxString& fileName, const CServerPath& remotePath, const CServer& server); // Remote files
 	bool RemoveAll(bool force);
 	bool RemoveAll(enum fileState state, const CServer* pServer = 0);
 
-	void FinishTransfer(bool successful, enum fileType type, const wxString& fileName);
+	void FinishTransfer(bool successful, const wxString& fileName);
+	void FinishTransfer(bool successful, const wxString& fileName, const CServerPath& remotePath, const CServer& server);
 
 	void CheckForModifications(
 #ifdef __WXMAC__
@@ -69,12 +72,14 @@ public:
 	 * The dangerous argument will be set to true on some filetypes,
 	 * e.g. executables.
 	 */
-	bool CanOpen(enum fileType type, const wxString& fileName, bool &dangerous);
-	bool StartEditing(enum fileType type, const wxString& file);
-
+	wxString CanOpen(enum fileType type, const wxString& fileName, bool &dangerous, bool& program_exists);
+	bool StartEditing(const wxString& file);
+	bool StartEditing(const wxString& file, const CServerPath& remotePath, const CServer& server);
+	
 	struct t_fileData
 	{
-		wxString name;
+		wxString name; // The name of the file
+		wxString file; // The actual local filename
 		fileState state;
 		wxDateTime modificationTime;
 		CServerPath remotePath;
@@ -83,10 +88,13 @@ public:
 
 	const std::list<t_fileData>& GetFiles(enum fileType type) const { wxASSERT(type != none); return m_fileDataList[type]; }
 
-	bool UploadFile(enum fileType type, const wxString& fileName, bool unedit);
+	bool UploadFile(const wxString& file, bool unedit);
+	bool UploadFile(const wxString& file, const CServerPath& remotePath, const CServer& server, bool unedit);
 
-	wxString GetOpenCommand(const wxString& file);
-	wxString GetSystemOpenCommand(const wxString& file);
+	// Returns command to open the file. If association is set but
+	// program does not exist, program_exists is set to false.
+	wxString GetOpenCommand(const wxString& file, bool& program_exists);
+	wxString GetSystemOpenCommand(const wxString& file, bool &program_exists);
 
 protected:
 	CEditHandler();
@@ -98,14 +106,18 @@ protected:
 
 	bool StartEditing(enum fileType type, t_fileData &data);
 
-	wxString GetCustomOpenCommand(const wxString& file);
+	wxString GetCustomOpenCommand(const wxString& file, bool& program_exists);
 
 	void SetTimerState();
 
+	bool UploadFile(enum fileType type, std::list<t_fileData>::iterator iter, bool unedit);
+
 	std::list<t_fileData> m_fileDataList[2];
 
-	std::list<t_fileData>::iterator GetFile(enum fileType type, const wxString& fileName);
-	std::list<t_fileData>::const_iterator GetFile(enum fileType type, const wxString& fileName) const;
+	std::list<t_fileData>::iterator GetFile(const wxString& fileName);
+	std::list<t_fileData>::const_iterator GetFile(const wxString& fileName) const;
+	std::list<t_fileData>::iterator GetFile(const wxString& fileName, const CServerPath& remotePath, const CServer& server);
+	std::list<t_fileData>::const_iterator GetFile(const wxString& fileName, const CServerPath& remotePath, const CServer& server) const;
 
 	CQueueView* m_pQueue;
 
@@ -113,6 +125,12 @@ protected:
 
 	void RemoveTemporaryFiles(const wxString& temp);
 
+	wxString GetTemporaryFile(wxString name);
+	wxString TruncateFilename(const wxString path, const wxString& name, int max);
+	bool FilenameExists(const wxString& file);
+
+	int DisplayChangeNotification(fileType type, std::list<t_fileData>::const_iterator iter, bool& remove);
+
 #ifdef __WXMSW__
 	HANDLE m_lockfile_handle;
 #else
@@ -136,6 +154,8 @@ public:
 protected:
 	void SetCtrlState();
 
+	CEditHandler::t_fileData* GetDataFromItem(int item, CEditHandler::fileType &type);
+
 	wxWindow* m_pParent;
 
 	DECLARE_EVENT_TABLE()
@@ -145,4 +165,22 @@ protected:
 	void OnEdit(wxCommandEvent& event);
 };
 
+class CNewAssociationDialog : protected wxDialogEx
+{
+public:
+	CNewAssociationDialog(wxWindow* parent);
+
+	bool Show(const wxString& file);
+
+protected:
+	void SetCtrlState();
+	wxWindow* m_pParent;
+	wxString m_ext;
+
+	DECLARE_EVENT_TABLE();
+	void OnRadioButton(wxCommandEvent& event);
+	void OnOK(wxCommandEvent& event);
+	void OnBrowseEditor(wxCommandEvent& event);
+};
+
 #endif //__EDITHANDLER_H__
