@@ -319,13 +319,14 @@ void CFileItem::SaveItem(TiXmlElement* pElement) const
 	AddTextElement(&file, "LocalFile", m_localFile);
 	AddTextElement(&file, "RemoteFile", m_remoteFile);
 	AddTextElement(&file, "RemotePath", m_remotePath.GetSafePath());
-	AddTextElement(&file, "Download", m_download ? _T("1") : _T("0"));
+	AddTextElementRaw(&file, "Download", m_download ? "1" : "0");
 	if (m_size != -1)
 		AddTextElement(&file, "Size", m_size.ToString());
 	if (m_errorCount)
 		AddTextElement(&file, "ErrorCount", m_errorCount);
-	AddTextElement(&file, "Priority", m_priority);
-	AddTextElement(&file, "TransferMode", m_transferSettings.binary ? _T("1") : _T("0"));
+	if (m_priority != priority_normal)
+		AddTextElement(&file, "Priority", m_priority);
+	AddTextElementRaw(&file, "TransferMode", m_transferSettings.binary ? "1" : "0");
 
 	pElement->InsertEndChild(file);
 }
@@ -339,6 +340,11 @@ bool CFileItem::TryRemoveAll()
 	return false;
 }
 
+void CFileItem::SetLocalFile(const wxString file)
+{
+	m_localFile = file;
+}
+
 CFolderItem::CFolderItem(CServerItem* parent, bool queued, const wxString& localFile)
 	: CFileItem(parent, queued, true, localFile, _T(""), CServerPath(), -1)
 {
@@ -360,7 +366,7 @@ void CFolderItem::SaveItem(TiXmlElement* pElement) const
 		AddTextElement(&file, "RemoteFile", m_remoteFile);
 		AddTextElement(&file, "RemotePath", m_remotePath.GetSafePath());
 	}
-	AddTextElement(&file, "Download", m_download ? _T("1") : _T("0"));
+	AddTextElementRaw(&file, "Download", m_download ? "1" : "0");
 
 	pElement->InsertEndChild(file);
 }
@@ -688,10 +694,7 @@ CFolderScanItem::CFolderScanItem(CServerItem* parent, bool queued, bool download
 	m_remove = false;
 	m_active = false;
 	m_count = 0;
-	t_dirPair pair;
-	pair.localPath = localPath.c_str();
-	pair.remotePath.SetSafePath(remotePath.GetSafePath().c_str());
-	m_dirsToCheck.push_back(pair);
+	m_dir_is_empty = false;
 
 	m_defaultFileExistsAction = CFileExistsNotification::unknown;
 }
@@ -1330,7 +1333,7 @@ bool CQueueViewBase::RemoveItem(CQueueItem* pItem, bool destroy, bool updateItem
 		if (oldCount > m_itemCount)
 		{
 			bool eraseBackground = GetTopItem() + GetCountPerPage() + 1 >= m_itemCount;
-			Refresh(eraseBackground);
+			RefreshListOnly(eraseBackground);
 			if (eraseBackground)
 				Update();
 		}
@@ -1344,7 +1347,13 @@ void CQueueViewBase::RefreshItem(const CQueueItem* pItem)
 	wxASSERT(pItem);
 	int index = GetItemIndex(pItem);
 
+#ifdef __WXMSW__
+	wxRect rect;
+	GetItemRect(index, rect);
+	RefreshRect(rect, false);
+#else
 	wxListCtrl::RefreshItem(index);
+#endif
 }
 
 void CQueueViewBase::OnNavigationKey(wxNavigationKeyEvent& event)
