@@ -156,6 +156,8 @@ public:
 	virtual bool TryRemoveAll(); // Removes a inactive childrens, queues active children for removal.
 								 // Returns true if item can be removed itself
 
+	void SetLocalFile(const wxString file);
+
 	bool m_queued;
 	int m_errorCount;
 	bool m_remove;
@@ -218,15 +220,12 @@ public:
 
 	int m_count;
 
-	struct t_dirPair
-	{
-		wxString localPath;
-		CServerPath remotePath;
-	};
-	std::list<t_dirPair> m_dirsToCheck;
-
 	enum CFileExistsNotification::OverwriteAction m_defaultFileExistsAction;
 
+	bool m_dir_is_empty;
+	wxString m_current_local_path;
+	CServerPath m_current_remote_path;
+
 protected:
 	wxString m_localPath;
 	CServerPath m_remotePath;
