@@ -70,7 +70,7 @@ void CQueueViewFailed::OnRemoveAll(wxCommandEvent& event)
 
 	DisplayNumberQueuedFiles();
 
-	Refresh();
+	RefreshListOnly();
 
 	if (!m_itemCount && m_pQueue->GetQueueView()->GetItemCount())
 		m_pQueue->SetSelection(0);
@@ -117,9 +117,19 @@ void CQueueViewFailed::OnRemoveSelected(wxCommandEvent& event)
 			if (pFileItem->m_edit == CEditHandler::remote && pEditHandler)
 			{
 				wxFileName fn(pFileItem->GetLocalFile());
-				enum CEditHandler::fileState state = pEditHandler->GetFileState(CEditHandler::remote, fn.GetFullName());
-				if (state == CEditHandler::upload_and_remove_failed)
-					pEditHandler->Remove(pFileItem->m_edit, fn.GetFullName());
+				if (pFileItem->m_edit == CEditHandler::local)
+				{
+					enum CEditHandler::fileState state = pEditHandler->GetFileState(pFileItem->GetLocalFile());
+					if (state == CEditHandler::upload_and_remove_failed)
+						pEditHandler->Remove(pFileItem->GetLocalFile());
+				}
+				else
+				{
+					CServerItem* pServerItem = (CServerItem*)pFileItem->GetTopLevelItem();
+					enum CEditHandler::fileState state = pEditHandler->GetFileState(pFileItem->GetRemoteFile(), pFileItem->GetRemotePath(), pServerItem->GetServer());
+					if (state == CEditHandler::upload_and_remove_failed)
+						pEditHandler->Remove(pFileItem->GetRemoteFile(), pFileItem->GetRemotePath(), pServerItem->GetServer());
+				}
 			}
 		}
 
@@ -134,7 +144,7 @@ void CQueueViewFailed::OnRemoveSelected(wxCommandEvent& event)
 	}
 	DisplayNumberQueuedFiles();
 	SaveSetItemCount(m_itemCount);
-	Refresh();
+	RefreshListOnly();
 
 	if (!m_itemCount && m_pQueue->GetQueueView()->GetItemCount())
 		m_pQueue->SetSelection(0);
@@ -205,22 +215,23 @@ void CQueueViewFailed::OnRequeueSelected(wxCommandEvent& event)
 						delete pFileItem;
 						continue;
 					}
-					wxFileName fn(pFileItem->GetLocalFile());
-					enum CEditHandler::fileState state = pEditHandler->GetFileState(CEditHandler::remote, fn.GetFullName());
+					enum CEditHandler::fileState state = pEditHandler->GetFileState(pFileItem->GetRemoteFile(), pFileItem->GetRemotePath(), pServerItem->GetServer());
 					if (state == CEditHandler::unknown)
 					{
 						wxASSERT(pFileItem->Download());
-						if (!pEditHandler->AddFile(CEditHandler::remote, fn.GetFullName(), pFileItem->GetRemotePath(), pServerItem->GetServer()))
+						wxString file = pFileItem->GetRemoteFile();
+						if (!pEditHandler->AddFile(CEditHandler::remote, file, pFileItem->GetRemotePath(), pServerItem->GetServer()))
 						{
 							failedToRequeueAll = true;
 							delete pFileItem;
 							continue;
 						}
+						pFileItem->SetLocalFile(file);
 					}
 					else if (state == CEditHandler::upload_and_remove_failed)
 					{
 						wxASSERT(!pFileItem->Download());
-						if (!pEditHandler->UploadFile(pFileItem->m_edit, fn.GetFullName(), true))
+						if (!pEditHandler->UploadFile(pFileItem->GetRemoteFile(), pFileItem->GetRemotePath(), pServerItem->GetServer(), true))
 							failedToRequeueAll = true;
 						delete pFileItem;
 						continue;
@@ -289,12 +300,12 @@ void CQueueViewFailed::OnRequeueSelected(wxCommandEvent& event)
 					delete pItem;
 					continue;
 				}
-				wxFileName fn(pFileItem->GetLocalFile());
-				enum CEditHandler::fileState state = pEditHandler->GetFileState(CEditHandler::remote, fn.GetFullName());
+				enum CEditHandler::fileState state = pEditHandler->GetFileState(pFileItem->GetRemoteFile(), pFileItem->GetRemotePath(), pServerItem->GetServer());
 				if (state == CEditHandler::unknown)
 				{
 					wxASSERT(pFileItem->Download());
-					if (!pEditHandler->AddFile(CEditHandler::remote, fn.GetFullName(), pFileItem->GetRemotePath(), pServerItem->GetServer()))
+					wxString file = pFileItem->GetRemoteFile();
+					if (!pEditHandler->AddFile(CEditHandler::remote, file, pFileItem->GetRemotePath(), pServerItem->GetServer()))
 					{
 						if (!pServerItem->GetChildrenCount(false))
 						{
@@ -306,11 +317,12 @@ void CQueueViewFailed::OnRequeueSelected(wxCommandEvent& event)
 						delete pItem;
 						continue;
 					}
+					pFileItem->SetLocalFile(file);
 				}
 				else if (state == CEditHandler::upload_and_remove_failed)
 				{
 					wxASSERT(!pFileItem->Download());
-					if (!pEditHandler->UploadFile(pFileItem->m_edit, fn.GetFullName(), true))
+					if (!pEditHandler->UploadFile(pFileItem->GetRemoteFile(), pFileItem->GetRemotePath(), pServerItem->GetServer(), true))
 						failedToRequeueAll = true;
 
 					if (!pServerItem->GetChildrenCount(false))
@@ -346,7 +358,7 @@ void CQueueViewFailed::OnRequeueSelected(wxCommandEvent& event)
 
 	DisplayNumberQueuedFiles();
 	SaveSetItemCount(m_itemCount);
-	Refresh();
+	RefreshListOnly();
 
 	if (!m_itemCount && m_pQueue->GetQueueView()->GetItemCount())
 		m_pQueue->SetSelection(0);
