@@ -9,6 +9,7 @@
 #include "RemoteListView.h"
 #include "recursive_operation.h"
 #include "statusbar.h"
+#include "local_filesys.h"
 
 CState::CState(CMainFrame* pMainFrame)
 {
@@ -16,11 +17,14 @@ CState::CState(CMainFrame* pMainFrame)
 
 	m_pDirectoryListing = 0;
 	m_pServer = 0;
+	m_successful_connect = 0;
 
 	m_pEngine = 0;
 	m_pCommandQueue = 0;
 
 	m_pRecursiveOperation = new CRecursiveOperation(this);
+
+	m_sync_browse.is_changing = false;
 }
 
 CState::~CState()
@@ -41,99 +45,69 @@ CState::~CState()
 	delete m_pRecursiveOperation;
 }
 
-wxString CState::GetLocalDir() const
+CLocalPath CState::GetLocalDir() const
 {
 	return m_localDir;
 }
 
-wxString CState::Canonicalize(wxString oldDir, wxString newDir, wxString *error /*=0*/)
+bool CState::SetLocalDir(const wxString& dir, wxString *error /*=0*/)
 {
-#ifdef __WXMSW__
-	if (newDir == _T("\\") || newDir == _T("/") || newDir == _T(""))
-		return _T("\\");
-
-	// "Go up one level" is a little bit difficult under Windows due to
-	// things like "My Computer" and "Desktop"
-	if (newDir == _T(".."))
+	if (m_sync_browse.is_changing)
 	{
-		newDir = oldDir;
-		if (newDir != _T("\\"))
-		{
-			newDir.RemoveLast();
-			int pos = newDir.Find('\\', true);
-			if (pos == -1)
-				return _T("\\");
-			else
-				newDir = newDir.Left(pos + 1);
-		}
-	}
-	else
-#endif
-	{
-		wxFileName dir(newDir, _T(""));
-		{
-			wxLogNull noLog;
-			if (!dir.MakeAbsolute(oldDir))
-				return _T("");
-		}
-		newDir = dir.GetFullPath();
-		if (newDir.Right(1) != wxFileName::GetPathSeparator())
-			newDir += wxFileName::GetPathSeparator();
+		wxMessageBox(_T("Cannot change directory, there already is a synchronized browsing operation in progress."), _("Synchronized browsing"));
+		return false;
 	}
 
-	// Check for partial UNC paths
-	if (newDir.Left(2) == _T("\\\\"))
-	{
-		int pos = newDir.Mid(2).Find('\\');
-		if (pos == -1 || pos + 3 == (int)newDir.Len())
-		{
-			// Partial UNC path, no share given
-			return newDir;
-		}
-
-		pos = newDir.Mid(pos + 3).Find('\\');
-		if (pos == -1)
-		{
-			// Partial UNC path, no full share yet, skip further processing
-			return _T("");
-		}
-	}
-
-	if (!wxDir::Exists(newDir))
-	{
-		if (!error)
-			return _T("");
-
-		*error = wxString::Format(_("'%s' does not exist or cannot be accessed."), newDir.c_str());
-
-#ifdef __WXMSW__
-		if (newDir[0] == '\\')
-			return _T("");
-
-		// Check for removable drive, display a more specific error message in that case
-		if (::GetLastError() != ERROR_NOT_READY)
-			return _T("");
-		int type = GetDriveType(newDir.Left(3));
-		if (type == DRIVE_REMOVABLE || type == DRIVE_CDROM)
-
-			*error = wxString::Format(_("Cannot access '%s', no media inserted or drive not ready."), newDir.c_str());
-#endif
-
-		return _T("");
-	}
-
-	return newDir;
-}
-
-bool CState::SetLocalDir(wxString dir, wxString *error /*=0*/)
-{
-	dir = Canonicalize(m_localDir, dir, error);
-	if (dir == _T(""))
+	CLocalPath p(m_localDir);
+	if (!p.ChangePath(dir))
 		return false;
 
-	m_localDir = dir;
+	if (!p.Exists(error))
+		return false;
 
-	COptions::Get()->SetOption(OPTION_LASTLOCALDIR, dir);
+	if (!m_sync_browse.local_root.empty())
+	{
+		wxASSERT(m_pServer);
+		
+		if (p != m_sync_browse.local_root && !p.IsSubdirOf(m_sync_browse.local_root))
+		{
+			wxString msg = wxString::Format(_("The local directory '%s' is not below the synchronization root (%s).\nDisable synchronized browsing and continue changing the local directory?"),
+					p.GetPath().c_str(),
+					m_sync_browse.local_root.GetPath().c_str());
+			if (wxMessageBox(msg, _("Synchronized browsing"), wxICON_QUESTION | wxYES_NO) != wxYES)
+				return false;
+			SetSyncBrowse(false);
+		}
+		else if (!IsRemoteIdle())
+		{
+			wxString msg(_("A remote operation is in progress and synchronized browsing is enabled.\nDisable synchronized browsing and continue changing the local directory?"));
+			if (wxMessageBox(msg, _("Synchronized browsing"), wxICON_QUESTION | wxYES_NO) != wxYES)
+				return false;
+			SetSyncBrowse(false);
+		}
+		else
+		{
+			CServerPath remote_path = GetSynchronizedDirectory(p);
+			if (remote_path.IsEmpty())
+			{
+				SetSyncBrowse(false);
+				wxString msg = wxString::Format(_("Could not obtain corresponding remote directory for the local directory '%s'.\nSynchronized browsing has been disabled."),
+					p.GetPath().c_str());
+				wxMessageBox(msg, _("Synchronized browsing"));
+				return false;
+			}
+
+			m_sync_browse.is_changing = true;
+			CListCommand *pCommand = new CListCommand(remote_path);
+			m_pCommandQueue->ProcessCommand(pCommand);
+
+			return true;
+		}
+	}
+	
+	m_localDir = p;
+
+	COptions::Get()->SetOption(OPTION_LASTLOCALDIR, m_localDir.GetPath());
 
 	NotifyHandlers(STATECHANGE_LOCAL_DIR);
 
@@ -144,6 +118,7 @@ bool CState::SetRemoteDir(const CDirectoryListing *pDirectoryListing, bool modif
 {
 	if (!pDirectoryListing)
 	{
+		SetSyncBrowse(false);
 		if (modified)
 			return false;
 
@@ -157,6 +132,8 @@ bool CState::SetRemoteDir(const CDirectoryListing *pDirectoryListing, bool modif
 		return true;
 	}
 
+	wxASSERT(pDirectoryListing->m_firstListTime.IsValid());
+
 	if (modified)
 	{
 		if (!m_pDirectoryListing || m_pDirectoryListing->path != pDirectoryListing->path)
@@ -187,6 +164,45 @@ bool CState::SetRemoteDir(const CDirectoryListing *pDirectoryListing, bool modif
 
 	delete pOldListing;
 
+	if (m_sync_browse.is_changing)
+	{
+		m_sync_browse.is_changing = false;
+		if (m_pDirectoryListing->path != m_sync_browse.remote_root && !m_pDirectoryListing->path.IsSubdirOf(m_sync_browse.remote_root, false))
+		{
+			SetSyncBrowse(false);
+			wxString msg = wxString::Format(_("Current remote directory (%s) is not below the synchronization root (%s).\nSynchronized browsing has been disabled."),
+					m_pDirectoryListing->path.GetPath().c_str(),
+					m_sync_browse.remote_root.GetPath().c_str());
+			wxMessageBox(msg, _("Synchronized browsing"));
+		}
+		else
+		{
+			CLocalPath local_path = GetSynchronizedDirectory(m_pDirectoryListing->path);
+			if (local_path.empty())
+			{
+				SetSyncBrowse(false);
+				wxString msg = wxString::Format(_("Could not obtain corresponding local directory for the remote directory '%s'.\nSynchronized browsing has been disabled."),
+					m_pDirectoryListing->path.GetPath().c_str());
+				wxMessageBox(msg, _("Synchronized browsing"));
+				return true;
+			}
+
+			wxString error;
+			if (!local_path.Exists(&error))
+			{
+				SetSyncBrowse(false);
+				wxString msg = error + _T("\n") + _("Synchronized browsing has been disabled.");
+				wxMessageBox(msg, _("Synchronized browsing"));
+				return true;
+			}
+
+			m_localDir = local_path;
+
+			COptions::Get()->SetOption(OPTION_LASTLOCALDIR, m_localDir.GetPath());
+
+			NotifyHandlers(STATECHANGE_LOCAL_DIR);
+		}
+	}
 	return true;
 }
 
@@ -210,43 +226,40 @@ void CState::RefreshLocal()
 
 void CState::RefreshLocalFile(wxString file)
 {
-	wxFileName fn(file);
-	if (!fn.IsOk())
-		return;
-	if (!fn.IsAbsolute())
+	wxString file_name;
+	CLocalPath path(file, &file_name);
+	if (path.empty())
 		return;
 
-	wxString path = fn.GetPath(wxPATH_GET_VOLUME | wxPATH_GET_SEPARATOR);
-	if (path == _T(""))
-		return;
-
-	if (path == m_localDir)
+	if (file_name.empty())
 	{
-		file = fn.GetFullName();
-		if (file == _T(""))
+		if (!path.HasParent())
 			return;
+		path.MakeParent(&file_name);
+		wxASSERT(!file_name.empty());
 	}
-	else if (path.Left(m_localDir.Len()) == m_localDir)
-	{
-		path = path.Mid(m_localDir.Len());
-		int pos = path.Find(wxFileName::GetPathSeparator());
-		if (pos < 1)
-			return;
+	if (path != m_localDir)
+		return;
 
-		file = path.Left(pos);
-	}
-
-	NotifyHandlers(STATECHANGE_LOCAL_REFRESH_FILE, file);
+	NotifyHandlers(STATECHANGE_LOCAL_REFRESH_FILE, file_name);
 }
 
 void CState::SetServer(const CServer* server)
 {
 	if (m_pServer)
 	{
+		if (server && *server == *m_pServer)
+		{
+			// Nothing changes
+			return;
+		}
+
 		SetRemoteDir(0);
 		delete m_pServer;
 	}
 
+	m_successful_connect = false;
+
 	CStatusBar* const pStatusBar = m_pMainFrame->GetStatusBar();
 	if (pStatusBar)
 	{
@@ -284,7 +297,7 @@ bool CState::Connect(const CServer& server, bool askBreak, const CServerPath& pa
 	}
 
 	m_pCommandQueue->ProcessCommand(new CConnectCommand(server));
-	m_pCommandQueue->ProcessCommand(new CListCommand(path));
+	m_pCommandQueue->ProcessCommand(new CListCommand(path, _T(""), LIST_FLAG_FALLBACK_CURRENT));
 
 	COptions::Get()->SetLastServer(server);
 	COptions::Get()->SetOption(OPTION_LASTSERVERPATH, path.GetSafePath());
@@ -431,9 +444,9 @@ void CState::UploadDroppedFiles(const wxFileDataObject* pFileDataObject, const C
 		else if (wxDir::Exists(files[i]))
 		{
 			wxString dir = files[i];
-			if (dir.Last() == wxFileName::GetPathSeparator() && dir.Len() > 1)
+			if (dir.Last() == CLocalPath::path_separator && dir.Len() > 1)
 				dir.RemoveLast();
-			int pos = dir.Find(wxFileName::GetPathSeparator(), true);
+			int pos = dir.Find(CLocalPath::path_separator, true);
 			if (pos != -1 && pos != (int)dir.Len() - 1)
 			{
 				wxString lastSegment = dir.Mid(pos + 1);
@@ -445,11 +458,8 @@ void CState::UploadDroppedFiles(const wxFileDataObject* pFileDataObject, const C
 	}
 }
 
-void CState::HandleDroppedFiles(const wxFileDataObject* pFileDataObject, wxString path, bool copy)
+void CState::HandleDroppedFiles(const wxFileDataObject* pFileDataObject, const CLocalPath& path, bool copy)
 {
-	if (path.Last() != wxFileName::GetPathSeparator())
-		path += wxFileName::GetPathSeparator();
-
 	const wxArrayString &files = pFileDataObject->GetFilenames();
 	if (!files.Count())
 		return;
@@ -460,6 +470,8 @@ void CState::HandleDroppedFiles(const wxFileDataObject* pFileDataObject, wxStrin
 	for (unsigned int i = 0; i < files.Count(); i++)
 		len += files[i].Len() + 1;
 
+	// SHFILEOPSTRUCT's pTo and pFrom accept null-terminated lists
+	// of null-terminated filenames.
 	wxChar* from = new wxChar[len];
 	wxChar* p = from;
 	for (unsigned int i = 0; i < files.Count(); i++)
@@ -467,11 +479,11 @@ void CState::HandleDroppedFiles(const wxFileDataObject* pFileDataObject, wxStrin
 		wxStrcpy(p, files[i]);
 		p += files[i].Len() + 1;
 	}
-	*p = 0;
+	*p = 0; // End of list
 
-	wxChar* to = new wxChar[path.Len() + 2];
-	wxStrcpy(to, path);
-	to[path.Len() + 1] = 0;
+	wxChar* to = new wxChar[path.GetPath().Len() + 2];
+	wxStrcpy(to, path.GetPath());
+	to[path.GetPath().Len() + 1] = 0; // End of list
 
 	SHFILEOPSTRUCT op = {0};
 	op.pFrom = from;
@@ -488,26 +500,26 @@ void CState::HandleDroppedFiles(const wxFileDataObject* pFileDataObject, wxStrin
 		const wxString& file = files[i];
 		if (wxFile::Exists(file))
 		{
-			int pos = file.Find(wxFileName::GetPathSeparator(), true);
+			int pos = file.Find(CLocalPath::path_separator, true);
 			if (pos == -1 || pos == (int)file.Len() - 1)
 				continue;
 			const wxString& name = file.Mid(pos + 1);
 			if (copy)
-				wxCopyFile(file, path + name);
+				wxCopyFile(file, path.GetPath() + name);
 			else
-				wxRenameFile(file, path + name);
+				wxRenameFile(file, path.GetPath() + name);
 		}
 		else if (wxDir::Exists(file))
 		{
 			if (copy)
-				RecursiveCopy(file, path);
+				RecursiveCopy(file, path.GetPath());
 			else
 			{
-				int pos = file.Find(wxFileName::GetPathSeparator(), true);
+				int pos = file.Find(CLocalPath::path_separator, true);
 				if (pos == -1 || pos == (int)file.Len() - 1)
 					continue;
 				const wxString& name = file.Mid(pos + 1);
-				wxRenameFile(file, path + name);
+				wxRenameFile(file, path.GetPath() + name);
 			}
 		}
 	}
@@ -516,66 +528,65 @@ void CState::HandleDroppedFiles(const wxFileDataObject* pFileDataObject, wxStrin
 	RefreshLocal();
 }
 
-bool CState::RecursiveCopy(wxString source, wxString target)
+bool CState::RecursiveCopy(CLocalPath source, const CLocalPath& target)
 {
-	if (source == _T(""))
+	if (source.empty() || target.empty())
 		return false;
 
-	if (target == _T(""))
+	if (source == target)
 		return false;
 
-	if (target.Last() != wxFileName::GetPathSeparator())
-		target += wxFileName::GetPathSeparator();
-
-	if (source.Last() == wxFileName::GetPathSeparator())
-		source.RemoveLast();
-
-	if (source + wxFileName::GetPathSeparator() == target)
+	if (source.IsParentOf(target))
 		return false;
 
-	if (target.Len() > source.Len() && source == target.Left(source.Len()) && target[source.Len()] == wxFileName::GetPathSeparator())
+	if (!source.HasParent())
 		return false;
 
-	int pos = source.Find(wxFileName::GetPathSeparator(), true);
-	if (pos == -1 || pos == (int)source.Len() - 1)
+	wxString last_segment;
+	if (!source.MakeParent(&last_segment))
 		return false;
 
 	std::list<wxString> dirsToVisit;
-	dirsToVisit.push_back(source.Mid(pos + 1) + wxFileName::GetPathSeparator());
-	source = source.Left(pos + 1);
+	dirsToVisit.push_back(last_segment + CLocalPath::path_separator);
 
 	// Process any subdirs which still have to be visited
 	while (!dirsToVisit.empty())
 	{
 		wxString dirname = dirsToVisit.front();
 		dirsToVisit.pop_front();
-		wxMkdir(target + dirname);
-		wxDir dir;
-		if (!dir.Open(source + dirname))
+		wxMkdir(target.GetPath() + dirname);
+
+		CLocalFileSystem fs;
+		if (!fs.BeginFindFiles(source.GetPath() + dirname, false))
 			continue;
 
+		bool is_dir, is_link;
 		wxString file;
-		for (bool found = dir.GetFirst(&file); found; found = dir.GetNext(&file))
+		while (fs.GetNextFile(file, is_link, is_dir, 0, 0, 0))
 		{
 			if (file == _T(""))
 			{
 				wxGetApp().DisplayEncodingWarning();
 				continue;
 			}
-			if (wxFileName::DirExists(source + dirname + file))
+
+			if (is_dir)
 			{
-				const wxString subDir = dirname + file + wxFileName::GetPathSeparator();
+				if (is_link)
+					continue;
+
+				const wxString subDir = dirname + file + CLocalPath::path_separator;
 				dirsToVisit.push_back(subDir);
 			}
 			else
-				wxCopyFile(source + dirname + file, target + dirname + file);
+				wxCopyFile(source.GetPath() + dirname + file, target.GetPath() + dirname + file);
 		}
 	}
 
 	return true;
 }
 
-bool CState::DownloadDroppedFiles(const CRemoteDataObject* pRemoteDataObject, wxString path, bool queueOnly /*=false*/)
+bool CState::DownloadDroppedFiles(const CRemoteDataObject* pRemoteDataObject, const CLocalPath& path, bool queueOnly /*=false*/)
 {
 	bool hasDirs = false;
 	bool hasFiles = false;
@@ -595,7 +606,7 @@ bool CState::DownloadDroppedFiles(const CRemoteDataObject* pRemoteDataObject, wx
 	}
 
 	if (hasFiles)
-		m_pMainFrame->GetQueue()->QueueFiles(queueOnly, path, *pRemoteDataObject);
+		m_pMainFrame->GetQueue()->QueueFiles(queueOnly, path.GetPath(), *pRemoteDataObject);
 
 	if (!hasDirs)
 		return true;
@@ -608,7 +619,7 @@ bool CState::IsRemoteConnected() const
 	if (!m_pEngine)
 		return false;
 
-	return m_pEngine->IsConnected();
+	return m_pServer != 0;
 }
 
 bool CState::IsRemoteIdle() const
@@ -622,42 +633,6 @@ bool CState::IsRemoteIdle() const
 	return m_pCommandQueue->Idle();
 }
 
-bool CState::LocalDirHasParent(const wxString& dir)
-{
-#ifdef __WXMSW__
-	if (dir.Left(2) == _T("\\\\"))
-	{
-		int pos = dir.Mid(2).Find('\\');
-		if (pos == -1 || pos + 3 == (int)dir.Len())
-			return false;
-	}
-	if (dir == _T("\\") || dir == _T("//"))
-		return false;
-#endif
-
-	if (dir == _T("/"))
-		return false;
-
-	return true;
-}
-
-bool CState::LocalDirIsWriteable(const wxString& dir)
-{
-#ifdef __WXMSW__
-	if (dir == _T("\\") || dir == _T("//"))
-		return false;
-
-	if (dir.Left(2) == _T("\\\\"))
-	{
-		int pos = dir.Mid(2).Find('\\');
-		if (pos == -1 || pos + 3 == (int)dir.Len())
-			return false;
-	}
-#endif
-
-	return true;
-}
-
 wxString CState::GetAsURL(const wxString& dir)
 {
 	// Cheap URL encode
@@ -715,3 +690,153 @@ wxString CState::GetAsURL(const wxString& dir)
 
 	return _T("file://") + encoded;
 }
+
+void CState::ListingFailed(int error)
+{
+	m_sync_browse.is_changing = false;
+
+	// Let the recursive operation handler know if a LIST command failed,
+	// so that it may issue the next command in recursive operations.
+	m_pRecursiveOperation->ListingFailed(error);
+}
+
+void CState::LinkIsNotDir(const CServerPath& path, const wxString& subdir)
+{
+	m_sync_browse.is_changing = false;
+
+	if (m_pRecursiveOperation->GetOperationMode() != CRecursiveOperation::recursive_none)
+		m_pRecursiveOperation->LinkIsNotDir();
+	else
+		m_pMainFrame->GetRemoteListView()->LinkIsNotDir(path, subdir);
+}
+
+bool CState::ChangeRemoteDir(const CServerPath& path, const wxString& subdir /*=_T("")*/, int flags /*=0*/)
+{
+	if (!m_pServer || !m_pCommandQueue)
+		return false;
+
+	if (!m_sync_browse.local_root.empty())
+	{
+		CServerPath p(path);
+		if (!subdir.empty() && !p.ChangePath(subdir))
+		{
+			wxString msg = wxString::Format(_("Could not get full remote path."));
+			wxMessageBox(msg, _("Synchronized browsing"));
+			return false;
+		}
+
+		if (p != m_sync_browse.remote_root && !p.IsSubdirOf(m_sync_browse.remote_root, false))
+		{
+			wxString msg = wxString::Format(_("The remote directory '%s' is not below the synchronization root (%s).\nDisable synchronized browsing and continue changing the remote directory?"),
+					p.GetPath().c_str(),
+					m_sync_browse.remote_root.GetPath().c_str());
+			if (wxMessageBox(msg, _("Synchronized browsing"), wxICON_QUESTION | wxYES_NO) != wxYES)
+				return false;
+			SetSyncBrowse(false);
+		}
+		else if (!IsRemoteIdle())
+		{
+			wxString msg(_("Another remote operation is already progress, cannot change directory now"));
+			wxMessageBox(msg, _("Synchronized browsing"), wxICON_EXCLAMATION);
+			return false;
+		}
+		else
+		{
+			wxString error;
+			CLocalPath local_path = GetSynchronizedDirectory(p);
+			if (local_path.empty())
+			{
+				wxString msg = wxString::Format(_("Could not obtain corresponding local directory for the remote directory '%s'.\nDisable synchronized browsing and continue changing the remote directory?"),
+					p.GetPath().c_str());
+				if (wxMessageBox(msg, _("Synchronized browsing"), wxICON_QUESTION | wxYES_NO) != wxYES)
+					return false;
+				SetSyncBrowse(false);
+			}
+			else if (!local_path.Exists(&error))
+			{
+				wxString msg = error + _T("\n") + _("Disable synchronized browsing and continue changing the remote directory?");
+				if (wxMessageBox(msg, _("Synchronized browsing"), wxICON_QUESTION | wxYES_NO) != wxYES)
+					return false;
+				SetSyncBrowse(false);
+			}
+			else
+				m_sync_browse.is_changing = true;
+		}
+	}
+
+	CListCommand *pCommand = new CListCommand(path, subdir, flags);
+	m_pCommandQueue->ProcessCommand(pCommand);
+
+	return true;
+}
+
+bool CState::SetSyncBrowse(bool enable)
+{
+	if (enable != m_sync_browse.local_root.empty())
+		return enable;
+
+	if (!enable)
+	{
+		m_sync_browse.local_root.clear();
+		m_sync_browse.remote_root.Clear();
+		m_sync_browse.is_changing = false;
+
+		NotifyHandlers(STATECHANGE_SYNC_BROWSE);
+		return false;
+	}
+
+	if (!m_pDirectoryListing)
+		return false;
+
+	m_sync_browse.is_changing = false;
+	m_sync_browse.local_root = m_localDir;
+	m_sync_browse.remote_root = m_pDirectoryListing->path;
+
+	while (m_sync_browse.local_root.HasParent() && m_sync_browse.remote_root.HasParent() &&
+		m_sync_browse.local_root.GetLastSegment() == m_sync_browse.remote_root.GetLastSegment())
+	{
+		m_sync_browse.local_root.MakeParent();
+		m_sync_browse.remote_root = m_sync_browse.remote_root.GetParent();
+	}
+
+	NotifyHandlers(STATECHANGE_SYNC_BROWSE);
+	return true;
+}
+
+CLocalPath CState::GetSynchronizedDirectory(CServerPath remote_path)
+{
+	std::list<wxString> segments;
+	while (remote_path.HasParent() && remote_path != m_sync_browse.remote_root)
+	{
+		segments.push_front(remote_path.GetLastSegment());
+		remote_path = remote_path.GetParent();
+	}
+	if (remote_path != m_sync_browse.remote_root)
+		return CLocalPath();
+
+	CLocalPath local_path = m_sync_browse.local_root;
+	for (std::list<wxString>::const_iterator iter = segments.begin(); iter != segments.end(); iter++)
+		local_path.AddSegment(*iter);
+
+	return local_path;
+}
+
+
+CServerPath CState::GetSynchronizedDirectory(CLocalPath local_path)
+{
+	std::list<wxString> segments;
+	while (local_path.HasParent() && local_path != m_sync_browse.local_root)
+	{
+		wxString last;
+		local_path.MakeParent(&last);
+		segments.push_front(last);
+	}
+	if (local_path != m_sync_browse.local_root)
+		return CServerPath();
+
+	CServerPath remote_path = m_sync_browse.remote_root;
+	for (std::list<wxString>::const_iterator iter = segments.begin(); iter != segments.end(); iter++)
+		remote_path.AddSegment(*iter);
+
+	return remote_path;
+}
