@@ -210,11 +210,44 @@ void wxStatusBarEx::SetStatusText(const wxString& text, int number /*=0*/)
 }
 #endif
 
-class CEncryptionIndicator : public wxStaticBitmap
+#ifdef __WXMSW__
+class wxStaticBitmapEx : public wxStaticBitmap
+{
+public:
+	wxStaticBitmapEx(wxWindow* parent, int id, const wxBitmap& bmp)
+		: wxStaticBitmap(parent, id, bmp)
+	{
+	};
+
+protected:
+	DECLARE_EVENT_TABLE();
+
+	// Make sure it is truly transparent, i.e. also works with
+	// themed status bars.
+	void OnErase(wxEraseEvent& event)
+	{
+	}
+
+	void OnPaint(wxPaintEvent& event)
+	{
+		wxPaintDC dc(this);
+		dc.DrawBitmap(GetBitmap(), 0, 0, true);
+	}
+};
+
+BEGIN_EVENT_TABLE(wxStaticBitmapEx, wxStaticBitmap)
+EVT_ERASE_BACKGROUND(wxStaticBitmapEx::OnErase)
+EVT_PAINT(wxStaticBitmapEx::OnPaint)
+END_EVENT_TABLE()
+#else
+#define wxStaticBitmapEx wxStaticBitmap
+#endif
+
+class CEncryptionIndicator : public wxStaticBitmapEx
 {
 public:
 	CEncryptionIndicator(CStatusBar* pStatusBar, const wxBitmap& bmp)
-		: wxStaticBitmap(pStatusBar, wxID_ANY, bmp)
+		: wxStaticBitmapEx(pStatusBar, wxID_ANY, bmp)
 	{
 		m_pStatusBar = pStatusBar;
 	}
@@ -229,7 +262,7 @@ protected:
 	}
 };
 
-BEGIN_EVENT_TABLE(CEncryptionIndicator, wxStaticBitmap)
+BEGIN_EVENT_TABLE(CEncryptionIndicator, wxStaticBitmapEx)
 EVT_LEFT_UP(CEncryptionIndicator::OnMouseUp)
 END_EVENT_TABLE()
 
@@ -241,6 +274,9 @@ CStatusBar::CStatusBar(wxTopLevelWindow* pParent)
 	m_pCertificate = 0;
 	m_pSftpEncryptionInfo = 0;
 
+	m_size = 0;
+	m_hasUnknownFiles = false;
+
 	const int count = 5;
 	SetFieldsCount(count);
 	int array[count];
@@ -265,12 +301,14 @@ CStatusBar::~CStatusBar()
 
 void CStatusBar::DisplayQueueSize(wxLongLong totalSize, bool hasUnknown)
 {
+	m_size = totalSize;
+	m_hasUnknownFiles = hasUnknown;
+
 	if (totalSize == 0 && !hasUnknown)
 	{
 		SetStatusText(_("Queue: empty"), FIELD_QUEUESIZE);
 		return;
 	}
-
 	int divider;
 	if (m_sizeFormat == 3)
 		divider = 1000;
@@ -353,7 +391,7 @@ void CStatusBar::DisplayDataType(const CServer* const pServer)
 		wxBitmap bmp = wxArtProvider::GetBitmap(name, wxART_OTHER, wxSize(16, 16));
 		if (!m_pDataTypeIndicator)
 		{
-			m_pDataTypeIndicator = new wxStaticBitmap(this, wxID_ANY, bmp);
+			m_pDataTypeIndicator = new wxStaticBitmapEx(this, wxID_ANY, bmp);
 			AddChild(-4, m_pDataTypeIndicator, 22);
 
 			if (m_pEncryptionIndicator)
@@ -414,6 +452,8 @@ void CStatusBar::UpdateSizeFormat()
 	m_sizeFormat = COptions::Get()->GetOptionVal(OPTION_SIZE_FORMAT);
 	if (!m_sizeFormat)
 		m_sizeFormat = 1;
+
+	DisplayQueueSize(m_size, m_hasUnknownFiles);
 }
 
 void CStatusBar::OnHandleClick(wxWindow* pWnd)
