@@ -2,6 +2,7 @@
 #include "xmlfunctions.h"
 #include "filezillaapp.h"
 #include "Options.h"
+#include <wx/ffile.h>
 
 CXmlFile::CXmlFile(const wxString& fileName)
 {
@@ -179,7 +180,21 @@ void AddTextElement(TiXmlElement* node, const char* name, const wxString& value)
 
 void AddTextElement(TiXmlElement* node, const char* name, int value)
 {
-	AddTextElement(node, name, wxString::Format(_T("%d"), value));
+	char buffer[sizeof(int) * 8]; // Always big enough
+	sprintf(buffer, "%d", value);
+	AddTextElementRaw(node, name, buffer);
+}
+
+void AddTextElementRaw(TiXmlElement* node, const char* name, const char* value)
+{
+	wxASSERT(node);
+	wxASSERT(value && *value);
+
+	TiXmlElement element(name);
+
+    element.InsertEndChild(TiXmlText(value));
+
+	node->InsertEndChild(element);
 }
 
 void AddTextElement(TiXmlElement* node, const wxString& value)
@@ -205,7 +220,35 @@ void AddTextElement(TiXmlElement* node, const wxString& value)
 
 void AddTextElement(TiXmlElement* node, int value)
 {
-	AddTextElement(node, wxString::Format(_T("%d"), value));
+	char buffer[sizeof(int)]; // Always big enough
+	sprintf(buffer, "%d", value);
+	AddTextElementRaw(node, buffer);
+}
+
+void AddTextElementRaw(TiXmlElement* node, const char* value)
+{
+	wxASSERT(node);
+	wxASSERT(value && *value);
+
+	for (TiXmlNode* pChild = node->FirstChild(); pChild; pChild = pChild->NextSibling())
+	{
+		if (!pChild->ToText())
+			continue;
+
+		node->RemoveChild(pChild);
+		break;
+	}
+
+    node->InsertEndChild(TiXmlText(value));
+}
+
+wxString GetTextElement_Trimmed(TiXmlElement* node, const char* name)
+{
+	wxString t = GetTextElement(node, name);
+	t.Trim(true);
+	t.Trim(false);
+
+	return t;
 }
 
 wxString GetTextElement(TiXmlElement* node, const char* name)
@@ -223,6 +266,15 @@ wxString GetTextElement(TiXmlElement* node, const char* name)
 	return ConvLocal(textNode->Value());
 }
 
+wxString GetTextElement_Trimmed(TiXmlElement* node)
+{
+	wxString t = GetTextElement(node);
+	t.Trim(true);
+	t.Trim(false);
+
+	return t;
+}
+
 wxString GetTextElement(TiXmlElement* node)
 {
 	wxASSERT(node);
@@ -310,21 +362,61 @@ wxLongLong GetTextElementLongLong(TiXmlElement* node, const char* name, int defV
 	return negative ? -value : value;
 }
 
+bool GetTextElementBool(TiXmlElement* node, const char* name, bool defValue /*=false*/)
+{
+	wxASSERT(node);
+
+	TiXmlElement* element = node->FirstChildElement(name);
+	if (!element)
+		return defValue;
+
+	TiXmlNode* textNode = element->FirstChild();
+	if (!textNode || !textNode->ToText())
+		return defValue;
+
+	const char* str = textNode->Value();
+	if (!str)
+		return defValue;
+	
+	switch (str[0])
+	{
+	case '0':
+		return false;
+	case '1':
+		return true;
+	default:
+		return defValue;
+	}
+}
+
+TiXmlDocument* LoadXmlDocument(const wxString& file)
+{
+	wxFFile f(file, _T("rb"));
+	if (!f.IsOpened())
+		return 0;
+
+	TiXmlDocument* pXmlDocument = new TiXmlDocument();
+	pXmlDocument->SetCondenseWhiteSpace(false);
+	if (!pXmlDocument->LoadFile(f.fp()))
+	{
+		delete pXmlDocument;
+		return 0;
+	}
+	return pXmlDocument;
+}
+
 // Opens the specified XML file if it exists or creates a new one otherwise.
 // Returns 0 on error.
-TiXmlElement* GetXmlFile(wxFileName file)
+TiXmlElement* GetXmlFile(wxFileName file, bool create /*=true*/)
 {
 	if (wxFileExists(file.GetFullPath()) && file.GetSize() > 0)
 	{
 		// File does exist, open it
 
-		TiXmlDocument* pXmlDocument = new TiXmlDocument();
-		pXmlDocument->SetCondenseWhiteSpace(false);
-		if (!pXmlDocument->LoadFile(file.GetFullPath().mb_str()))
-		{
-			delete pXmlDocument;
+		TiXmlDocument* pXmlDocument = LoadXmlDocument(file.GetFullPath());
+		if (!pXmlDocument)
 			return 0;
-		}
+
 		if (!pXmlDocument->FirstChildElement("FileZilla3"))
 		{
 			delete pXmlDocument;
@@ -342,15 +434,18 @@ TiXmlElement* GetXmlFile(wxFileName file)
 	}
 	else
 	{
-		// File does not exist, create new XML document
-
+		// File does not exist
+		if (!create)
+			return 0;
+		
+		// create new XML document
 		TiXmlDocument* pXmlDocument = new TiXmlDocument();
 		pXmlDocument->SetCondenseWhiteSpace(false);
 		pXmlDocument->InsertEndChild(TiXmlDeclaration("1.0", "UTF-8", "yes"));
 
 		pXmlDocument->InsertEndChild(TiXmlElement("FileZilla3"));
 
-		if (!pXmlDocument->SaveFile(file.GetFullPath().mb_str()))
+		if (!SaveXmlFile(file, pXmlDocument, 0))
 		{
 			delete pXmlDocument;
 			return 0;
@@ -360,7 +455,7 @@ TiXmlElement* GetXmlFile(wxFileName file)
 	}
 }
 
-bool SaveXmlFile(const wxFileName& file, TiXmlNode* node, wxString* error /*=0*/)
+bool SaveXmlFile(const wxFileName& file, TiXmlNode* node, wxString* error /*=0*/, bool move /*=false*/)
 {
 	if (!node)
 		return true;
@@ -373,7 +468,12 @@ bool SaveXmlFile(const wxFileName& file, TiXmlNode* node, wxString* error /*=0*/
 	if (wxFileExists(fullPath))
 	{
 		exists = true;
-		if (!wxCopyFile(fullPath, fullPath + _T("~")))
+		bool res;
+		if (!move)
+			res = wxCopyFile(fullPath, fullPath + _T("~"));
+		else
+			res = wxRenameFile(fullPath, fullPath + _T("~"));
+		if (!res)
 		{
 			const wxString msg = _("Failed to create backup copy of xml file");
 			if (error)
@@ -384,7 +484,8 @@ bool SaveXmlFile(const wxFileName& file, TiXmlNode* node, wxString* error /*=0*/
 		}
 	}
 
-	if (!pDocument->SaveFile(fullPath.mb_str()))
+	wxFFile f(fullPath, _T("w"));
+	if (!f.IsOpened() || !pDocument->SaveFile(f.fp()))
 	{
 		wxRemoveFile(fullPath);
 		if (exists)
@@ -513,7 +614,10 @@ bool GetServer(TiXmlElement *node, CServer& server)
 	}
 
 	server.SetBypassProxy(GetTextElementInt(node, "BypassProxy", false) == 1);
-	server.SetName(GetTextElement(node, "Name"));
+	server.SetName(GetTextElement_Trimmed(node, "Name"));
+
+	if (server.GetName().empty())
+		server.SetName(GetTextElement_Trimmed(node));
 
 	return true;
 }
@@ -568,13 +672,13 @@ void SetServer(TiXmlElement *node, const CServer& server)
 	switch (server.GetPasvMode())
 	{
 	case MODE_PASSIVE:
-		AddTextElement(node, "PasvMode", _T("MODE_PASSIVE"));
+		AddTextElementRaw(node, "PasvMode", "MODE_PASSIVE");
 		break;
 	case MODE_ACTIVE:
-		AddTextElement(node, "PasvMode", _T("MODE_ACTIVE"));
+		AddTextElementRaw(node, "PasvMode", "MODE_ACTIVE");
 		break;
 	default:
-		AddTextElement(node, "PasvMode", _T("MODE_DEFAULT"));
+		AddTextElementRaw(node, "PasvMode", "MODE_DEFAULT");
 		break;
 	}
 	AddTextElement(node, "MaximumMultipleConnections", server.MaximumMultipleConnections());
@@ -582,13 +686,13 @@ void SetServer(TiXmlElement *node, const CServer& server)
 	switch (server.GetEncodingType())
 	{
 	case ENCODING_AUTO:
-		AddTextElement(node, "EncodingType", _T("Auto"));
+		AddTextElementRaw(node, "EncodingType", "Auto");
 		break;
 	case ENCODING_UTF8:
-		AddTextElement(node, "EncodingType", _T("UTF-8"));
+		AddTextElementRaw(node, "EncodingType", "UTF-8");
 		break;
 	case ENCODING_CUSTOM:
-		AddTextElement(node, "EncodingType", _T("Custom"));
+		AddTextElementRaw(node, "EncodingType", "Custom");
 		AddTextElement(node, "CustomEncoding", server.GetCustomEncoding());
 		break;
 	}
@@ -605,7 +709,7 @@ void SetServer(TiXmlElement *node, const CServer& server)
 		}
 	}
 
-	AddTextElement(node, "BypassProxy", server.GetBypassProxy() ? 1 : 0);
+	AddTextElementRaw(node, "BypassProxy", server.GetBypassProxy() ? "1" : "0");
 	const wxString& name = server.GetName();
 	if (name != _T(""))
 		AddTextElement(node, "Name", name);
