@@ -70,30 +70,38 @@ TiXmlElement* FindElementWithAttribute(TiXmlElement* node, const char* element,
 // Add a new child element with the specified name and value to the xml document
 void AddTextElement(TiXmlElement* node, const char* name, const wxString& value);
 void AddTextElement(TiXmlElement* node, const char* name, int value);
+void AddTextElementRaw(TiXmlElement* node, const char* name, const char* value);
 
 // Set the current element's text value
 void AddTextElement(TiXmlElement* node, const wxString& value);
 void AddTextElement(TiXmlElement* node, int value);
+void AddTextElementRaw(TiXmlElement* node, const char* value);
 
 // Get string from named child element
 wxString GetTextElement(TiXmlElement* node, const char* name);
+wxString GetTextElement_Trimmed(TiXmlElement* node, const char* name);
 
 // Get string from current element
 wxString GetTextElement(TiXmlElement* node);
+wxString GetTextElement_Trimmed(TiXmlElement* node);
 
 // Get (64-bit) integer from named element
 int GetTextElementInt(TiXmlElement* node, const char* name, int defValue = 0);
 wxLongLong GetTextElementLongLong(TiXmlElement* node, const char* name, int defValue = 0);
 
+bool GetTextElementBool(TiXmlElement* node, const char* name, bool defValue = false);
+
 // Opens the specified XML file if it exists or creates a new one otherwise.
 // Returns 0 on error.
-TiXmlElement* GetXmlFile(wxFileName file);
+TiXmlElement* GetXmlFile(wxFileName file, bool create = true);
 
 // Save the XML document to the given file
-bool SaveXmlFile(const wxFileName& file, TiXmlNode* node, wxString* error = 0);
+bool SaveXmlFile(const wxFileName& file, TiXmlNode* node, wxString* error = 0, bool move = false);
 
 // Functions to save and retrieve CServer objects to the XML file
 void SetServer(TiXmlElement *node, const CServer& server);
 bool GetServer(TiXmlElement *node, CServer& server);
 
+TiXmlDocument* LoadXmlDocument(const wxString& file);
+
 #endif //__XMLFUNCTIONS_H__
