@@ -37,6 +37,29 @@
 # define X11_UNIX_PATH "/tmp/.X11-unix/X"
 #endif
 
+/*
+ * We used to typedef struct Socket_tag *Socket.
+ *
+ * Since we have made the networking abstraction slightly more
+ * abstract, Socket no longer means a tcp socket (it could mean
+ * an ssl socket).  So now we must use Actual_Socket when we know
+ * we are talking about a tcp socket.
+ */
+typedef struct Socket_tag *Actual_Socket;
+
+/*
+ * Mutable state that goes with a SockAddr: stores information
+ * about where in the list of candidate IP(v*) addresses we've
+ * currently got to.
+ */
+typedef struct SockAddrStep_tag SockAddrStep;
+struct SockAddrStep_tag {
+#ifndef NO_IPV6
+    struct addrinfo *ai;	       /* steps along addr->ais */
+#endif
+    int curraddr;
+};
+
 struct Socket_tag {
     struct socket_function_table *fn;
     /* the above variable absolutely *must* be the first in this structure */
@@ -60,42 +83,63 @@ struct Socket_tag {
     int nodelay, keepalive;            /* for connect()-type sockets */
     int privport, port;                /* and again */
     SockAddr addr;
+    SockAddrStep step;
 
     unsigned int lastSendNotificationLow;
     time_t lastSendNotificationHigh;
     unsigned int lastRecvNotificationLow;
     time_t lastRecvNotificationHigh;
+    /*
+     * We sometimes need pairs of Socket structures to be linked:
+     * if we are listening on the same IPv6 and v4 port, for
+     * example. So here we define `parent' and `child' pointers to
+     * track this link.
+     */
+    Actual_Socket parent, child;
 };
 
-/*
- * We used to typedef struct Socket_tag *Socket.
- *
- * Since we have made the networking abstraction slightly more
- * abstract, Socket no longer means a tcp socket (it could mean
- * an ssl socket).  So now we must use Actual_Socket when we know
- * we are talking about a tcp socket.
- */
-typedef struct Socket_tag *Actual_Socket;
-
 struct SockAddr_tag {
+    int refcount;
     const char *error;
-    /*
-     * Which address family this address belongs to. AF_INET for
-     * IPv4; AF_INET6 for IPv6; AF_UNSPEC indicates that name
-     * resolution has not been done and a simple host name is held
-     * in this SockAddr structure.
-     */
-    int family;
+    enum { UNRESOLVED, UNIX, IP } superfamily;
 #ifndef NO_IPV6
     struct addrinfo *ais;	       /* Addresses IPv6 style. */
-    struct addrinfo *ai;	       /* steps along the linked list */
 #else
     unsigned long *addresses;	       /* Addresses IPv4 style. */
-    int naddresses, curraddr;
+    int naddresses;
 #endif
     char hostname[512];		       /* Store an unresolved host name. */
 };
 
+/*
+ * Which address family this address belongs to. AF_INET for IPv4;
+ * AF_INET6 for IPv6; AF_UNSPEC indicates that name resolution has
+ * not been done and a simple host name is held in this SockAddr
+ * structure.
+ */
+#ifndef NO_IPV6
+#define SOCKADDR_FAMILY(addr, step) \
+    ((addr)->superfamily == UNRESOLVED ? AF_UNSPEC : \
+     (addr)->superfamily == UNIX ? AF_UNIX : \
+     (step).ai ? (step).ai->ai_family : AF_INET)
+#else
+#define SOCKADDR_FAMILY(addr, step) \
+    ((addr)->superfamily == UNRESOLVED ? AF_UNSPEC : \
+     (addr)->superfamily == UNIX ? AF_UNIX : AF_INET)
+#endif
+
+/*
+ * Start a SockAddrStep structure to step through multiple
+ * addresses.
+ */
+#ifndef NO_IPV6
+#define START_STEP(addr, step) \
+    ((step).ai = (addr)->ais, (step).curraddr = 0)
+#else
+#define START_STEP(addr, step) \
+    ((step).curraddr = 0)
+#endif
+
 // 1/10th of a second in microseconds
 const static int notificationDelay = 1000000 / 10;
 
@@ -212,9 +256,9 @@ static int cmpfortree(void *av, void *bv)
     if (as > bs)
 	return +1;
     if (a < b)
-	return -1;
+       return -1;
     if (a > b)
-	return +1;
+       return +1;
     return 0;
 }
 
@@ -261,9 +305,10 @@ SockAddr sk_namelookup(const char *host, char **canonicalname, int address_famil
 
     /* Clear the structure and default to IPv4. */
     memset(ret, 0, sizeof(struct SockAddr_tag));
-    ret->family = 0;		       /* We set this one when we have resolved the host. */
+    ret->superfamily = UNRESOLVED;
     *realhost = '\0';
     ret->error = NULL;
+    ret->refcount = 1;
 
 #ifndef NO_IPV6
     hints.ai_flags = AI_CANONNAME;
@@ -277,15 +322,14 @@ SockAddr sk_namelookup(const char *host, char **canonicalname, int address_famil
     hints.ai_canonname = NULL;
     hints.ai_next = NULL;
     err = getaddrinfo(host, NULL, &hints, &ret->ais);
-    ret->ai = ret->ais;
     if (err != 0) {
 	ret->error = gai_strerror(err);
 	return ret;
     }
-    ret->family = ret->ai->ai_family;
+    ret->superfamily = IP;
     *realhost = '\0';
-    if (ret->ai->ai_canonname != NULL)
-	strncat(realhost, ret->ai->ai_canonname, sizeof(realhost) - 1);
+    if (ret->ais->ai_canonname != NULL)
+	strncat(realhost, ret->ais->ai_canonname, sizeof(realhost) - 1);
     else
 	strncat(realhost, host, sizeof(realhost) - 1);
 #else
@@ -294,12 +338,12 @@ SockAddr sk_namelookup(const char *host, char **canonicalname, int address_famil
 	 * Otherwise use the IPv4-only gethostbyname... (NOTE:
 	 * we don't use gethostbyname as a fallback!)
 	 */
-	if (ret->family == 0) {
+	if (ret->superfamily == UNRESOLVED) {
 	    /*debug(("Resolving \"%s\" with gethostbyname() (IPv4 only)...\n", host)); */
 	    if ( (h = gethostbyname(host)) )
-		ret->family = AF_INET;
+		ret->superfamily = IP;
 	}
-	if (ret->family == 0) {
+	if (ret->superfamily == UNRESOLVED) {
 	    ret->error = (h_errno == HOST_NOT_FOUND ||
 			  h_errno == NO_DATA ||
 			  h_errno == NO_ADDRESS ? "Host does not exist" :
@@ -322,12 +366,11 @@ SockAddr sk_namelookup(const char *host, char **canonicalname, int address_famil
 	 * This must be a numeric IPv4 address because it caused a
 	 * success return from inet_addr.
 	 */
-	ret->family = AF_INET;
+	ret->superfamily = IP;
 	strncpy(realhost, host, sizeof(realhost));
 	ret->addresses = snew(unsigned long);
 	ret->naddresses = 1;
 	ret->addresses[0] = ntohl(a);
-	ret->curraddr = 0;
     }
 #endif
     realhost[lenof(realhost)-1] = '\0';
@@ -340,7 +383,7 @@ SockAddr sk_nonamelookup(const char *host)
 {
     SockAddr ret = snew(struct SockAddr_tag);
     ret->error = NULL;
-    ret->family = AF_UNSPEC;
+    ret->superfamily = UNRESOLVED;
     strncpy(ret->hostname, host, lenof(ret->hostname));
     ret->hostname[lenof(ret->hostname)-1] = '\0';
 #ifndef NO_IPV6
@@ -348,45 +391,47 @@ SockAddr sk_nonamelookup(const char *host)
 #else
     ret->addresses = NULL;
 #endif
+    ret->refcount = 1;
     return ret;
 }
 
-static int sk_nextaddr(SockAddr addr)
+static int sk_nextaddr(SockAddr addr, SockAddrStep *step)
 {
 #ifndef NO_IPV6
-    if (addr->ai && addr->ai->ai_next) {
-	addr->ai = addr->ai->ai_next;
-	addr->family = addr->ai->ai_family;
+    if (step->ai && step->ai->ai_next) {
+	step->ai = step->ai->ai_next;
 	return TRUE;
     } else
 	return FALSE;
 #else
-    if (addr->curraddr+1 < addr->naddresses) {
-	addr->curraddr++;
+    if (step->curraddr+1 < addr->naddresses) {
+	step->curraddr++;
 	return TRUE;
     } else {
 	return FALSE;
     }
-#endif
+#endif    
 }
 
 void sk_getaddr(SockAddr addr, char *buf, int buflen)
 {
 
-    if (addr->family == AF_UNSPEC) {
+    if (addr->superfamily == UNRESOLVED) {
 	strncpy(buf, addr->hostname, buflen);
 	buf[buflen-1] = '\0';
     } else {
 #ifndef NO_IPV6
-	if (getnameinfo(addr->ai->ai_addr, addr->ai->ai_addrlen, buf, buflen,
+	if (getnameinfo(addr->ais->ai_addr, addr->ais->ai_addrlen, buf, buflen,
 			NULL, 0, NI_NUMERICHOST) != 0) {
 	    buf[0] = '\0';
 	    strncat(buf, "<unknown>", buflen - 1);
 	}
 #else
 	struct in_addr a;
-	assert(addr->family == AF_INET);
-	a.s_addr = htonl(addr->addresses[addr->curraddr]);
+	SockAddrStep step;
+	START_STEP(addr, step);
+	assert(SOCKADDR_FAMILY(addr, step) == AF_INET);
+	a.s_addr = htonl(addr->addresses[0]);
 	strncpy(buf, inet_ntoa(a), buflen);
 	buf[buflen-1] = '\0';
 #endif
@@ -427,15 +472,17 @@ static int sockaddr_is_loopback(struct sockaddr *sa)
 int sk_address_is_local(SockAddr addr)
 {
 
-    if (addr->family == AF_UNSPEC)
+    if (addr->superfamily == UNRESOLVED)
 	return 0;                      /* we don't know; assume not */
     else {
 #ifndef NO_IPV6
-	return sockaddr_is_loopback(addr->ai->ai_addr);
+	return sockaddr_is_loopback(addr->ais->ai_addr);
 #else
 	struct in_addr a;
-	assert(addr->family == AF_INET);
-	a.s_addr = htonl(addr->addresses[addr->curraddr]);
+	SockAddrStep step;
+	START_STEP(addr, step);
+	assert(SOCKADDR_FAMILY(addr, step) == AF_INET);
+	a.s_addr = htonl(addr->addresses[0]);
 	return ipv4_is_loopback(a);
 #endif
     }
@@ -443,37 +490,47 @@ int sk_address_is_local(SockAddr addr)
 
 int sk_addrtype(SockAddr addr)
 {
-    return (addr->family == AF_INET ? ADDRTYPE_IPV4 :
+    SockAddrStep step;
+    int family;
+    START_STEP(addr, step);
+    family = SOCKADDR_FAMILY(addr, step);
+
+    return (family == AF_INET ? ADDRTYPE_IPV4 :
 #ifndef NO_IPV6
-	    addr->family == AF_INET6 ? ADDRTYPE_IPV6 :
+	    family == AF_INET6 ? ADDRTYPE_IPV6 :
 #endif
 	    ADDRTYPE_NAME);
 }
 
 void sk_addrcopy(SockAddr addr, char *buf)
 {
+    SockAddrStep step;
+    int family;
+    START_STEP(addr, step);
+    family = SOCKADDR_FAMILY(addr, step);
 
 #ifndef NO_IPV6
-    if (addr->family == AF_INET)
-	memcpy(buf, &((struct sockaddr_in *)addr->ai->ai_addr)->sin_addr,
+    if (family == AF_INET)
+	memcpy(buf, &((struct sockaddr_in *)step.ai->ai_addr)->sin_addr,
 	       sizeof(struct in_addr));
-    else if (addr->family == AF_INET6)
-	memcpy(buf, &((struct sockaddr_in6 *)addr->ai->ai_addr)->sin6_addr,
+    else if (family == AF_INET6)
+	memcpy(buf, &((struct sockaddr_in6 *)step.ai->ai_addr)->sin6_addr,
 	       sizeof(struct in6_addr));
     else
 	assert(FALSE);
 #else
     struct in_addr a;
 
-    assert(addr->family == AF_INET);
-    a.s_addr = htonl(addr->addresses[addr->curraddr]);
+    assert(family == AF_INET);
+    a.s_addr = htonl(addr->addresses[step.curraddr]);
     memcpy(buf, (char*) &a.s_addr, 4);
 #endif
 }
 
 void sk_addr_free(SockAddr addr)
 {
-
+    if (--addr->refcount > 0)
+	return;
 #ifndef NO_IPV6
     if (addr->ais != NULL)
 	freeaddrinfo(addr->ais);
@@ -483,6 +540,12 @@ void sk_addr_free(SockAddr addr)
     sfree(addr);
 }
 
+SockAddr sk_addr_dup(SockAddr addr)
+{
+    addr->refcount++;
+    return addr;
+}
+
 static Plug sk_tcp_plug(Socket sock, Plug p)
 {
     Actual_Socket s = (Actual_Socket) sock;
@@ -540,6 +603,7 @@ Socket sk_register(OSSocket sockfd, Plug plug)
     ret->pending_error = 0;
     ret->oobpending = FALSE;
     ret->listener = 0;
+    ret->parent = ret->child = NULL;
     ret->addr = NULL;
     ret->connected = 1;
 
@@ -569,7 +633,7 @@ static int try_connect(Actual_Socket sock)
     const struct sockaddr *sa;
     int err = 0;
     short localport;
-    int fl, salen;
+    int fl, salen, family;
 
     /*
      * Remove the socket from the tree before we overwrite its
@@ -587,8 +651,9 @@ static int try_connect(Actual_Socket sock)
     /*
      * Open socket.
      */
-    assert(sock->addr->family != AF_UNSPEC);
-    s = socket(sock->addr->family, SOCK_STREAM, 0);
+    family = SOCKADDR_FAMILY(sock->addr, sock->step);
+    assert(family != AF_UNSPEC);
+    s = socket(family, SOCK_STREAM, 0);
     sock->s = s;
 
     if (s < 0) {
@@ -629,13 +694,13 @@ static int try_connect(Actual_Socket sock)
 
     /* We don't try to bind to a local address for UNIX domain sockets.  (Why
      * do we bother doing the bind when localport == 0 anyway?) */
-    if(sock->addr->family != AF_UNIX) {
+    if (family != AF_UNIX) {
 	/* Loop round trying to bind */
 	while (1) {
 	    int retcode;
 
 #ifndef NO_IPV6
-	    if (sock->addr->family == AF_INET6) {
+	    if (family == AF_INET6) {
 		/* XXX use getaddrinfo to get a local address? */
 		a6.sin6_family = AF_INET6;
 		a6.sin6_addr = in6addr_any;
@@ -644,7 +709,7 @@ static int try_connect(Actual_Socket sock)
 	    } else
 #endif
 	    {
-		assert(sock->addr->family == AF_INET);
+		assert(family == AF_INET);
 		a.sin_family = AF_INET;
 		a.sin_addr.s_addr = htonl(INADDR_ANY);
 		a.sin_port = htons(localport);
@@ -658,14 +723,14 @@ static int try_connect(Actual_Socket sock)
 		if (err != EADDRINUSE) /* failed, for a bad reason */
 		  break;
 	    }
-
+	    
 	    if (localport == 0)
 	      break;		       /* we're only looping once */
 	    localport--;
 	    if (localport == 0)
 	      break;		       /* we might have got to the end */
 	}
-
+	
 	if (err)
 	    goto ret;
     }
@@ -673,25 +738,25 @@ static int try_connect(Actual_Socket sock)
     /*
      * Connect to remote address.
      */
-    switch(sock->addr->family) {
+    switch(family) {
 #ifndef NO_IPV6
       case AF_INET:
 	/* XXX would be better to have got getaddrinfo() to fill in the port. */
-	((struct sockaddr_in *)sock->addr->ai->ai_addr)->sin_port =
+	((struct sockaddr_in *)sock->step.ai->ai_addr)->sin_port =
 	    htons(sock->port);
-	sa = (const struct sockaddr *)sock->addr->ai->ai_addr;
-	salen = sock->addr->ai->ai_addrlen;
+	sa = (const struct sockaddr *)sock->step.ai->ai_addr;
+	salen = sock->step.ai->ai_addrlen;
 	break;
       case AF_INET6:
-	((struct sockaddr_in *)sock->addr->ai->ai_addr)->sin_port =
+	((struct sockaddr_in *)sock->step.ai->ai_addr)->sin_port =
 	    htons(sock->port);
-	sa = (const struct sockaddr *)sock->addr->ai->ai_addr;
-	salen = sock->addr->ai->ai_addrlen;
+	sa = (const struct sockaddr *)sock->step.ai->ai_addr;
+	salen = sock->step.ai->ai_addrlen;
 	break;
 #else
       case AF_INET:
 	a.sin_family = AF_INET;
-	a.sin_addr.s_addr = htonl(sock->addr->addresses[sock->addr->curraddr]);
+	a.sin_addr.s_addr = htonl(sock->addr->addresses[sock->step.curraddr]);
 	a.sin_port = htons((short) sock->port);
 	sa = (const struct sockaddr *)&a;
 	salen = sizeof a;
@@ -765,9 +830,11 @@ Socket sk_new(SockAddr addr, int port, int privport, int oobinline,
     ret->frozen_readable = 0;
     ret->localhost_only = 0;	       /* unused, but best init anyway */
     ret->pending_error = 0;
+    ret->parent = ret->child = NULL;
     ret->oobpending = FALSE;
     ret->listener = 0;
     ret->addr = addr;
+    START_STEP(ret->addr, ret->step);
     ret->s = -1;
     ret->oobinline = oobinline;
     ret->nodelay = nodelay;
@@ -778,7 +845,7 @@ Socket sk_new(SockAddr addr, int port, int privport, int oobinline,
     err = 0;
     do {
         err = try_connect(ret);
-    } while (err && sk_nextaddr(ret->addr));
+    } while (err && sk_nextaddr(ret->addr, &ret->step));
 
     if (err)
         ret->error = strerror(err);
@@ -786,7 +853,7 @@ Socket sk_new(SockAddr addr, int port, int privport, int oobinline,
     return (Socket) ret;
 }
 
-Socket sk_newlistener(char *srcaddr, int port, Plug plug, int local_host_only, int address_family)
+Socket sk_newlistener(char *srcaddr, int port, Plug plug, int local_host_only, int orig_address_family)
 {
     int s;
 #ifndef NO_IPV6
@@ -799,6 +866,7 @@ Socket sk_newlistener(char *srcaddr, int port, Plug plug, int local_host_only, i
     struct sockaddr_in a;
     Actual_Socket ret;
     int retcode;
+    int address_family;
     int on = 1;
 
     /*
@@ -815,6 +883,7 @@ Socket sk_newlistener(char *srcaddr, int port, Plug plug, int local_host_only, i
     ret->frozen_readable = 0;
     ret->localhost_only = local_host_only;
     ret->pending_error = 0;
+    ret->parent = ret->child = NULL;
     ret->oobpending = FALSE;
     ret->listener = 1;
     ret->addr = NULL;
@@ -823,9 +892,9 @@ Socket sk_newlistener(char *srcaddr, int port, Plug plug, int local_host_only, i
      * Translate address_family from platform-independent constants
      * into local reality.
      */
-    address_family = (address_family == ADDRTYPE_IPV4 ? AF_INET :
+    address_family = (orig_address_family == ADDRTYPE_IPV4 ? AF_INET :
 #ifndef NO_IPV6
-		      address_family == ADDRTYPE_IPV6 ? AF_INET6 :
+		      orig_address_family == ADDRTYPE_IPV6 ? AF_INET6 :
 #endif
 		      AF_UNSPEC);
 
@@ -937,6 +1006,32 @@ Socket sk_newlistener(char *srcaddr, int port, Plug plug, int local_host_only, i
 	return (Socket) ret;
     }
 
+#ifndef NO_IPV6
+    /*
+     * If we were given ADDRTYPE_UNSPEC, we must also create an
+     * IPv4 listening socket and link it to this one.
+     */
+    if (address_family == AF_INET6 && orig_address_family == ADDRTYPE_UNSPEC) {
+        Actual_Socket other;
+
+        other = (Actual_Socket) sk_newlistener(srcaddr, port, plug,
+                                               local_host_only, ADDRTYPE_IPV4);
+
+        if (other) {
+            if (!other->error) {
+                other->parent = ret;
+                ret->child = other;
+            } else {
+                /* If we couldn't create a listening socket on IPv4 as well
+                 * as IPv6, we must return an error overall. */
+                close(s);
+                sfree(ret);
+                return (Socket) other;
+            }
+        }
+    }
+#endif
+
     ret->s = s;
 
     uxsel_tell(ret);
@@ -949,6 +1044,9 @@ static void sk_tcp_close(Socket sock)
 {
     Actual_Socket s = (Actual_Socket) sock;
 
+    if (s->child)
+        sk_tcp_close((Socket)s->child);
+
     uxsel_del(s->s);
     del234(sktree, s);
     close(s->s);
@@ -1163,7 +1261,7 @@ static int net_select_result(int fd, int event)
 				    strerror(errno), errno, 0);
 	    } else {
 		UpdateQuota(0, ret);
-                RecvNotification(s);
+		RecvNotification(s);
                 /*
                  * Receiving actual data on a socket means we can
                  * stop falling back through the candidate
@@ -1262,7 +1360,7 @@ static int net_select_result(int fd, int event)
             int err = errno;
 	    if (s->addr) {
 		plug_log(s->plug, 1, s->addr, s->port, strerror(err), err);
-		while (s->addr && sk_nextaddr(s->addr)) {
+		while (s->addr && sk_nextaddr(s->addr, &s->step)) {
 		    err = try_connect(s);
 		}
 	    }
@@ -1272,7 +1370,7 @@ static int net_select_result(int fd, int event)
 	    return plug_closing(s->plug, NULL, 0, 0);
 	} else {
 	    UpdateQuota(0, ret);
-            RecvNotification(s);
+	    RecvNotification(s);
             /*
              * Receiving actual data on a socket means we can
              * stop falling back through the candidate
@@ -1323,7 +1421,7 @@ void net_pending_errors(void)
      * others to be closed. (I can't think of any reason this might
      * happen in current SSH implementation, but to maintain
      * generality of this network layer I'll assume the worst.)
-     *
+     * 
      * So what we'll do is search the socket list for _one_ socket
      * with a pending error, and then handle it, and then search
      * the list again _from the beginning_. Repeat until we make a
@@ -1417,40 +1515,36 @@ int net_service_lookup(char *service)
 	return 0;
 }
 
-SockAddr platform_get_x11_unix_address(const char *display, int displaynum,
-				       char **canonicalname)
+SockAddr platform_get_x11_unix_address(const char *sockpath, int displaynum)
 {
     SockAddr ret = snew(struct SockAddr_tag);
     int n;
 
     memset(ret, 0, sizeof *ret);
-    ret->family = AF_UNIX;
+    ret->superfamily = UNIX;
     /*
-     * Mac OS X Leopard uses an innovative X display naming
-     * convention in which the entire display name is the path to
-     * the Unix socket, including the trailing :0 which only
-     * _looks_ like a display number. Heuristically, I think
-     * detecting this by means of a leading slash ought to be
-     * adequate.
+     * In special circumstances (notably Mac OS X Leopard), we'll
+     * have been passed an explicit Unix socket path.
      */
-    if (display[0] == '/') {
+    if (sockpath) {
 	n = snprintf(ret->hostname, sizeof ret->hostname,
-		     "%s", display);
+		     "%s", sockpath);
     } else {
 	n = snprintf(ret->hostname, sizeof ret->hostname,
 		     "%s%d", X11_UNIX_PATH, displaynum);
     }
-    if(n < 0)
+
+    if (n < 0)
 	ret->error = "snprintf failed";
-    else if(n >= sizeof ret->hostname)
+    else if (n >= sizeof ret->hostname)
 	ret->error = "X11 UNIX name too long";
-    else
-	*canonicalname = dupstr(ret->hostname);
+
 #ifndef NO_IPV6
-    ret->ai = ret->ais = NULL;
+    ret->ais = NULL;
 #else
     ret->addresses = NULL;
-    ret->curraddr = ret->naddresses = 0;
+    ret->naddresses = 0;
 #endif
+    ret->refcount = 1;
     return ret;
 }
